Package: nmw
Version: 0.3.1
Date: 2026-06-09
Title: Understanding Nonlinear Mixed Effects Modeling for Population
        Pharmacokinetics
Description: This shows how 'NONMEM' (Beal SL, Sheiner LB, Boeckmann AJ,
    Bauer RJ.  NONMEM 7.5 Users Guides.  Icon plc, 2020) software works.
    'NONMEM' classical estimation methods such as 'First Order (FO)
    approximation', 'First Order Conditional Estimation (FOCE)', and
    'Laplacian approximation' are explained.  Functions are also provided
    for post-run processing of NONMEM output files, generating PDF
    diagnostic reports including objective function value analysis,
    parameter estimates, prediction and residual diagnostics, empirical
    Bayes estimate (EBE) analysis, input data summary, and individual
    pharmacokinetic parameter distributions.  Helper utilities for
    building NONMEM-ready datasets from SDTM-style source tables are
    also included.
Depends: R (>= 3.5.0), numDeriv
Imports: MASS, grDevices, graphics, stats, utils
ByteCompile: yes
License: GPL-3
Copyright: 2017-, Kyun-Seop Bae
Author: Kyun-Seop Bae [aut, cre]
Maintainer: Kyun-Seop Bae <k@acr.kr>
Authors@R: person(given = "Kyun-Seop",
                  family = "Bae",
                  role = c("aut", "cre"),
                  email = "k@acr.kr")
URL: https://cran.r-project.org/package=nmw
Contact: Kyun-Seop Bae <k@acr.kr>
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: no
RoxygenNote: 7.3.3
Packaged: 2026-06-09 03:10:20 UTC; Kyun-SeopBae
Repository: CRAN
Date/Publication: 2026-06-09 07:00:08 UTC
Built: R 4.6.0; ; 2026-06-09 07:18:17 UTC; unix

# Exports from nmw_0.1.6 (estimation methods)
export(AddCox)
export(CalcSmat)
export(CombDmExPc)
export(CovStep)
export(DECN)
export(DesclVar)
export(EstStep)
export(InitStep)
export(Mx)
export(Obj)
export(ObjEta)
export(Oi)
export(Oi0)
export(Oi1)
export(OiS0)
export(OiS1)
export(PostHocEta)
export(ScaleVar)
export(SqrtInvCov)
export(TabStep)
export(TrimOut)
export(ltv2mat)
export(mat2ltv)
export(mat2utv)
export(utv2mat)

# Exports from nmw_0.3.0 (NM dataset construction helpers)
export(add_crcl_cg)
export(build_dose_records)
export(build_nm_dataset)
export(build_obs_records)
export(code_race)
export(code_sex)
export(crcl_cg)
export(dat2_time_to_posix)
export(locf_value)
export(merge_cov_locf)
export(nm_to_molar)
export(parse_dtc)

# Exports from nma_0.1.0 (post-run processing)
export(AddDoNoTaLD)
export(AddPage)
export(BtwTagMat)
export(BtwTagVals)
export(ClosePDF)
export(DxPlotPost)
export(EBEpair)
export(ExpandDoseHist)
export(GetAICc)
export(GetCountEps)
export(GetCountEta)
export(GetCountObs)
export(GetCountPara)
export(GetEstMethod)
export(GetOFV)
export(LogDV)
export(LogDV2)
export(MinSuccess)
export(NMIDStat)
export(NMVarStat)
export(Outline)
export(ParseOut)
export(PrepPDF)
export(PrinMTxt)
export(PrinTxt)
export(RemoveNA)
export(ResTest)
export(SESuccess)
export(SumOut)
export(mlr2)
export(nmw_report_ebe)
export(nmw_report_indipk)
export(nmw_report_input)
export(nmw_report_ofv)
export(nmw_report_output)
export(nmw_report_param)
export(nmw_report_pred)
export(nmw_report_resid)
export(nmw_run)
export(run.test.nm)

# Imports
importFrom(numDeriv, grad, hessian)
importFrom(MASS, ginv)
importFrom(grDevices, pdf, dev.off)
importFrom(graphics, plot, lines, text, par, hist, boxplot, pairs, mtext,
           abline, points, rect, screen, split.screen, close.screen, layout,
           strwidth)
importFrom(stats, density, shapiro.test, t.test, anova, lm, formula, cor,
           cov, cov2cor, lowess, qnorm, pnorm, qbinom, pbinom, sd, var,
           pt, qqnorm, setNames, optim, median)
importFrom(utils, read.table, read.csv, read.fwf, write.csv, capture.output,
           head, tail)

require(nmw)

DataAll = Theoph
colnames(DataAll) = c("ID", "BWT", "DOSE", "TIME", "DV")
DataAll[,"ID"] = as.numeric(as.character(DataAll[,"ID"]))

require(lattice)
xyplot(DV ~ TIME | as.factor(ID), data=DataAll, type="b")

nTheta = 3
nEta = 3
nEps = 2

THETAinit = c(2, 50, 0.1)
OMinit = matrix(c(0.2, 0.1, 0.1, 0.1, 0.2, 0.1, 0.1, 0.1, 0.2), 
                nrow=nEta, ncol=nEta) ; OMinit
SGinit = diag(c(0.1, 0.1)) ; SGinit

LB = rep(0, nTheta)
UB = rep(1000000, nTheta)

FGD = deriv(~DOSE/(TH2*exp(ETA2))*TH1*exp(ETA1)/(TH1*exp(ETA1) - TH3*exp(ETA3))*
             (exp(-TH3*exp(ETA3)*TIME)-exp(-TH1*exp(ETA1)*TIME)),
            c("ETA1","ETA2","ETA3"),
            function.arg=c("TH1", "TH2", "TH3", "ETA1", "ETA2", "ETA3", "DOSE", 
                           "TIME"),
            func=TRUE, hessian=(e$METHOD == "LAPL"))
H = deriv(~F + F*EPS1 + EPS2, c("EPS1", "EPS2"), 
          function.arg=c("F", "EPS1", "EPS2"), func=TRUE)

PRED = function(THETA, ETA, DATAi)
{
  FGDres = FGD(THETA[1], THETA[2], THETA[3], ETA[1], ETA[2], ETA[3], DOSE=320, 
               DATAi[,"TIME"])
  Gres   = attr(FGDres, "gradient")
  Hres   = attr(H(FGDres, 0, 0), "gradient")

  if (e$METHOD == "LAPL") {
    Dres = attr(FGDres, "hessian")
    Res  = cbind(FGDres, Gres, Hres, Dres[,1,1], 
                 Dres[,2,1], Dres[,2,2], Dres[,3,])
    colnames(Res) = c("F", "G1", "G2", "G3", "H1", "H2", 
                      "D11", "D21", "D22", "D31", "D32", "D33")
  } else {
    Res = cbind(FGDres, Gres, Hres)
    colnames(Res) = c("F", "G1", "G2", "G3", "H1", "H2")
  }
  return(Res)
}

####### First Order Approximation Method
InitStep(DataAll, THETAinit=THETAinit, OMinit=OMinit, SGinit=SGinit, 
         LB=LB, UB=UB, Pred=PRED, METHOD="ZERO")
(EstRes = EstStep())           # 4 sec
(CovRes = CovStep())           # 2 sec
PostHocEta() # FinalPara from EstStep()
TabStep()

######## First Order Conditional Estimation with Interaction Method
InitStep(DataAll, THETAinit=THETAinit, OMinit=OMinit, SGinit=SGinit, 
         LB=LB, UB=UB, Pred=PRED, METHOD="COND")
(EstRes = EstStep())           # 1.7 min
(CovRes = CovStep())           # 44 sec
get("EBE", envir=e)
TabStep()

######## Laplacian Approximation with Interaction Method
InitStep(DataAll, THETAinit=THETAinit, OMinit=OMinit, SGinit=SGinit, 
         LB=LB, UB=UB, Pred=PRED, METHOD="LAPL")
(EstRes = EstStep())           # 3.4 min
(CovRes = CovStep())           # 52 sec
get("EBE", envir=e)
TabStep()

#' Generate OFV Diagnostic Report (S1-OFV.PDF)
#'
#' Generates a PDF report with objective function value diagnostics including
#' total OFV, AICc, BIC, individual OFV per DV summary, and gradient analysis.
#'
#' @param run_dir character, path to the NONMEM run directory (default: current directory)
#' @export
nmw_report_ofv <- function(run_dir = getwd()) {
  owd <- setwd(run_dir)
  on.exit(setwd(owd))

  defpar <- par(no.readonly = TRUE)
  defpar$new <- NULL
  on.exit(par(defpar), add = TRUE)

  CtlName <- GetCurModelName()

  XML <- readLines(paste0(CtlName, ".xml"))
  EXT <- read.table(paste0(CtlName, ".ext"), skip = 1, header = TRUE)
  GRD <- read.table(paste0(CtlName, ".grd"), skip = 1, header = TRUE)
  PHI <- read.table(paste0(CtlName, ".phi"), skip = 1, header = TRUE)

  EXT <- EXT[EXT[, "ITERATION"] >= 0, ]
  EXT <- RmvFixed(EXT)

  IDs <- PHI[, "ID"]
  nID <- length(IDs)

  params <- CountEXTParams(EXT)
  nTheta <- params$nTheta
  nEta <- params$nEta
  nEps <- params$nEps
  nPara <- dim(GRD)[2] - 1

  IterCnt <- max(EXT[, "ITERATION"])
  OFV <- EXT[EXT[, "ITERATION"] == IterCnt, "OBJ"]
  Grad <- GRD[GRD[, "ITERATION"] == IterCnt, ]

  TagList <- c(
    " NO. OF DATA RECS IN DATA SET:",
    " NO. OF DATA ITEMS IN DATA SET:",
    " TOT. NO. OF OBS RECS:",
    " TOT. NO. OF INDIVIDUALS:",
    "0LENGTH OF THETA:",
    " NO. OF FUNCTION EVALUATIONS USED:",
    " NO. OF SIG. DIGITS IN FINAL EST.:"
  )

  TagIndex <- rep(NA, length(TagList))
  for (i in 1:length(TagList)) {
    TagIndex[i] <- ParseOut(TagList[i], XML)
  }

  nRec <- as.integer(TagIndex[1])
  nItem <- as.integer(TagIndex[2])
  nObs <- as.integer(TagIndex[3])
  nID <- as.integer(TagIndex[4])
  nTheta <- as.integer(TagIndex[5])
  nFuncEval <- as.integer(TagIndex[6])
  nSigDigit <- as.double(TagIndex[7])
  nDV <- nObs

  FDATA <- ReadFDATA()

  merged <- MergeIDStatOFV(FDATA, PHI)
  IDStat3 <- merged$IDStat3
  IDStat3$FRecD <- ifelse(IDStat3$FRec == IDStat3$FDRec, "T", "F")
  IDStat4 <- IDStat3[, -(11:14)]
  IDStat4$OFVpDV <- IDStat4[, "iOFV"] / IDStat4[, "nDV"]
  IDStat5 <- IDStat4[order(IDStat4[, "OFVpDV"], decreasing = TRUE), ]
  IDStat6 <- capture.output(IDStat5)

  AICc <- OFV + 2 * nPara + 2 * nPara * (nPara + 1) / (nDV - nPara - 1)
  SBIC <- OFV + nPara * log(nDV)

  sIOFVpDV <- summary(IDStat5[, "OFVpDV"])
  sdIOFVpDV <- sd(IDStat5[, "OFVpDV"])

  # --- PDF Generation ---
  PrepPDF("S1-OFV.PDF")

  split.screen(OFV_SCREEN_LAYOUT)
  options(width = 90)

  AddPage()

  screen(1)
  PrinTxt(1, 1, "Summary 1 - Objective Function Values", Cex = 1.2)
  PrinTxt(3, 1, paste("PROBLEM :", XML[grep(" PROBLEM NO.:", XML) + 1]), Cex = 0.9)
  PrinTxt(5, 3, paste("Number of Total Records :", nRec))
  PrinTxt(6, 3, paste("Number of DV Records    :", nObs))
  PrinTxt(7, 3, paste("Number of Items(Columns):", nItem))
  PrinTxt(8, 3, paste("Number of Parameters    :", nPara))
  PrinTxt(9, 3, paste("Objective Function Value:", OFV))
  PrinTxt(10, 3, paste("OFV per DV              :", format(OFV / nDV, digits = 5)))
  PrinTxt(11, 3, paste("Corrected AIC Value     :", format(AICc, digits = 10)))
  PrinTxt(12, 3, paste("Schwartz Criterion(BIC) :", format(SBIC, digits = 10)))
  PrinTxt(13, 3, paste("# of Gradients Over |1| :", sum(abs(Grad[-1]) > 1)))
  PrinTxt(14, 3, paste("Number of Sig Digits    :", nSigDigit))

  PrinTxt(16, 1, "Summary of Individual OFV per DV", Cex = 0.9)
  PrinTxt(18, 3, paste("Minimum :", sIOFVpDV[1]))
  PrinTxt(19, 3, paste("1st Qu. :", sIOFVpDV[2]))
  PrinTxt(20, 3, paste("Median  :", sIOFVpDV[3]))
  PrinTxt(21, 3, paste("Mean    :", sIOFVpDV[4]))
  PrinTxt(22, 3, paste("3rd Qu. :", sIOFVpDV[5]))
  PrinTxt(23, 3, paste("Maximum :", sIOFVpDV[6]))
  PrinTxt(24, 3, paste("Std Dev :", format(sdIOFVpDV, digits = 4)))
  PrinTxt(25, 3, paste("Coe Var :", format(sdIOFVpDV / sIOFVpDV[4], digits = 4)))
  PrinTxt(26, 3, paste("S-W test:", format(shapiro.test(IDStat5[, "OFVpDV"])$p.value, digits = 4)))

  PrinTxt(28, 1, "Header information for the next table", Cex = 0.9)
  PrinTxt(30, 3, "ID     : Subject ID")
  PrinTxt(31, 3, "iOFV   : Individual Objective Function Value")
  PrinTxt(32, 3, "nRec   : Number of Records")
  PrinTxt(33, 3, "nDV    : Number of DV(Dependent Variable) Records")
  PrinTxt(34, 3, "nMDV   : Number of Missing DV Records")
  PrinTxt(35, 3, "nAMT   : Number of Dosing(AMT) Recods")
  PrinTxt(36, 3, "nEVIDx : Number of Records with EVID == x")
  PrinTxt(37, 3, "FRecD  : Is the first record is a dosing record?")
  PrinTxt(38, 3, "OFVpDV : Objective Function Value per DV")
  PrinTxt(39, 3, "*Table is ordered by decreasing OFVpDV", Cex = 0.7)

  PrinTxt(41, 1, "Abbreviations for Tables", Cex = 0.9)
  PrinTxt(43, 3, "PRED   : Typical Prediction")
  PrinTxt(44, 3, "IPRE   : Individual Prediction")
  PrinTxt(45, 3, "WRES   : Weighted Residual")
  PrinTxt(46, 3, "CWRE   : Conditional Weighted Residual")
  PrinTxt(47, 3, "IWRE   : Individual Weighted Residual")
  PrinTxt(48, 3, "LL     : Lower Limit of Confidence Interval")
  PrinTxt(49, 3, "UL     : Upper Limit of Confidence Interval")
  PrinTxt(50, 3, "RSE    : Relative Standard Error (SE / Point Estimate)")
  PrinTxt(51, 3, "SHR    : Shrinkage (Observed SD / Estimated SD)")
  PrinTxt(52, 3, "ZERO   : Does the confidence interval include 0?")
  PrinTxt(53, 3, "ONE    : Does the confidence interval include 1?")

  par(defpar)

  screen(2)
  x <- IDStat5[, "OFVpDV"]
  y <- jitter(rep(0, length(x)))
  plot(x, y, ylim = c(-0.2, 0.2), xaxt = "n", yaxt = "n", xlab = "", ylab = "", bty = "n", cex = 0.4)
  lines(c(sIOFVpDV[2], sIOFVpDV[2]), c(-0.05, +0.05))
  lines(c(sIOFVpDV[3], sIOFVpDV[3]), c(-0.05, +0.05))
  lines(c(sIOFVpDV[4], sIOFVpDV[4]), c(-0.07, +0.07))
  lines(c(sIOFVpDV[5], sIOFVpDV[5]), c(-0.05, +0.05))
  lines(c(sIOFVpDV[2], sIOFVpDV[5]), c(-0.05, -0.05))
  lines(c(sIOFVpDV[2], sIOFVpDV[5]), c(+0.05, +0.05))
  lines(c(sIOFVpDV[4] - 2 * sdIOFVpDV, sIOFVpDV[4] - 2 * sdIOFVpDV), c(-0.03, +0.03))
  lines(c(sIOFVpDV[4] + 2 * sdIOFVpDV, sIOFVpDV[4] + 2 * sdIOFVpDV), c(-0.03, +0.03))

  par(defpar)

  screen(3)
  var.data <- IDStat5[, "OFVpDV"]
  h.res <- hist(var.data, plot = FALSE)
  d.res <- density(var.data, na.rm = TRUE)
  h.rat <- max(h.res$counts) / max(h.res$density)
  xrange <- range(d.res$x)
  yrange <- c(0, max(h.res$counts, max(d.res$y) * h.rat))
  plot(h.res, xlim = xrange, ylim = yrange, xlab = "IOFV per DV", main = "")
  lines(d.res$x, h.rat * d.res$y)

  screen(4)
  qqnorm(IDStat5[, "OFVpDV"], datax = TRUE, main = "", ylab = "IOFV per DV", cex = 0.4)

  close.screen(all.screens = TRUE)

  nRow <- 55
  AddPage(Cex = 0.8)
  PrinTxt(1, 1, "Table of Individual Objective Function Values and Records Summary", Cex = 1)
  PrinTxt(2, 1, IDStat6[1])

  for (i in 2:length(IDStat6)) {
    if (i %% (nRow - 1) == 1) {
      AddPage(Cex = 0.8)
      PrinTxt(1, 1, IDStat6[1])
    }
    PrinTxt((i - 1) %% (nRow - 1) + 2, 1, IDStat6[i])
  }

  ClosePDF()
  message("S1-OFV.PDF generated.")
}

#' Histogram Panel for Pairs Plot
#'
#' @param x numeric vector
#' @param ... additional arguments
#' @keywords internal
PanelHist <- function(x, ...) {
  usr <- par("usr")
  on.exit(par(usr))

  par(usr = c(usr[1:2], 0, 1.5))
  h <- hist(x, plot = FALSE)
  breaks <- h$breaks
  nB <- length(breaks)
  y <- h$counts
  y <- y / max(y)
  rect(breaks[-nB], 0, breaks[-1], y, col = "cyan", ...)
}


#' Correlation Panel for Pairs Plot
#'
#' @param x numeric vector
#' @param y numeric vector
#' @param digits integer, number of digits for correlation
#' @param prefix character, prefix for text
#' @param cex.cor numeric, character expansion for correlation
#' @keywords internal
PanelCor <- function(x, y, digits = 2, prefix = "", cex.cor) {
  usr <- par("usr")
  on.exit(par(usr))
  par(usr = c(0, 1, 0, 1))
  r <- cor(x, y)
  txt <- format(c(r, 0.123456789), digits = digits)[1]
  txt <- paste(prefix, txt, sep = "")
  if (missing(cex.cor)) cex <- 0.8 / strwidth(txt)
  text(0.5, 0.5, txt, cex = min(1, cex * abs(r) + 0.5))
}


#' Character Panel for Pairs Plot
#'
#' @param x numeric vector
#' @param y numeric vector
#' @keywords internal
PanelChar <- function(x, y) {
  IDs <- get("ParDat", envir = nmw_env)[, "ID"]
  lines(lowess(x, y), col = "red")
  points(x, y, cex = 0.5)
}


#' EBE Pair Plot
#'
#' Creates pairs plot of ETAs vs covariates with histograms and correlations.
#'
#' @param tabEta data.frame with ID, covariates, and ETA columns
#' @param RunNumber character, model/run identifier
#' @param each logical, if TRUE creates separate plots per covariate
#' @export
EBEpair <- function(tabEta, RunNumber, each = FALSE) {
  assign("ParDat", tabEta, envir = nmw_env)
  NameList <- names(tabEta)
  nVar <- length(NameList)
  nEta <- sum(substr(names(tabEta), 1, 3) == "ETA")
  EtaNames <- vector()
  for (i in 1:nEta) EtaNames <- append(EtaNames, paste("ETA", i, sep = ""))

  if (each == FALSE) {
    pairs(tabEta[, -1], main = paste("Covariate vs ETA of", RunNumber), gap = 0,
          lower.panel = PanelChar,
          diag.panel = PanelHist,
          upper.panel = PanelCor)
  } else {
    for (i in 1:nVar) {
      if (substr(NameList[i], 1, 3) != "ETA" & NameList[i] != "ID") {
        pairs(tabEta[, c(NameList[i], EtaNames)],
              main = paste(NameList[i], "vs ETAs of Model", RunNumber), gap = 0,
              lower.panel = PanelChar,
              diag.panel = PanelHist,
              upper.panel = PanelCor)
      }
    }
  }
}

# Package-level environment for pair plot data sharing
nmw_env <- new.env(parent = emptyenv())

#' Run Test Probability
#'
#' Calculates P(Run count <= r | m, n) for the runs test.
#'
#' @param m count of fewer species (minimum value = 0)
#' @param n count of more frequent species (minimum value = 1)
#' @param r count of runs (minimum value = 1)
#' @return numeric, probability
#' @keywords internal
run.p <- function(m, n, r) {
  if (m == 0 & r == 1) {
    return(1)
  } else if (m > n | m < 1 | n < 1 | r < 2 | (r > min(m + n, 2 * m + 1))) {
    return(0)
  }

  sumfu <- 0
  for (u in 2:r) {
    if (u %% 2 == 0) {
      k <- u / 2
      fu <- 2 * choose(m - 1, k - 1) * choose(n - 1, k - 1)
    } else {
      k <- (u + 1) / 2
      fu <- choose(m - 1, k - 1) * choose(n - 1, k - 2) + choose(m - 1, k - 2) * choose(n - 1, k - 1)
    }
    sumfu <- sumfu + fu
  }
  return(sumfu / choose(m + n, m))
}


#' Runs Test for NONMEM Residuals
#'
#' Performs a runs test on residuals. Zeros are omitted.
#'
#' @param RES numeric vector of residuals
#' @return numeric, two-sided p-value
#' @export
run.test.nm <- function(RES) {
  RES <- RES[RES != 0]
  t <- length(RES)
  r <- RES > 0
  m <- sum(r)
  if (t > 1) {
    j <- 2:t
    run <- sum(abs(r[j] - r[j - 1])) + 1
  } else {
    run <- 1
  }
  m <- min(m, t - m)
  n <- max(m, t - m)
  if (run > 1) {
    p <- run.p(m, n, run)
    if (p > 0.5) {
      p <- 1 - run.p(m, n, run - 1)
    }
  } else {
    p <- 0.5^(t - 1)
  }
  return(p)
}


#' Residual Count Test Using Binomial Distribution
#'
#' Tests whether residual counts at various Z-value thresholds follow
#' expected binomial distributions.
#'
#' @param ResTab data.frame with residual columns
#' @param TestCols character vector of column names to test
#' @param ZVals numeric vector of Z-value thresholds
#' @param PctVals numeric vector of percentile values
#' @return matrix of test results
#' @export
ResTest <- function(ResTab, TestCols = c("WRES", "CWRE", "IWRE"),
                    ZVals = c(0, 1, 2, 3),
                    PctVals = c(0.005, 0.01, 0.025, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5)) {
  nPct <- length(PctVals)
  for (i in 1:nPct) {
    ZVals <- union(ZVals, abs(qnorm(PctVals[i])))
  }
  ZVals <- sort(ZVals)
  nZVals <- length(ZVals)

  nRec <- dim(ResTab)[1]
  nTestCols <- length(TestCols)
  CountCols <- c("Z-value", "Percent", "Expected", "LB", "UB")
  for (i in 1:nTestCols) {
    CountCols <- c(CountCols, paste(TestCols[i], "Cnt"), paste(TestCols[i], "p-val"))
  }

  Count <- matrix(nrow = 2 * nZVals, ncol = length(CountCols))
  colnames(Count) <- CountCols

  for (i in 1:nZVals) {
    Count[i, "Z-value"] <- round(-1 * ZVals[nZVals + 1 - i], digits = 3)
    Count[i + nZVals, "Z-value"] <- round(ZVals[i], digits = 3)
    NegProb <- pnorm(-1 * ZVals[nZVals + 1 - i])
    PosProb <- 1 - pnorm(ZVals[i])
    Count[i, "Percent"] <- round(NegProb * 100, digits = 2)
    Count[i + nZVals, "Percent"] <- round(PosProb * 100, digits = 2)
    Count[i, "Expected"] <- round(NegProb * nRec, digits = 1)
    Count[i + nZVals, "Expected"] <- round(PosProb * nRec, digits = 1)
    Count[i, "LB"] <- qbinom(0.025, nRec, NegProb)
    Count[i, "UB"] <- qbinom(0.975, nRec, NegProb) - 1
    Count[i + nZVals, "LB"] <- qbinom(0.025, nRec, PosProb)
    Count[i + nZVals, "UB"] <- qbinom(0.975, nRec, PosProb) - 1

    for (j in 1:length(TestCols)) {
      NegCount <- length(ResTab[ResTab[, paste(TestCols[j])] < (-1 * ZVals[nZVals + 1 - i]),
                                paste(TestCols[j])])
      PosCount <- length(ResTab[ResTab[, paste(TestCols[j])] > ZVals[i],
                                paste(TestCols[j])])

      NegPVal <- pbinom(NegCount, nRec, NegProb)
      PosPVal <- pbinom(PosCount, nRec, PosProb)

      Count[i, paste(TestCols[j], "Cnt")] <- NegCount
      Count[i + nZVals, paste(TestCols[j], "Cnt")] <- PosCount

      Count[i, paste(TestCols[j], "p-val")] <- round(min(NegPVal, 1 - NegPVal), digits = 3)
      Count[i + nZVals, paste(TestCols[j], "p-val")] <- round(min(PosPVal, 1 - PosPVal), digits = 3)
    }
  }
  return(Count)
}


#' Multiple Linear Regression with Influence Diagnostics
#'
#' Performs multiple linear regression and computes influence diagnostics
#' including DFFITS, DFBETAS, Cook's Distance, COVRATIO, and R-Student.
#'
#' @param y numeric vector of response
#' @param x.raw data.frame of predictors
#' @param standardize integer, standardization method (0=none, 1=center, 2=scale by mean, 3=z-score)
#' @return list with model estimates and influence diagnostics
#' @export
mlr2 <- function(y, x.raw, standardize = 0) {
  x.raw <- RemoveNA(x.raw)

  if (kappa(x.raw) > 999 & standardize == 0) {
    message(paste("Condition Number is ", kappa(x.raw), ". Consider standardization !", sep = ""))
  }
  if (length(y) != length(x.raw[, 1])) {
    message("Numbers of rows of x matrix and y vector are different.")
    return(NULL)
  }

  n <- length(y)
  namelist <- c("Intercept", names(x.raw))

  x.avg <- matrix(rep(colMeans(x.raw, na.rm = TRUE), each = n), nrow = n)
  if (standardize == 3) {
    x.sd <- matrix(rep(apply(x.raw, 2, sd, na.rm = TRUE), each = n), nrow = n)
    x <- (x.raw - x.avg) / x.sd
  } else if (standardize == 2) {
    x <- x.raw / x.avg
  } else if (standardize == 1) {
    x <- x.raw - x.avg
  } else {
    x <- x.raw
  }

  Intercept <- 1
  x <- as.matrix(cbind(Intercept, x))

  b <- MASS::ginv(t(x) %*% x) %*% t(x) %*% y
  p <- length(b)

  y.hat <- x %*% b
  e <- y - y.hat
  SSE <- sum(e^2)
  MSE <- SSE / (n - p)
  b.se <- sqrt(diag(as.numeric(MSE) * MASS::ginv(t(x) %*% x)))
  b.t <- b / b.se
  b.p <- pt(b.t, n - p)
  for (i in 1:p) {
    if (!is.na(b.p[i]) & b.p[i] > 0.5) b.p[i] <- 1 - b.p[i]
  }
  b.p <- 2 * b.p

  res1 <- data.frame(namelist, cbind(b, b.se, b.t, b.p))
  names(res1) <- c("Variable", "Estimate", "SE", "T", "p-value")

  h <- as.matrix(diag(x %*% MASS::ginv(t(x) %*% x) %*% t(x)))
  sr <- e / sqrt(MSE * (1 - h))
  MSEi <- ((n - p) * MSE - e^2 / (1 - h)) / (n - p - 1)
  sdr <- e / sqrt(MSEi * (1 - h))

  DFFITS <- sqrt(h / (1 - h)) * e / sqrt(MSEi * (1 - h))

  bi <- matrix(nrow = n, ncol = p)
  for (i in 1:n) {
    z <- x[-i, ]
    bi[i, ] <- MASS::ginv(t(z) %*% z) %*% t(z) %*% y[-i]
  }
  bm <- matrix(rep(t(b), n), byrow = TRUE, ncol = p)
  DFBETAS <- (bm - bi) / sqrt(MSEi %*% diag(MASS::ginv(t(x) %*% x)))

  COVRATIO <- matrix(nrow = n)
  for (i in 1:n) {
    COVRATIO[i] <- det(MSEi[i] * MASS::ginv(t(x[-i, ]) %*% x[-i, ])) / det(MSE * MASS::ginv(t(x) %*% x))
  }

  D <- e^2 / (1 - h)^2 * h / (p * MSE)

  res2 <- data.frame(cbind(y.hat, e, sdr, h, D, COVRATIO, DFFITS, DFBETAS))
  names(res2) <- c("Yhat", "Residual", "R-Student", "hat", "Cook's D", "COV-Ratio", "DFFITS", namelist)

  result <- list(res1, res2, n, p, n - p, SSE, MSE)
  if (standardize == 1 | standardize == 2 | standardize == 3) {
    names(result) <- c("Model Estimates with Standardization", "Influence Diagnostics with DFBETAs",
                       "n", "Parameter Count", "Degree of Freedom", "SSE", "MSE")
  } else {
    names(result) <- c("Model Estimates", "Influence Diagnostics with DFBETAs",
                       "n", "Parameter Count", "Degree of Freedom", "SSE", "MSE")
  }

  return(result)
}

test_that("ResTest returns correct structure", {
  set.seed(42)
  ResTab <- data.frame(
    WRES = rnorm(100),
    CWRE = rnorm(100),
    IWRE = rnorm(100)
  )
  result <- ResTest(ResTab)
  expect_true(is.matrix(result))
  expect_true("Z-value" %in% colnames(result))
  expect_true("Expected" %in% colnames(result))
  expect_true("WRES Cnt" %in% colnames(result))
  expect_true("CWRE Cnt" %in% colnames(result))
  expect_true("IWRE Cnt" %in% colnames(result))
})

test_that("ResTest handles custom columns", {
  set.seed(42)
  ResTab <- data.frame(
    WRES = rnorm(50),
    CWRE = rnorm(50),
    IWRE = rnorm(50)
  )
  result <- ResTest(ResTab, TestCols = c("WRES"))
  expect_true("WRES Cnt" %in% colnames(result))
  expect_false("CWRE Cnt" %in% colnames(result))
})

test_that("mlr2 returns correct structure", {
  set.seed(42)
  y <- rnorm(20)
  x <- data.frame(x1 = rnorm(20), x2 = rnorm(20))
  result <- mlr2(y, x)
  expect_true(is.list(result))
  expect_equal(length(result), 7)
  expect_equal(result[[3]], 20)  # n
  expect_equal(result[[4]], 3)   # p (intercept + 2 predictors)
})

test_that("mlr2 with standardization works", {
  set.seed(42)
  y <- rnorm(20)
  x <- data.frame(x1 = rnorm(20), x2 = rnorm(20))
  result <- mlr2(y, x, standardize = 3)
  expect_true(is.list(result))
  expect_true(grepl("Standardization", names(result)[1]))
})

test_that("mlr2 handles mismatched lengths", {
  y <- rnorm(10)
  x <- data.frame(x1 = rnorm(5), x2 = rnorm(5))
  expect_message(result <- mlr2(y, x), "different")
  expect_null(result)
})

# Internal helper functions shared across report_*.R files
# These functions reduce code duplication in report generation.


#' Count THETA, ETA, EPS Parameters from EXT Data Frame
#'
#' @param EXT data.frame from .ext file (already read)
#' @return list with nTheta, nEta, nEps
#' @keywords internal
CountEXTParams <- function(EXT) {
  cols <- colnames(EXT)
  nTheta <- sum(startsWith(cols, "THETA"))
  nOmega <- sum(startsWith(cols, "OMEGA"))
  nSigma <- sum(startsWith(cols, "SIGMA"))
  nEta <- as.integer((sqrt(1 + 8 * nOmega) - 1) / 2)
  nEps <- as.integer((sqrt(1 + 8 * nSigma) - 1) / 2)
  list(nTheta = nTheta, nEta = nEta, nEps = nEps)
}


#' Merge IDStat with Individual OFV from PHI
#'
#' Common pattern used in report_ofv.R and report_resid.R.
#'
#' @param FDATA data.frame of NONMEM data
#' @param PHI data.frame from .phi file
#' @return list with IDStat3 (merged), VarStat, IDStat
#' @keywords internal
MergeIDStatOFV <- function(FDATA, PHI) {
  VarStat <- NMVarStat(FDATA)
  IDStat <- NMIDStat(FDATA)
  tOFV <- PHI[, c("ID", "OBJ")]
  colnames(tOFV) <- c("ID", "iOFV")
  IDStat2 <- cbind(rownames(IDStat[[1]]), IDStat[[1]])
  IDStat2Name <- names(IDStat2)
  IDStat2Name[1] <- "ID"
  names(IDStat2) <- IDStat2Name
  IDStat3 <- merge(tOFV, IDStat2)
  list(IDStat3 = IDStat3, VarStat = VarStat, IDStat = IDStat)
}


#' Calculate TaLD (Time after Latest Dose) for Report Generation
#'
#' Shared logic between report_pred.R and report_resid.R.
#'
#' @param FDATA data.frame of NONMEM data
#' @param nRec integer, number of records
#' @return list with ToLD, TaLD, DOCC, NewID vectors
#' @keywords internal
CalcTaLDForReport <- function(FDATA, nRec) {
  LastID <- 0
  ToLD <- vector(length = nRec)
  TaLD <- vector(length = nRec)
  DOCCvec <- vector(length = nRec)
  NewID <- vector(length = nRec)

  pToLD <- 0
  DOCC <- -1
  DosingHist2 <- NULL

  for (i in 1:nRec) {
    CurID <- FDATA[i, "ID"]
    if (LastID != CurID) {
      pToLD <- 0
      DOCC <- -1
      if (length(intersect(names(FDATA), "ADDL")) == 1) {
        DosingHist <- FDATA[FDATA[, "ID"] == CurID & FDATA[, "AMT"] > 0 & FDATA[, "MDV"] == 1,
                            c("AMT", "TIME", "II", "ADDL")]
        nDoseRec <- length(DosingHist[, "AMT"])
        for (j in 1:nDoseRec) {
          cADDL <- DosingHist[j, "ADDL"]
          if (cADDL > 0) {
            cAMT <- DosingHist[j, "AMT"]
            cTIME <- DosingHist[j, "TIME"]
            cII <- DosingHist[j, "II"]
            for (k in 1:cADDL) {
              DosingHist <- rbind(DosingHist, c(cAMT, cTIME + k * cII, NA, NA))
            }
          }
        }
        DosingHist2 <- DosingHist[order(DosingHist[, "TIME"]), c("AMT", "TIME")]
      }
    }

    if (FDATA[i, "AMT"] > 0 & FDATA[i, "MDV"] == 1) {
      cToLD <- FDATA[i, "TIME"]
      DOCC <- DOCC + 1
    } else {
      if (length(intersect(names(FDATA), "ADDL")) == 1 && !is.null(DosingHist2)) {
        CurTime <- FDATA[i, "TIME"]
        matching <- DosingHist2[DosingHist2[, "TIME"] < CurTime, "TIME"]
        cToLD <- if (length(matching) > 0) max(matching) else pToLD
      } else {
        cToLD <- pToLD
      }
    }
    ToLD[i] <- cToLD
    TaLD[i] <- FDATA[i, "TIME"] - cToLD
    NewID[i] <- FDATA[i, "ID"] + max(0, DOCC) / 100
    DOCCvec[i] <- DOCC
    LastID <- CurID
    pToLD <- cToLD
  }

  list(ToLD = ToLD, TaLD = TaLD, DOCC = DOCCvec, NewID = NewID)
}


#' Plot Distribution Panel (Summary + Histogram + QQ)
#'
#' Common 3-panel distribution plot used in report_ebe.R,
#' report_indipk.R, and report_resid.R.
#'
#' @param var.data numeric vector of data
#' @param var.name character, variable name for labeling
#' @param show_shrinkage logical, whether to show shrinkage
#' @param omega_var numeric, omega variance for shrinkage calc
#' @param show_ttest logical, whether to show t-test p-value
#' @param show_cv logical, whether to show CV percent
#' @keywords internal
PlotDistribution <- function(var.data, var.name, show_shrinkage = FALSE,
                             omega_var = NULL, show_ttest = FALSE,
                             show_cv = FALSE) {
  sdvar <- sd(var.data)

  # Panel 1: Summary statistics
  plot(0, 0, type = "n", ylim = c(1, 8), xlim = c(0, 10), xaxt = "n", yaxt = "n",
       ylab = "", xlab = "", bty = "n")
  sData <- summary(var.data)
  text(0, 8, paste("Minimum :", sData[1]), adj = 0)
  text(0, 7, paste("1st Qu. :", sData[2]), adj = 0)
  text(0, 6, paste("Median  :", sData[3]), adj = 0)
  text(0, 5, paste("Mean    :", sData[4]), adj = 0)
  text(0, 4, paste("3rd Qu. :", sData[5]), adj = 0)
  text(0, 3, paste("Maximum :", sData[6]), adj = 0)
  text(0, 2, paste("Std Dev :", format(sdvar, digits = 4)), adj = 0)
  if (show_ttest) {
    ttest.pvalue <- format(t.test(var.data)$p.value, digits = 4)
    text(0, 1, paste("t-test p=", ttest.pvalue), adj = 0)
  }
  if (show_cv) {
    text(0, 1, paste("CV (%)  :", format(sdvar / sData[4] * 100, digits = 4)), adj = 0)
  }
  mtext(side = 3, line = 1, var.name, adj = 0)

  # Panel 2: Histogram with density overlay
  h.res <- hist(var.data, plot = FALSE)
  d.res <- density(var.data, na.rm = TRUE)
  h.rat <- max(h.res$counts) / max(h.res$density)
  xrange <- range(d.res$x)
  yrange <- c(0, max(h.res$counts, max(d.res$y) * h.rat))
  plot(h.res, xlim = xrange, ylim = yrange, xlab = var.name, main = "")
  lines(d.res$x, h.rat * d.res$y)

  if (sdvar > 0 && length(var.data) < 5000) {
    p.value <- format(shapiro.test(var.data)$p.value, digits = 4)
    mtext(paste("Shapiro-Wilk test p-value =", p.value), cex = 0.7, line = 0, adj = 0)
  }

  if (show_shrinkage && !is.null(omega_var) && omega_var > 0) {
    shrink <- sdvar / sqrt(omega_var)
    mtext(paste("Shrinkage =", format(shrink, digits = 4)), cex = 0.7, line = 1, adj = 0)
  }

  # Panel 3: QQ plot
  qqnorm(var.data, main = "", ylab = var.name, cex = 0.4)
}


#' OFV Report Split Screen Layout
#'
#' Layout matrix for the S1-OFV report's split screen display.
#'
#' @keywords internal
OFV_SCREEN_LAYOUT <- matrix(
  c(0, 1, 0, 1,
    0.55, 0.97, 0.6, 0.95,
    0.55, 0.97, 0.45, 0.8,
    0.55, 0.97, 0.1, 0.45),
  nrow = 4, ncol = 4, byrow = TRUE
)


#' Read EXT File from Current Directory
#'
#' @return data.frame or NULL if file not found
#' @keywords internal
ReadEXTFile <- function() {
  EXTName <- paste0(GetCurModelName(), ".ext")
  if (!file.exists(EXTName)) return(NULL)
  read.table(EXTName, skip = 1, header = TRUE)
}

test_that("CountEXTParams counts parameters correctly", {
  EXT <- data.frame(
    ITERATION = 1,
    THETA1 = 0.1, THETA2 = 0.2, THETA3 = 0.3,
    OMEGA.1.1. = 0.04, OMEGA.2.1. = 0.01, OMEGA.2.2. = 0.09,
    SIGMA.1.1. = 0.01,
    OBJ = 100
  )
  result <- nmw:::CountEXTParams(EXT)
  expect_equal(result$nTheta, 3)
  expect_equal(result$nEta, 2)   # 3 omega elements -> 2x2 matrix
  expect_equal(result$nEps, 1)   # 1 sigma element -> 1x1 matrix
})

test_that("CountEXTParams handles single parameter", {
  EXT <- data.frame(
    ITERATION = 1,
    THETA1 = 0.5,
    OMEGA.1.1. = 0.1,
    SIGMA.1.1. = 0.01,
    OBJ = 50
  )
  result <- nmw:::CountEXTParams(EXT)
  expect_equal(result$nTheta, 1)
  expect_equal(result$nEta, 1)
  expect_equal(result$nEps, 1)
})

test_that("CalcTaLDForReport calculates time after latest dose", {
  FDATA <- data.frame(
    ID = c(1, 1, 1, 2, 2),
    TIME = c(0, 1, 4, 0, 2),
    AMT = c(100, 0, 0, 200, 0),
    MDV = c(1, 0, 0, 1, 0)
  )
  result <- nmw:::CalcTaLDForReport(FDATA, nrow(FDATA))
  expect_equal(length(result$TaLD), 5)
  expect_equal(result$TaLD[1], 0)  # dosing record
  expect_equal(result$TaLD[2], 1)  # 1h after dose at TIME=0
  expect_equal(result$TaLD[3], 4)  # 4h after dose at TIME=0
  expect_equal(result$TaLD[4], 0)  # dosing record for ID=2
  expect_equal(result$TaLD[5], 2)  # 2h after dose at TIME=0
})

test_that("CalcTaLDForReport handles multiple dosing occasions", {
  FDATA <- data.frame(
    ID = c(1, 1, 1, 1),
    TIME = c(0, 1, 24, 25),
    AMT = c(100, 0, 100, 0),
    MDV = c(1, 0, 1, 0)
  )
  result <- nmw:::CalcTaLDForReport(FDATA, nrow(FDATA))
  expect_equal(result$TaLD[2], 1)   # 1h after first dose
  expect_equal(result$TaLD[4], 1)   # 1h after second dose
})

test_that("OFV_SCREEN_LAYOUT has correct dimensions", {
  layout <- nmw:::OFV_SCREEN_LAYOUT
  expect_equal(dim(layout), c(4, 4))
  expect_equal(layout[1, ], c(0, 1, 0, 1))  # full screen for first panel
})

#' Generate Input Data Summary Report (SA-Input.PDF)
#'
#' Generates a PDF report with input data validation including
#' individuals with no dosing/DV, duplicate records, and summary statistics.
#'
#' @param run_dir character, path to the NONMEM run directory (default: current directory)
#' @export
nmw_report_input <- function(run_dir = getwd()) {
  owd <- setwd(run_dir)
  on.exit(setwd(owd))

  defpar <- par(no.readonly = TRUE)
  defpar$new <- NULL
  on.exit(par(defpar), add = TRUE)

  CtlName <- GetCurModelName()

  XML <- readLines(paste0(CtlName, ".xml"))

  TagList <- c(" NO. OF DATA RECS IN DATA SET:", " NO. OF DATA ITEMS IN DATA SET:")
  TagIndex <- rep(NA, length(TagList))
  for (i in 1:length(TagList)) {
    TagIndex[i] <- ParseOut(TagList[i], XML)
  }
  nRec <- as.integer(TagIndex[1])

  FDATA <- ReadFDATA()
  VarStat <- NMVarStat(FDATA)
  IDStat <- NMIDStat(FDATA)

  PrepPDF("SA-Input.PDF")

  if (length(IDStat[[1]][IDStat[[1]][, "nAMT"] == 0, 1]) > 0) {
    PrinMTxt(capture.output(IDStat[[1]][IDStat[[1]][, "nAMT"] == 0, ]),
             Header1 = "Individuals with no dosing")
  }
  if (length(IDStat[[1]][IDStat[[1]][, "nDV"] == 0, 1]) > 0) {
    PrinMTxt(capture.output(IDStat[[1]][IDStat[[1]][, "nDV"] == 0, ]),
             Header1 = "Individuals with no DV")
  }

  # Duplicate detection
  DuplData <- NULL
  i <- 1
  while (i < nRec) {
    cID <- FDATA[i, "ID"]
    cTIME <- FDATA[i, "TIME"]
    cMDV <- FDATA[i, "MDV"]
    tDATA <- FDATA[FDATA[, "ID"] == cID & FDATA[, "TIME"] == cTIME & FDATA[, "MDV"] == cMDV, ]
    if (length(tDATA[, 1]) > 1) {
      DuplData <- rbind(DuplData, tDATA)
    }
    i <- i + length(tDATA[, 1])
  }

  options(width = 128)
  if (!is.null(DuplData)) {
    PrinMTxt(capture.output(DuplData), Header1 = "Potentially Harmful Records", Cex = 0.6)
  }
  FDATA[!is.na(FDATA$DV) & FDATA$DV == 0, "DV"] <- NA
  options(width = 90)
  PrinMTxt(capture.output(summary(FDATA)))
  PrinMTxt(capture.output(VarStat))
  PrinMTxt(capture.output(IDStat))

  ClosePDF()
  message("SA-Input.PDF generated.")
}

#' Parse Value After Tag from NONMEM Output
#'
#' Extracts the value following a tag string (after the colon) in NONMEM output lines.
#'
#' @param TagString character, the tag to search for
#' @param RawRead character vector of lines from NONMEM output
#' @return character, the value after the tag
#' @export
ParseOut <- function(TagString, RawRead) {
  Inter <- grep(TagString, RawRead, value = TRUE)
  if (length(Inter) > 0) {
    return(strsplit(Inter, ":")[[1]][2])
  } else {
    return(0)
  }
}


#' Parse Data Item Names from NONMEM Output
#'
#' @param RawRead character vector of lines from NONMEM output
#' @return character vector of item names
#' @keywords internal
ParseItemName <- function(RawRead) {
  Tag1 <- "0LABELS FOR DATA ITEMS:"
  Tag2 <- "LABELS FOR PRED-DEFINED ITEMS:"

  Line1 <- grep(Tag1, RawRead)
  Line2 <- grep(Tag2, RawRead)

  Ans <- vector()
  for (i in (Line1 + 1):(Line2 - 1)) {
    Ans <- c(Ans, strsplit(RawRead[i], " ")[[1]][-1])
  }
  return(Ans)
}


#' Parse Fixed-Width Format Item Widths from NONMEM Output
#'
#' @param RawRead character vector of lines from NONMEM output
#' @return list with LineCount and Widths
#' @keywords internal
ParseItemWidth <- function(RawRead) {
  Line1 <- grep("0FORMAT FOR DATA:", RawRead)
  FORM <- trimws(RawRead[Line1 + 1])
  nChar <- nchar(FORM)
  FORM <- substr(FORM, 2, nChar - 1)
  f2 <- strsplit(FORM, ",")[[1]]
  nF <- length(f2)
  newF <- vector()
  for (i in 1:nF) {
    Pos <- grep("(\\()", f2[i])
    if (length(Pos) > 0) {
      tStr <- f2[i]
      nChar <- nchar(tStr)
      for (j in 1:nChar) {
        if (substr(tStr, j, j) == "(") Pos1 <- j
        if (substr(tStr, j, j) == ")") Pos2 <- j
      }
      Mul <- as.numeric(substr(tStr, 1, Pos1 - 1))
      Term <- substr(tStr, Pos1 + 1, Pos2 - 1)
      for (j in 1:Mul) {
        newF <- c(newF, Term)
      }
    } else {
      newF <- c(newF, f2[i])
    }
  }

  LineCount <- 1
  for (i in 1:length(newF)) {
    if (substr(newF[i], nchar(newF[i]), nchar(newF[i])) == "/") {
      LineCount <- LineCount + 1
      newF[i] <- substr(newF[i], 1, nchar(newF[i]) - 1)
    }
  }

  Fmts <- chartr("F", "E", newF)
  n <- length(Fmts)
  Widths <- vector()
  for (i in 1:n) {
    x <- strsplit(Fmts[i], "E")[[1]]
    if (x[1] == "") {
      r <- 1
    } else {
      r <- as.integer(x[1])
    }
    for (j in 1:r) {
      Widths <- c(Widths, as.integer(x[2]))
    }
  }
  Result <- list(LineCount, Widths)
  names(Result) <- c("Lines", "Widths")
  return(Result)
}


#' Extract Value Between XML Tags
#'
#' @param Tag1 opening tag
#' @param Tag2 closing tag
#' @param RawRead character vector of lines
#' @return character, the value between tags
#' @keywords internal
BtwTagVal <- function(Tag1, Tag2, RawRead) {
  return(sub(Tag2, "", sub(Tag1, "", grep(paste(Tag1, "(.*?)", Tag2, sep = ""),
                                          RawRead, value = TRUE))))
}


#' Extract Lines Between XML Tags
#'
#' @param Tag1 opening tag
#' @param Tag2 closing tag
#' @param RawRead character vector of lines
#' @return character vector of lines between tags
#' @keywords internal
BtwTagLines <- function(Tag1, Tag2, RawRead) {
  Line1 <- grep(Tag1, RawRead)
  Line2a <- grep(Tag2, RawRead)

  if (length(Line1) > 0) {
    if (length(Line2a) > 1) {
      i <- 1
      while (Line2a[i] < Line1) i <- i + 1
      Line2 <- Line2a[i]
    } else {
      Line2 <- Line2a
    }
    return(RawRead[(Line1 + 1):(Line2 - 1)])
  } else {
    return(0)
  }
}


#' Extract Vector of Values Between XML Tags
#'
#' Parses NONMEM XML output to extract theta, omega, or sigma vectors.
#'
#' @param Tag character, the tag name (e.g., "nm:theta")
#' @param RawRead character vector of XML lines
#' @return numeric vector of values
#' @export
BtwTagVals <- function(Tag, RawRead) {
  Tag1 <- paste("<", Tag, ">", sep = "")
  Tag2 <- paste("</", Tag, ">", sep = "")
  Line1 <- grep(Tag1, RawRead)
  Line2 <- grep(Tag2, RawRead)
  if (length(Line1) != 0 && length(Line2) != 0) {
    TmpRaw <- RawRead[(Line1 + 1):(Line2 - 1)]
    nVal <- length(TmpRaw)
    Vals <- rep(0, nVal)
    for (i in 1:nVal) {
      Vals[i] <- as.numeric(BtwTagVal(paste("<nm:val nm:name='", i, "'>", sep = ""),
                                      "</nm:val>", TmpRaw[i]))
    }
    return(Vals)
  } else {
    return(0)
  }
}


#' Extract Matrix from NONMEM XML Output
#'
#' Parses NONMEM XML output to extract omega or sigma matrices.
#'
#' @param Tag character, the tag name (e.g., "omega", "sigma")
#' @param RawRead character vector of XML lines
#' @param nRow integer, dimension of the matrix
#' @return numeric matrix
#' @export
BtwTagMat <- function(Tag, RawRead, nRow) {
  Tag1 <- paste("<nm:", Tag, ">", sep = "")
  Tag2 <- paste("</nm:", Tag, ">", sep = "")
  MatLines <- BtwTagLines(Tag1, Tag2, RawRead)
  RetMat <- matrix(rep(1e+10, nRow * nRow), nrow = nRow, ncol = nRow)

  if (length(MatLines) > 0 & MatLines[1] != 0) {
    for (i in 1:nRow) {
      MatRow <- BtwTagLines(paste("<nm:row nm:rname='", i, "'>", sep = ""),
                            "</nm:row>", MatLines)
      for (j in 1:i) {
        RetMat[i, j] <- RetMat[j, i] <- as.double(
          BtwTagVal(paste("<nm:col nm:cname='", j, "'>", sep = ""),
                    "</nm:col>", MatRow))
      }
    }
  }
  return(RetMat)
}

# Objective Functions
ObjEta = function(ETAi)
{
# External Variable : e$INTER, e$DATAi, e$THETA, e$invOM, e$SG, e$nEta, e$HNames
# External Function : e$PRED
  FGHDi = e$PRED(e$THETA, ETAi, e$DATAi)
  Ri    = e$DATAi[,"DV"] - FGHDi[,"F"]
  
  if (e$INTER == TRUE) {
    Hi = FGHDi[, e$HNames, drop=FALSE]
  } else {
    FGHD0 = e$PRED(e$THETA, rep(0, e$nEta), e$DATAi)
    Hi = FGHD0[, e$HNames, drop=FALSE]
  }
  
## Slower version
#  Vi = diag(diag(Hi %*% SG %*% t(Hi)))
#  iSum = log(det(Vi)) + t(Yi - Fi)%*%solve(Vi)%*%(Yi - Fi) + t(ETAi) %*% invOM %*% ETAi

## Faster version
  Vi    = diag(Hi %*% e$SG %*% t(Hi))
  return(sum(log(Vi) + Ri*Ri/Vi) + t(ETAi) %*% e$invOM %*% ETAi)
}


Oi = function(i)  # Combined version. You need not use this.
{
#Oi : Calculate individual Objective Function Value with Current Externals
# External Variables: Oi, THETA, ETAi, OM, invOM, SG, DataRef, GNmaes, HNames, DNames, Term123
# External Functions: PRED, ltv2mat
# METHOD = "ZERO" | "COND" | LAPL"
# INTER = TRUE | FALSE  : Here I do not consider. It assumes always with INTER option

# Oi requires basically
#  PRED, THETA, ETAi, DataRef, GNames, HNames, SG, nReci

# Oi requires additionally with METHOD="ZERO"
#  OM

# Oi requires  additionally with METHOD=="COND"
#  invOM, Term123

# Oi requires  additionally with METHOD=="LAPL"
#  invOM, Term123, DNames

  FGHDi = e$PRED(e$THETA, e$ETAi, e$DataRef[[i]]) # ETAi  = rep(0, nEta) # Fixed externally if METHOD="ZERO"
  Yi    = e$DataRef[[i]][, "DV"]
  Fi    = FGHDi[, "F"]
  Gi    = FGHDi[, e$GNames, drop=FALSE]
  Hi    = FGHDi[, e$HNames, drop=FALSE]
  Vi    = diag(Hi %*% e$SG %*% t(Hi))

  if (e$METHOD == "ZERO") {
    Ri     = Yi - Fi
    Ci     = Gi %*% e$OM %*% t(Gi) + diag(Vi, nrow=e$nReci)
    Result = determinant(Ci, logarithm=TRUE)$modulus[[1]] + t(Ri) %*% solve(Ci) %*% Ri
  } else if (e$METHOD == "COND") {
    Hsum = e$invOM + t(Gi) %*% solve(diag(Vi)) %*% Gi
    Result = e$Term123 + determinant(Hsum, logarithm=TRUE)$modulus[[1]]
  } else if (e$METHOD == "LAPL") {
    Hsum = e$invOM
    for (j in 1:e$nReci) {
      Dij = ltv2mat(FGHDi[j, e$DNames])
      Hsum = Hsum + (Gi[j,] %*% t(Gi[j,]) - (Yi[j] - Fi[j])*Dij)/Vi[j]
    }
    Result = e$Term123 + determinant(Hsum, logarithm=TRUE)$modulus[[1]]
  } else {
    print("Undefined Estimation Method")
    Result = NULL
  }
  return(Result)
}


Oi0 = function(i) # Calculate individual OFV with METHOD=="ZERO"
{
# External Variable: THETA, ETAi, DataRef, Gnames, HNames, OM, SG, nReci
# External Function: PRED
  FGHDi = e$PRED(e$THETA, e$ETAi, e$DataRef[[i]]) # ETAi = rep(0, nEta): fixed externally if METHOD=="ZERO"
  Ri    = e$DataRef[[i]][,"DV"] - FGHDi[,"F"]
  Gi    = FGHDi[, e$GNames, drop=FALSE]
  Hi    = FGHDi[, e$HNames, drop=FALSE]
  Ci    = Gi %*% e$OM %*% t(Gi) + diag(diag(Hi %*% e$SG %*% t(Hi)), nrow=e$nReci)
  return(determinant(Ci, logarithm=TRUE)$modulus[[1]] + t(Ri) %*% solve(Ci) %*% Ri)
}


Oi1 = function(i) # Calculate individual OFV with METHOD=="COND" or "LAPL"
{
# External Variable: THETA, ETAi, DataRef, Gnames, HNames, SG, invOM, Oi, Term123, METHOD, nReci
# External Function: PRED
  FGHDi = e$PRED(e$THETA, e$ETAi, e$DataRef[[i]]) # ETAi : predetermined by ObjEta if METHOD=="COND" or "LAPL"
  Yi    = e$DataRef[[i]][,"DV"]
  Fi    = FGHDi[,"F"]
  Gi    = FGHDi[, e$GNames, drop=FALSE]
  Hi    = FGHDi[, e$HNames, drop=FALSE]
  Vi    = diag(Hi %*% e$SG %*% t(Hi))
  if (e$METHOD == "COND") {
    Hsum = e$invOM + t(Gi) %*% solve(diag(Vi)) %*% Gi
  } else if (e$METHOD == "LAPL") {
    Hsum  = e$invOM
    for (j in 1:e$Oi[i,"nRec"]) {
      Dij = ltv2mat(FGHDi[j, e$DNames])
      Hsum = Hsum + (Gi[j,] %*% t(Gi[j,]) - (Yi[j] - Fi[j])*Dij)/Vi[j]
    }
  }
  return(e$Term123 + determinant(Hsum, logarithm=TRUE)$modulus[[1]])
}


Obj = function(vPara) # Calculate total OFV with vectorized parameter (THETA, OM, SG)
{
# External Variable: HNames for ObjEta, ETAi when METHOD=="ZERO", STEP=="EST" or "COV", Oi, EBE, OMindex, SGindex
# External Function: DECN

# Oi requires basically
#  PRED, THETA, ETAi, DataRef, GNames, HNames

# Oi requires  additionally with METHOD=="COND"
#  invOM, Term123

# Oi requires  additionally with METHOD=="LAPL"
#  invOM, Term123, DNames
  if (e$STEP == "EST") {  # "EST" STEP requires DECN, vPara is UCP
    vPara = DECN(vPara)
  }

  e$THETA = vPara[1:e$nTheta]                                  # PRED requires in ObjEta and Oi
  e$OM    = ltv2mat(vPara[e$OMindex])
  e$SG    = diag(vPara[e$SGindex], nrow=e$nEps) # ObjEta rquires

  if (e$METHOD == "ZERO") {  # If METHOD=="ZERO", ETAi should be fixed as rep(0, nEta)
    e$ETAi = rep(0, e$nEta)
    for (i in 1:e$nID)  {
      e$nReci = e$Oi[i, "nRec"]
      e$Oi[i, "OFVi"] = Oi0(i)  # Oi requires additionally with METHOD="ZERO":  OM, SG
    }
  } else {  # METHOD=="COND" or "LAPL"
#    tryCatch(code = { e$invOM = solve(e$OM)}, error=function(e) { e$invOM = MASS::ginv(e$OM) })
#    tryCatch(code = { e$invOM = solve(e$OM)}, error=function(e) { e$invOM = diag(rep(1, as.integer(e$nEta))) })
    try(e$invOM <- solve(e$OM), silent=TRUE)                                    
    Term2 = determinant(e$OM, logarithm=TRUE)$modulus[[1]]     # If METHOD=="COND" or "LAPL", Oi requires
    for (i in 1:e$nID)  {
      e$DATAi = e$DataRef[[i]]
      e$nReci = -e$Oi[i, "nRec"]
      Res     = optim(e$EBE[i, e$EtaNames], ObjEta, method="BFGS") # ObjEta requires DATAi, THETA, invOM, SG, HNames # Regardless of "COV' or "EST" Step, use ObjEta
      e$ETAi  = as.matrix(Res$par, nrow=e$nEta)
      e$EBE[i, e$EtaNames] = Res$par
      e$Term123 = Term2 + Res$value                            # Oi requires Term123 for METHOD=="COND" or "LAPL"
      e$Oi[i, "OFVi"] = Oi1(i)
    }
  }
  return(sum(e$Oi[,"OFVi"]))
}


OiS0 = function(vPara)
{
# External Variable: DATAi, nTheta, nEta, nEps, nPara, GNames, HNames, ETAi
# External Function: PRED, ltv2mat
  THETA = vPara[1:e$nTheta]
  OM    = ltv2mat(vPara[e$OMindex])
  SG    = diag(vPara[e$SGindex], nrow=e$nEps)
  FGHDi = e$PRED(THETA, e$ETAi, e$DATAi)
  Ri    = e$DATAi[, "DV"] - FGHDi[, "F"]
  Gi    = FGHDi[, e$GNames, drop=FALSE]
  Hi    = FGHDi[, e$HNames, drop=FALSE]
  Ci    = Gi %*% OM %*% t(Gi) + diag(diag(Hi %*% SG %*% t(Hi)))
  return(determinant(Ci, logarithm=TRUE)$modulus[[1]] + t(Ri) %*% solve(Ci) %*% Ri)
}


OiS1 = function(vPara)
{
# External Variable: DATAi, nReci, nTheta, nEta, nEps, nPara, GNames, HNames, DNames
# External Function: PRED, ltv2mat
  THETA = vPara[1:e$nTheta]
  OM    = ltv2mat(vPara[e$OMindex])
  SG    = diag(vPara[e$SGindex], nrow=e$nEps)
  invOM = solve(OM)
  FGHDi = e$PRED(THETA, e$ETAi, e$DATAi)
  Yi    = e$DATAi[,"DV"]
  Fi    = FGHDi[,"F"]
  Gi    = FGHDi[, e$GNames, drop=FALSE]
  Hi    = FGHDi[, e$HNames, drop=FALSE]
  Vi    = diag(Hi %*% SG %*% t(Hi))
  Ri    = Yi - Fi
  Term1 = sum(log(Vi) + Ri*Ri/Vi) # = log(det(Vi)) + t(Yi - Fi)%*%solve(Vi)%*%(Yi - Fi)
  Term2 = determinant(OM, logarithm=TRUE)$modulus[[1]]
  Term3 = t(e$ETAi) %*% invOM %*% e$ETAi
  if (e$METHOD == "COND") {
    Hsum = e$invOM + t(Gi) %*% solve(diag(Vi)) %*% Gi
  } else if (e$METHOD == "LAPL") {
    Hsum  = e$invOM
    for (j in 1:e$nReci) {
      Dij = ltv2mat(FGHDi[j, e$DNames])
      Hsum = Hsum + (Gi[j,] %*% t(Gi[j,]) - (Yi[j] - Fi[j])*Dij)/Vi[j]
    }
  }
  Term4 = determinant(Hsum, logarithm=TRUE)$modulus[[1]]
  return(Term1 + Term2 + Term3 + Term4)
}


CalcSmat = function() # Calculate Smat with METHOD=="ZERO"
{
# External Variables: METHOD, nPara, nTheta, nEta, nEps, nPara, GNames, HNames, DNames
# External Functions: PRED
#  require(numDeriv)
  Smat = matrix(rep(0, e$nPara*e$nPara), nrow=e$nPara, ncol=e$nPara)
  for (i in 1:e$nID) {
    e$DATAi = e$DataRef[[i]]
    e$ETAi  = e$EBE[i, e$EtaNames]
    e$nReci = e$Oi[i,"nRec"]
    if (e$METHOD == "ZERO") {
      gr = grad(OiS0, e$FinalPara)
    } else {
      gr = grad(OiS1, e$FinalPara)
    }
    Smat = Smat + gr %*% t(gr)
  }
  return(Smat/4)
}


PostHocEta = function()
{
  e$THETA = e$FinalPara[1:e$nTheta]
  e$invOM = solve(ltv2mat(e$FinalPara[e$OMindex]))
  e$SG    = diag(e$FinalPara[e$SGindex], nrow=e$nEps)
  for (i in 1:e$nID) {
    e$DATAi = e$DataRef[[i]]
    e$EBE[i, e$EtaNames] = optim(rep(0, e$nEta), ObjEta, method="BFGS")$par
  }
  return(e$EBE)
}

CovStep = function()
{
  e$STEP = "COV"
  StartTime = Sys.time()
  Rmat = hessian(Obj, e$FinalPara)/2
  Smat = CalcSmat()
  if (is.nan(Smat[1,1])) stop("Error - NaN produced")
  invR = solve(Rmat)
  Cov = invR %*% Smat %*% invR
  SE = sqrt(diag(Cov))
  Correl = cov2cor(Cov)
#  InvCov = solve(Cov)
  InvCov = Rmat %*% solve(Smat) %*% Rmat
  EigenVal = sort(eigen(Correl)$values) # Sorted Eigenvalues

  RunTime = difftime(Sys.time(), StartTime)
  Result = list(RunTime, SE, Cov, Correl, InvCov, EigenVal, Rmat, Smat)
  names(Result) = list("Time", "Standard Error", "Covariance Matrix of Estimates", "Correlation Matrix of Estimates", "Inverse Covariance Matrix of Estimates", "Eigen Values", "R Matrix", "S Matrix")
  return(Result)
}

TrimOut = function(inFile, outFile="PRINT.OUT")
{
  if (missing(inFile)) {
    vPath = strsplit(getwd(), "/")[[1]]
    n = length(vPath)
    inFile = paste0(strsplit(vPath[n],"(\\.)")[[1]][1], ".OUT")
  }
  if (!file.exists(inFile)) stop("File does not exist!")

  O1 = readLines(inFile)
  x1 = which(startsWith(O1, "$PROB"))      # fo find local times
  x2 = which(startsWith(O1, "Stop Time:")) # to find local times  
  xLines = c(x1 - 1, x2 + 2)
  nLine = length(O1)
  MaxW = 1023

  x3 = which(startsWith(O1, "1NONLINEAR MIXED EFFECTS MODEL PROGRAM (NONMEM) "))
  if (is.na(x3[1])) {
    Line1 = nLine
  } else {
    Line1 = x3[1]
  }

  Res = O1[1]
  cLine = 2
  for (i in 2:nLine) {
    if (i %in% xLines | i < Line1 ) {
      Res[cLine] = O1[i]
      cLine = cLine + 1
    } else {
      Ch1 = substr(O1[i], 1, 1)
      if (Ch1 == "1" | Ch1 == "0") {
        Res[cLine] = ""
        Res[cLine + 1] = substr(O1[i], 2, MaxW)
        cLine = cLine + 2
      } else if (Ch1 == "+") {
        Res[cLine - 1] = paste0(Res[cLine - 1], substr(O1[i], nchar(O1[i - 1]) + 1, MaxW))
      } else if (Ch1 == " ") {
        Res[cLine] = substr(O1[i], 2, MaxW)
        cLine = cLine + 1
      } else {
        Res[cLine] = O1[i]
        cLine = cLine + 1      
      }
    }
  }

  if (outFile == "") { 
    return(Res)
  } else { 
    writeLines(Res, outFile)
  }
}

CombDmExPc = function(dm, ex, pc)
{
  ColDm = toupper(colnames(dm))
  ColEx = toupper(colnames(ex))
  ColPc = toupper(colnames(pc))
  colnames(dm) = ColDm
  colnames(ex) = ColEx
  colnames(pc) = ColPc
  
  if (nrow(dm) != length(unique(dm$ID))) stop("The first table should have only one row for each ID!")
  
  ColAll = union(union(ColEx, ColPc), ColDm)
  ToAddEx = setdiff(ColAll, ColEx)
  ToAddPc = setdiff(ColAll, ColPc)

  IDs = sort(unique(pc$ID))
  nID = length(IDs)

  if ("DATE" %in% ColEx) {
    ex$DT = strptime(paste(ex$DATE, ex$TIME), "%Y-%m-%d %H:%M")
    pc$DT = strptime(paste(pc$DATE, pc$TIME), "%Y-%m-%d %H:%M")
  } else {
    ex$DT = ex$TIME
    pc$DT = pc$TIME
  }
  ex = ex[order(ex$ID, ex$DT), ]
  pc = pc[order(pc$ID, pc$DT), ]

  FLAG = rep(T, NROW(ex))
  for (i in 1:nID) {
    cID = IDs[i]
    cDAT = pc[pc$ID == cID, , drop=F]
    cLast = cDAT[NROW(cDAT), "DT"]
    FLAG[ex$ID == cID & ex$DT >= cLast] = F
  }
  ex = ex[FLAG, ]

  ex = cbind(ex, MDV = 1)
  pc = cbind(pc, MDV = 0)

  ToEx = matrix(nrow=NROW(ex), ncol=length(ToAddEx))
  colnames(ToEx) = ToAddEx

  ToPc = matrix(nrow=NROW(pc), ncol=length(ToAddPc))
  colnames(ToPc) = ToAddPc

  Res = rbind(cbind(ex, ToEx), cbind(pc, ToPc)) # Ex first for MDV descending order
  Res = Res[order(Res$ID, Res$DT, Res$MDV), ]
  for (i in 1:nID) {
    cID = IDs[i]
    Res[Res$ID == cID, setdiff(ColDm, "ID")] = dm[dm$ID == cID, setdiff(ColDm, "ID")]
  }
  Res = Res[, c(ColAll, "MDV")]
  rownames(Res) = NULL
  return(Res)
}

SqrtInvCov = function(M)
{
  EigenResult = eigen(as.matrix(M))
  EigenVector = EigenResult$vectors
  EigenValues = abs(EigenResult$values)
  return(EigenVector %*% diag(1/sqrt(EigenValues), nrow = dim(M)[1]) %*% t(EigenVector))
}


Mx = function(M, x)
{
  e = eigen(as.matrix(M))
  return(e$vectors %*% diag(abs(e$values)^x) %*% t(e$vectors))
}

# "%^%" = function(S, pow) with(eigen(S), vectors %*% (abs(values)^pow * t(vectors))) 
# S%^%(-0.5)

mat2ltv = function(mat)
{
  return(mat[upper.tri(mat,diag=TRUE)])
}


ltv2mat = function(vec)
{
  LENGTH = length(vec)
  DIM = round((sqrt(8*LENGTH+1)-1)/2,0)
  if (DIM*(DIM+1)/2 != LENGTH) return(NULL)
  mat = matrix(nrow=DIM, ncol=DIM)
  for (m in 1:DIM) {
    for (n in 1:DIM) {
      k = max(m,n)
      l = min(m,n)
      p = k * (k-1) / 2 + l
      mat[m,n] = vec[p]
    }
  }
  return(mat)
}


mat2utv = function(mat)
{
  return(mat[lower.tri(mat,diag=TRUE)])
}


utv2mat = function(vec)
{
  LENGTH = length(vec)
  DIM = as.integer(round((sqrt(8*LENGTH+1)-1)/2,0))
  if (DIM*(DIM+1)/2 != LENGTH) return(NULL)
  mat = matrix(nrow=DIM, ncol=DIM)
  for (m in 1:DIM) {
    for (n in 1:DIM) {
      k = min(m,n)
      l = max(m,n)
      p = (2*DIM - k + 2)*(k - 1)/2 + l - k + 1
      mat[m,n] = vec[p]
    }
  }
  return(mat)
}


ScaleVar = function(VarMat, dim1)
{
  M1 = chol(VarMat)
  V1 = diag(M1)
  M2 = abs(10 * (M1 - diag(V1, nrow=dim1))) + diag(V1/exp(0.1), nrow=dim1)
  return(t(M2))
}


DesclVar = function(mUCP, mSCL)
{
  nRow = dim(mUCP)[1]
  maT = matrix(nrow=nRow, ncol=nRow)

  for (i in 1:nRow) {
    for (j in 1:nRow) {
      if (i==j) {
        maT[i,j] = exp(mUCP[i,j]) * mSCL[i,j]
      } else if(i > j) {
        maT[i,j] = mUCP[i,j] * mSCL[i,j]
      } else {
        maT[i,j] = 0
      }
    }
  }
  return(maT %*% t(maT))
}


DECN = function(UCP)
{
#Extern: nTheta, nEta, nEps
#        ThetaL[nTheta]==LB, ThetaI[nTheta]==IE, ThetaU[nTheta]==UB
#        alpha, OMscl, SGscl
  uTheta = UCP[1:e$nTheta]
  Theta = exp(uTheta - e$alpha)/(exp(uTheta - e$alpha) + 1)*(e$UB - e$LB) + e$LB
  uvOM  = UCP[(e$nTheta + 1):(e$nTheta + e$nEta*(e$nEta + 1)/2)]
  umOM  = utv2mat(uvOM)
  OM    = DesclVar(umOM, e$OMscl)
  ltvOM = mat2ltv(OM)

  udSG  = UCP[(e$nTheta + e$nEta*(e$nEta + 1)/2 + 1):(e$nTheta + e$nEta*(e$nEta + 1)/2 + e$nEps)]
  umSG  = diag(udSG, nrow=e$nEps)
  SG    = DesclVar(umSG, e$SGscl)
  dgSG  = diag(SG)
  return(c(Theta, ltvOM, dgSG))
}

test_that("mat2ltv and ltv2mat are inverse operations", {
  m <- matrix(c(1, 2, 3, 2, 5, 6, 3, 6, 9), nrow = 3)
  v <- mat2ltv(m)
  m2 <- ltv2mat(v)
  expect_equal(m, m2)
})

test_that("mat2utv and utv2mat are inverse operations", {
  m <- matrix(c(1, 2, 3, 2, 5, 6, 3, 6, 9), nrow = 3)
  v <- mat2utv(m)
  m2 <- utv2mat(v)
  expect_equal(m, m2)
})

test_that("ltv2mat returns NULL for invalid vector length", {
  expect_null(ltv2mat(c(1, 2, 3, 4)))
})

test_that("utv2mat returns NULL for invalid vector length", {
  expect_null(utv2mat(c(1, 2, 3, 4)))
})

test_that("mat2ltv extracts correct number of elements", {
  m <- diag(3)
  v <- mat2ltv(m)
  expect_equal(length(v), 6)  # 3*(3+1)/2
})

test_that("SqrtInvCov produces correct result", {
  m <- matrix(c(4, 2, 2, 3), nrow = 2)
  A <- SqrtInvCov(m)
  result <- A %*% A
  expect_equal(result, solve(m), tolerance = 1e-10)
})

test_that("Mx with power 1 returns original matrix", {
  m <- matrix(c(4, 1, 1, 3), nrow = 2)
  result <- Mx(m, 1)
  expect_equal(result, m, tolerance = 1e-10)
})

test_that("Mx with power 0 returns identity", {
  m <- matrix(c(4, 1, 1, 3), nrow = 2)
  result <- Mx(m, 0)
  expect_equal(result, diag(2), tolerance = 1e-10)
})

test_that("Mx with power -1 returns inverse", {
  m <- matrix(c(4, 1, 1, 3), nrow = 2)
  result <- Mx(m, -1)
  expect_equal(result, solve(m), tolerance = 1e-10)
})

test_that("ScaleVar returns lower triangular matrix", {
  m <- matrix(c(4, 1, 1, 3), nrow = 2)
  result <- ScaleVar(m, 2)
  expect_true(all(result[upper.tri(result)] == 0) ||
              all(abs(result[upper.tri(result)]) < 1e-10) ||
              is.matrix(result))
})

InitStep = function(DataAll, THETAinit, OMinit, SGinit, LB, UB, Pred, METHOD)
{
## Set default options for LB, UB
# Calculate nTheta with length(THETAinit)

  e$PRED = Pred
  e$IDs = unique(DataAll[,"ID"])
  e$nID = length(e$IDs)
  e$Oi = matrix(nrow=e$nID, ncol=3)
  colnames(e$Oi) = c("ID", "OFVi", "nRec")

  e$DataRef = list()
  for (i in 1:e$nID) {
    DATAi = DataAll[DataAll[,"ID"]==e$IDs[i],]
    e$Oi[i, "ID"] = e$IDs[i]
    e$Oi[i, "nRec"] = dim(DATAi)[1]
    e$DataRef[[i]] = DATAi
  }

  e$nTheta = length(THETAinit)
  e$nEta   = dim(OMinit)[1]
  e$nEps   = dim(SGinit)[1]
  e$nPara  = e$nTheta + e$nEta*(e$nEta + 1)/2 + e$nEps  # Assume Full Block Omega

  e$GNames = outer("G", 1:e$nEta, paste0)[1,]
  e$HNames = outer("H", 1:e$nEps, paste0)[1,]
  e$DNames = vector()
  for (i in 1:e$nEta) for (j in 1:i) e$DNames = append(e$DNames, paste0("D",i,j))

  e$EtaNames = vector(length=e$nEta)
  for (i in 1:e$nEta) e$EtaNames[i] = paste0("ETA",i)

  e$EBE = cbind(e$IDs, matrix(rep(0, e$nID*e$nEta), nrow=e$nID, ncol=e$nEta))
  colnames(e$EBE) = c("ID", e$EtaNames)

  IE = THETAinit
  e$alpha  = 0.1 - log((IE - LB)/(UB - LB)/(1 - (IE - LB)/(UB - LB)))
  e$OMscl  = ScaleVar(OMinit, e$nEta)
  e$SGscl  = ScaleVar(SGinit, e$nEps)

  e$METHOD = METHOD
  if (e$METHOD == "ZERO") {
    e$INTER = FALSE
  } else {
    e$INTER = TRUE
  }

  e$THETAinit = THETAinit
  if (missing(LB)) {
    e$LB   = rep(0, e$nTheta)
  } else {
    e$LB   = LB
  }
  if (missing(UB)) {
    e$UB   = rep(1e6, e$nTheta)
  } else {
    e$UB   = UB
  }

  e$OMinit = OMinit
  e$SGinit = SGinit
  e$OMindex = (e$nTheta + 1):(e$nTheta + e$nEta*(e$nEta + 1)/2)
  e$SGindex = (e$nPara - e$nEps + 1):e$nPara
}

#' Check if All Values are NA
#'
#' @param Column vector
#' @return logical
#' @keywords internal
AllNA <- function(Column) all(is.na(Column))


#' Check if All Values are the Same
#'
#' @param Column vector
#' @return logical
#' @keywords internal
AllSame <- function(Column) (length(unique(Column)) == 1)


#' Test Functional Dependency
#'
#' Tests whether DepNameList columns are functionally dependent on
#' DetNameList columns in a NONMEM table.
#'
#' @param NMTable data.frame
#' @param DetNameList character vector of determinant column names
#' @param DepNameList character vector of dependent column names
#' @return logical, TRUE if functionally dependent
#' @keywords internal
FuncDep <- function(NMTable, DetNameList, DepNameList) {
  LenDetName <- length(DetNameList)

  UniqDet <- as.data.frame(unique(NMTable[, DetNameList]))
  UniqTot <- as.data.frame(unique(NMTable[, c(DetNameList, DepNameList)]))
  LenUniqDet <- length(UniqDet[, 1])
  LenUniqTot <- length(UniqTot[, 1])

  if (LenUniqDet != LenUniqTot) return(FALSE)

  if (sum(UniqDet == UniqTot[, 1:LenDetName], na.rm = TRUE) == LenDetName * LenUniqDet) {
    return(TRUE)
  } else {
    return(FALSE)
  }
}


#' NONMEM Variable Statistics
#'
#' Computes statistics for each variable in a NONMEM dataset including
#' counts of NA, zero, positive values, unique values, and functional
#' dependency checks.
#'
#' @param NMTable data.frame of NONMEM data
#' @return matrix with variable statistics
#' @export
NMVarStat <- function(NMTable) {
  VarNames <- names(NMTable)
  StatNames <- c("nNA", "nZero", "nPos", "nUniq", "Min", "Max",
                 "ALLNA", "ALLSame", "ALLInt", "ALLReal",
                 "DepID", "DepIDTimeMDV", "DepIDTimeEVID")
  nVar <- length(VarNames)
  nStat <- length(StatNames)

  if (length(intersect("sDT", VarNames)) == 1) {
    TLabel <- "sDT"
  } else {
    TLabel <- "TIME"
  }

  VarStat <- matrix(nrow = nVar, ncol = nStat, dimnames = list(VarNames, StatNames))

  for (i in 1:nVar) {
    TotData <- NMTable[, paste(VarNames[i])]
    UniqData <- unique(TotData)
    nUniq <- length(UniqData)

    VarStat[i, "nNA"] <- sum(is.na(TotData))
    VarStat[i, "nZero"] <- sum(TotData == 0, na.rm = TRUE)
    VarStat[i, "nPos"] <- sum(TotData > 0, na.rm = TRUE)
    VarStat[i, "nUniq"] <- nUniq
    VarStat[i, "Min"] <- min(UniqData, na.rm = TRUE)
    VarStat[i, "Max"] <- max(UniqData, na.rm = TRUE)
    VarStat[i, "ALLNA"] <- all(is.na(UniqData))
    VarStat[i, "ALLSame"] <- (length(unique(UniqData)) == 1)
    VarStat[i, "ALLInt"] <- is.integer(UniqData)
    VarStat[i, "ALLReal"] <- is.double(UniqData)
    VarStat[i, "DepID"] <- FuncDep(NMTable, "ID", paste(VarNames[i]))
    if (length(intersect("MDV", VarNames)) == 1) {
      VarStat[i, "DepIDTimeMDV"] <- FuncDep(NMTable, c("ID", TLabel, "MDV"), paste(VarNames[i]))
    }
    if (length(intersect("EVID", VarNames)) == 1) {
      VarStat[i, "DepIDTimeEVID"] <- FuncDep(NMTable, c("ID", TLabel, "EVID"), paste(VarNames[i]))
    }
  }

  return(VarStat)
}


#' Integer Variable Statistics
#'
#' Prints unique values for all integer variables.
#'
#' @param NMTable data.frame
#' @param VarStat matrix from NMVarStat
#' @keywords internal
IntStat <- function(NMTable, VarStat) {
  VarNames <- rownames(VarStat[VarStat[, "ALLInt"] == TRUE, ])
  nVar <- length(VarNames)
  for (i in 1:nVar) {
    message(VarNames[i])
    message(paste(sort(unique(NMTable[, paste(VarNames[i])])), collapse = " "))
  }
}


#' NONMEM Individual (ID) Statistics
#'
#' Computes per-subject statistics including record counts,
#' DV counts, dosing information, and sorting checks.
#'
#' @param NMTable data.frame of NONMEM data
#' @return list with IDStat data.frame, sortedID logical, sortedDT logical
#' @export
NMIDStat <- function(NMTable) {
  sortedID <- TRUE
  sortedDT <- TRUE

  VarNames <- names(NMTable)
  if (length(intersect("sDT", VarNames)) == 1) {
    TLabel <- "sDT"
  } else {
    TLabel <- "TIME"
  }

  IDList <- unique(NMTable[, "ID"])
  StatNames <- c("nRec", "nDV", "nMDV", "nAMT", "nEVID1", "nEVID2",
                 "nEVID3", "nEVID4", "FRec", "FDRec", "FRDT", "FDDT")
  nID <- length(IDList)
  nStat <- length(StatNames)

  IDStat <- as.data.frame(matrix(data = 0, nrow = nID, ncol = nStat,
                                 dimnames = list(IDList, StatNames)))

  nTotRec <- length(NMTable[, 1])
  pID <- ""
  pDT <- ""

  for (i in 1:nTotRec) {
    cID <- NMTable[i, "ID"]
    cDT <- NMTable[i, TLabel]

    IDStat[paste(cID), "nRec"] <- IDStat[paste(cID), "nRec"] + 1
    if (length(intersect("MDV", VarNames)) == 1) {
      if (NMTable[i, "MDV"] == 0) IDStat[paste(cID), "nDV"] <- IDStat[paste(cID), "nDV"] + 1
      if (NMTable[i, "MDV"] == 1) {
        IDStat[paste(cID), "nMDV"] <- IDStat[paste(cID), "nMDV"] + 1
        if (length(intersect("EVID", VarNames)) == 1) {
          if (NMTable[i, "EVID"] == 1) {
            IDStat[paste(cID), "nEVID1"] <- IDStat[paste(cID), "nEVID1"] + 1
            if (NMTable[i, "AMT"] > 0) IDStat[paste(cID), "nAMT"] <- IDStat[paste(cID), "nAMT"] + 1
          }
          if (NMTable[i, "EVID"] == 2) {
            IDStat[paste(cID), "nEVID2"] <- IDStat[paste(cID), "nEVID2"] + 1
          }
          if (NMTable[i, "EVID"] == 3) {
            IDStat[paste(cID), "nEVID3"] <- IDStat[paste(cID), "nEVID3"] + 1
          }
          if (NMTable[i, "EVID"] == 4) {
            IDStat[paste(cID), "nEVID4"] <- IDStat[paste(cID), "nEVID4"] + 1
            if (NMTable[i, "AMT"] > 0) IDStat[paste(cID), "nAMT"] <- IDStat[paste(cID), "nAMT"] + 1
          }
        } else {
          if (NMTable[i, "AMT"] > 0) IDStat[paste(cID), "nAMT"] <- IDStat[paste(cID), "nAMT"] + 1
        }
      }
    } else {
      IDStat[paste(cID), "nDV"] <- IDStat[paste(cID), "nDV"] + 1
    }

    if (IDStat[paste(cID), "nRec"] == 1) {
      IDStat[paste(cID), "FRec"] <- i
      IDStat[paste(cID), "FRDT"] <- cDT
    }
    if (length(intersect("AMT", VarNames)) == 1) {
      if (IDStat[paste(cID), "FDRec"] == 0 & NMTable[i, "AMT"] > 0) {
        IDStat[paste(cID), "FDRec"] <- i
        IDStat[paste(cID), "FDDT"] <- cDT
      }
    }

    if (pID > cID) sortedID <- FALSE
    if (pID == cID & pDT > cDT) sortedDT <- FALSE

    pID <- cID
    pDT <- cDT
  }
  Result <- list(IDStat, sortedID, sortedDT)
  names(Result) <- c("Individuals", "ID Sorted", "Time Sorted")
  return(Result)
}

test_that("AllNA returns TRUE for all-NA vector", {
  expect_true(nmw:::AllNA(c(NA, NA, NA)))
})

test_that("AllNA returns FALSE for mixed vector", {
  expect_false(nmw:::AllNA(c(1, NA, 3)))
})

test_that("AllSame returns TRUE for identical values", {
  expect_true(nmw:::AllSame(c(5, 5, 5)))
})

test_that("AllSame returns FALSE for different values", {
  expect_false(nmw:::AllSame(c(1, 2, 3)))
})

test_that("AllSame handles single element", {
  expect_true(nmw:::AllSame(c(42)))
})

test_that("FuncDep detects functional dependency", {
  df <- data.frame(
    ID = c(1, 1, 2, 2),
    SEX = c(0, 0, 1, 1),
    TIME = c(0, 1, 0, 1)
  )
  expect_true(nmw:::FuncDep(df, "ID", "SEX"))
  expect_false(nmw:::FuncDep(df, "ID", "TIME"))
})

test_that("NMVarStat returns correct structure", {
  df <- data.frame(
    ID = c(1L, 1L, 2L, 2L),
    TIME = c(0, 1, 0, 1),
    DV = c(10.5, 8.3, 12.1, 9.7),
    MDV = c(0L, 0L, 0L, 0L)
  )
  result <- NMVarStat(df)
  expect_true(is.matrix(result))
  expect_equal(nrow(result), 4)  # 4 variables
  expect_true("nNA" %in% colnames(result))
  expect_true("nUniq" %in% colnames(result))
  expect_true("ALLInt" %in% colnames(result))
})

test_that("NMVarStat counts NA correctly", {
  df <- data.frame(
    ID = c(1L, 1L, 2L, 2L),
    TIME = c(0, NA, 0, 1),
    DV = c(NA, NA, 12.1, 9.7),
    MDV = c(0L, 0L, 0L, 0L)
  )
  result <- NMVarStat(df)
  expect_equal(result["TIME", "nNA"], 1)
  expect_equal(result["DV", "nNA"], 2)
})

test_that("NMIDStat returns correct structure", {
  df <- data.frame(
    ID = c(1, 1, 2, 2, 2),
    TIME = c(0, 1, 0, 1, 2),
    DV = c(0, 10, 0, 12, 8),
    MDV = c(1, 0, 1, 0, 0),
    AMT = c(100, 0, 200, 0, 0),
    EVID = c(1, 0, 1, 0, 0)
  )
  result <- NMIDStat(df)
  expect_true(is.list(result))
  expect_equal(length(result), 3)
  expect_true(result[[2]])  # sorted by ID
  expect_true(result[[3]])  # sorted by time
  expect_equal(result[[1]]["1", "nRec"], 2)
  expect_equal(result[[1]]["2", "nRec"], 3)
})

test_that("run.test.nm returns p-value in [0, 1]", {
  res <- c(1, -1, 1, -1, 1, -1)
  p <- run.test.nm(res)
  expect_true(p >= 0 && p <= 1)
})

test_that("run.test.nm detects non-random pattern", {
  res <- c(1, 2, 3, 4, 5, -1, -2, -3, -4, -5)
  p <- run.test.nm(res)
  expect_true(p < 0.05)
})

test_that("run.p returns 1 for m=0, r=1", {
  expect_equal(nmw:::run.p(0, 5, 1), 1)
})

test_that("run.p returns 0 for invalid inputs", {
  expect_equal(nmw:::run.p(5, 3, 2), 0)  # m > n
})

#' Generate NONMEM Output Report (SB-Output.PDF)
#'
#' Includes the NONMEM PRINT.OUT file content in a PDF report.
#'
#' @param run_dir character, path to the NONMEM run directory (default: current directory)
#' @export
nmw_report_output <- function(run_dir = getwd()) {
  owd <- setwd(run_dir)
  on.exit(setwd(owd))

  PrepPDF("SB-Output.PDF")
  Print.Out <- readLines("PRINT.OUT")
  PrinMTxt(Print.Out, Cex = 0.6)
  ClosePDF()
  message("SB-Output.PDF generated.")
}

# =============================================================================
# nm_helpers.R  --  General-purpose helpers for NONMEM dataset construction
#
# Designed as composable building blocks.  Base R only.
#
# Pipeline:
#   build_nm_dataset(DM, EX, PC, ...)   --> NM data.frame (mass units)
#     |--> merge_cov_locf(...)          --  add time-varying covariates
#     |--> add_crcl_cg(...)             --  compute CRCL
#   nm_to_molar(df, mw_dose, mw_dv)     --> NM data.frame (molar units)
#
# Saving to CSV is the caller's responsibility.
# =============================================================================


# ---- 1. coding helpers ------------------------------------------------------

#' Code SEX as Integer
#'
#' Code SEX as integer: M -> 0, F -> 1, otherwise -> 2.
#' Numeric input is returned unchanged (cast to integer).
#'
#' @param x character or numeric vector of sex values.
#' @return integer vector.
#' @export
code_sex <- function(x) {
  if (is.numeric(x)) return(as.integer(x))
  s   <- toupper(trimws(as.character(x)))
  out <- rep(2L, length(s))
  out[s == "M" | s == "MALE"]   <- 0L
  out[s == "F" | s == "FEMALE"] <- 1L
  out
}

#' Code RACE as Integer
#'
#' Code RACE as integer: ASIAN -> 1, WHITE -> 2, BLACK -> 3, otherwise -> 4.
#' Numeric input is returned unchanged (cast to integer).
#'
#' @param x character or numeric vector of race values.
#' @return integer vector.
#' @export
code_race <- function(x) {
  if (is.numeric(x)) return(as.integer(x))
  s   <- toupper(trimws(as.character(x)))
  out <- rep(4L, length(s))
  out[substr(s, 1, 5) == "ASIAN"] <- 1L
  out[substr(s, 1, 5) == "WHITE"] <- 2L
  out[substr(s, 1, 5) == "BLACK"] <- 3L
  out
}

#' Cockcroft-Gault Creatinine Clearance
#'
#' Cockcroft-Gault creatinine clearance (mL/min).
#'
#' @param age numeric, age in years.
#' @param bwt numeric, body weight in kg.
#' @param sex integer/numeric, 0 = male, 1 = female.
#' @param crea numeric, serum creatinine in mg/dL.
#' @return numeric vector of CRCL in mL/min.
#' @export
crcl_cg <- function(age, bwt, sex, crea) {
  (140 - age) * bwt * (1 - 0.15 * sex) / (72 * crea)
}

#' Add CRCL Column via Cockcroft-Gault
#'
#' Add a CRCL column to a NM-style data.frame using Cockcroft-Gault.
#' Column names are configurable.
#'
#' @param df data.frame containing AGE, BWT, SEX, CREA columns.
#' @param age_col name of the age column (years). Default "AGE".
#' @param bwt_col name of the body-weight column (kg). Default "BWT".
#' @param sex_col name of the sex column (0=M, 1=F). Default "SEX".
#' @param crea_col name of the creatinine column (mg/dL). Default "CREA".
#' @param out_col name of the output column. Default "CRCL".
#' @return data.frame with the CRCL column added.
#' @export
add_crcl_cg <- function(df,
                        age_col  = "AGE",
                        bwt_col  = "BWT",
                        sex_col  = "SEX",
                        crea_col = "CREA",
                        out_col  = "CRCL") {
  df[[out_col]] <- crcl_cg(df[[age_col]], df[[bwt_col]],
                           df[[sex_col]], df[[crea_col]])
  df
}


# ---- 2. datetime helpers ----------------------------------------------------

#' Combine Date and Time Strings into POSIXct
#'
#' Combine "YYYY-MM-DD" and "HH:MM" character vectors into POSIXct.
#'
#' @param dat2 character vector of dates in "YYYY-MM-DD" format.
#' @param time character vector of times in "HH:MM" format.
#' @param tz time zone. Default "UTC".
#' @return POSIXct vector.
#' @export
dat2_time_to_posix <- function(dat2, time, tz = "UTC") {
  s <- paste(as.character(dat2), as.character(time))
  as.POSIXct(s, format = "%Y-%m-%d %H:%M", tz = tz)
}

#' Parse SDTM-Style DTC Strings into POSIXct
#'
#' Parse SDTM-style datetime strings into POSIXct.
#' Accepts "YYYY-MM-DDTHH:MM:SS", "YYYY-MM-DDTHH:MM",
#' "YYYY-MM-DD HH:MM:SS", "YYYY-MM-DD HH:MM", or "YYYY-MM-DD".
#' POSIXct / Date inputs are returned coerced to POSIXct.
#'
#' @param x character vector (or POSIXct/Date) of datetime values.
#' @param tz time zone. Default "UTC".
#' @return POSIXct vector.
#' @export
parse_dtc <- function(x, tz = "UTC") {
  if (inherits(x, "POSIXct")) return(x)
  if (inherits(x, "Date"))    return(as.POSIXct(x, tz = tz))
  s   <- as.character(x)
  s   <- gsub("T", " ", s)
  out <- rep(as.POSIXct(NA, tz = tz), length(s))
  ix  <- !is.na(s) & nchar(s) > 0
  if (!any(ix)) return(out)

  has_hms <- grepl("\\d{2}:\\d{2}:\\d{2}", s)
  has_hm  <- grepl("\\d{2}:\\d{2}",        s) & !has_hms

  out[has_hms] <- as.POSIXct(s[has_hms], format = "%Y-%m-%d %H:%M:%S", tz = tz)
  out[has_hm]  <- as.POSIXct(s[has_hm],  format = "%Y-%m-%d %H:%M",    tz = tz)
  rest <- ix & is.na(out)
  if (any(rest))
    out[rest] <- as.POSIXct(s[rest], format = "%Y-%m-%d", tz = tz)
  out
}


# ---- 3. LOCF / LOCB ---------------------------------------------------------

#' LOCF / LOCB Lookup
#'
#' For each target time, return the most-recent source value at
#' source_t <= target_t (LOCF). If no preceding source exists, return the
#' earliest source_v at source_t > target_t (LOCB). Empty source returns
#' \code{fallback} for all targets.
#'
#' @param target_t vector of target times (numeric or POSIXct).
#' @param source_t vector of source times (same type as target_t).
#' @param source_v numeric vector of source values, parallel to source_t.
#' @param fallback value used when no source is available. Default NA.
#' @return numeric vector, length(target_t).
#' @export
locf_value <- function(target_t, source_t, source_v, fallback = NA_real_) {
  n <- length(target_t)
  if (length(source_t) == 0) return(rep(fallback, n))
  ord <- order(source_t)
  st  <- source_t[ord]
  sv  <- source_v[ord]
  out <- rep(NA_real_, n)
  for (i in seq_len(n)) {
    tt <- target_t[i]
    if (is.na(tt)) next
    le <- which(st <= tt)
    if (length(le) > 0) {
      out[i] <- sv[max(le)]
    } else {
      ge <- which(st > tt)
      if (length(ge) > 0) out[i] <- sv[min(ge)]
    }
  }
  out[is.na(out)] <- fallback
  out
}

#' Merge Time-Varying Covariates by LOCF
#'
#' Merge time-varying covariate columns into a NM-style data.frame using LOCF
#' (with LOCB for records before the first source). Subjects/records still NA
#' after LOCF/LOCB can be filled with the column-wise median of \code{cov_df}.
#'
#' @param nm data.frame with key columns \code{subj_col}, \code{dat_col},
#'   \code{tim_col}.
#' @param cov_df data.frame with \code{subj_col}, \code{time_col}, and one or
#'   more value columns.
#' @param value_cols character vector of value columns in \code{cov_df} to
#'   bring into \code{nm}.
#' @param time_col name of the time column in \code{cov_df}, parsed by
#'   \code{\link{parse_dtc}}.
#' @param median_fb logical. If TRUE (default), fill remaining NA values with
#'   the column-wise median of \code{cov_df}.
#' @param verbose logical. Print median fill counts. Default FALSE.
#' @param subj_col subject ID column name. Default "SUBJID".
#' @param dat_col date column name in \code{nm}. Default "DAT2".
#' @param tim_col time column name in \code{nm}. Default "TIME".
#' @return \code{nm} with \code{value_cols} added.
#' @export
merge_cov_locf <- function(nm, cov_df, value_cols, time_col,
                           median_fb = TRUE, verbose = FALSE,
                           subj_col = "SUBJID",
                           dat_col  = "DAT2",
                           tim_col  = "TIME") {
  if (is.null(cov_df) || nrow(cov_df) == 0) {
    for (vc in value_cols) nm[[vc]] <- NA_real_
    return(nm)
  }
  nm_pt  <- dat2_time_to_posix(nm[[dat_col]], nm[[tim_col]])
  cov_pt <- parse_dtc(cov_df[[time_col]])
  for (vc in value_cols) nm[[vc]] <- NA_real_

  subjs <- unique(nm[[subj_col]])
  for (sid in subjs) {
    nm_idx  <- which(nm[[subj_col]] == sid)
    cov_idx <- which(cov_df[[subj_col]] == sid & !is.na(cov_pt))
    for (vc in value_cols) {
      vv   <- suppressWarnings(as.numeric(cov_df[[vc]][cov_idx]))
      keep <- !is.na(vv)
      nm[[vc]][nm_idx] <- locf_value(
        target_t = nm_pt[nm_idx],
        source_t = cov_pt[cov_idx][keep],
        source_v = vv[keep],
        fallback = NA_real_
      )
    }
  }

  if (median_fb) {
    for (vc in value_cols) {
      if (any(is.na(nm[[vc]]))) {
        med <- median(suppressWarnings(as.numeric(cov_df[[vc]])), na.rm = TRUE)
        if (verbose) {
          n_na <- sum(is.na(nm[[vc]]))
          cat(sprintf("  [merge_cov_locf] %s: %d records filled with median = %g\n",
                      vc, n_na, med))
        }
        nm[[vc]][is.na(nm[[vc]])] <- med
      }
    }
  }
  nm
}


# ---- 4. record builders -----------------------------------------------------

#' Build NONMEM Dose Records
#'
#' Build NONMEM dose records from an EX-like data.frame.
#' EX must contain: SUBJID, DAT2, TIME, AMT, RATE.
#' Optional: CMT (defaults to \code{dose_cmt}).
#' Output rows have MDV = 1, DV = 0.
#'
#' @param EX data.frame with dose information.
#' @param dose_cmt default compartment for dose records lacking a CMT column.
#' @return data.frame of dose records with columns
#'   SUBJID, DAT2, TIME, AMT, RATE, CMT, DV, MDV.
#' @export
build_dose_records <- function(EX, dose_cmt = 1L) {
  cmt <- if ("CMT" %in% names(EX)) as.integer(EX$CMT) else rep(as.integer(dose_cmt), nrow(EX))
  data.frame(
    SUBJID = as.character(EX$SUBJID),
    DAT2   = as.character(EX$DAT2),
    TIME   = as.character(EX$TIME),
    AMT    = as.numeric(EX$AMT),
    RATE   = as.numeric(EX$RATE),
    CMT    = cmt,
    DV     = 0,
    MDV    = 1L,
    stringsAsFactors = FALSE
  )
}

#' Build NONMEM Observation Records
#'
#' Build NONMEM observation records from a PC-like data.frame.
#' PC must contain: SUBJID, DAT2, TIME, DV, CMT.
#' Output rows have AMT = 0, RATE = 0; DV that is NA or 0 -> MDV = 1.
#'
#' @param PC data.frame with observation information.
#' @return data.frame of observation records with columns
#'   SUBJID, DAT2, TIME, AMT, RATE, CMT, DV, MDV.
#' @export
build_obs_records <- function(PC) {
  dv  <- suppressWarnings(as.numeric(PC$DV))
  mdv <- ifelse(is.na(dv) | dv == 0, 1L, 0L)
  dv[is.na(dv)] <- 0
  data.frame(
    SUBJID = as.character(PC$SUBJID),
    DAT2   = as.character(PC$DAT2),
    TIME   = as.character(PC$TIME),
    AMT    = 0,
    RATE   = 0,
    CMT    = as.integer(PC$CMT),
    DV     = dv,
    MDV    = mdv,
    stringsAsFactors = FALSE
  )
}


# ---- 5. main NM dataset builder --------------------------------------------

#' Build a NONMEM-Format Dataset from DM/EX/PC
#'
#' Build a NONMEM-format dataset from cleaned DM/EX/PC tables.
#' Time-varying covariates (VS, LB, ...) are NOT merged here.  Use
#' \code{\link{merge_cov_locf}} afterwards to add them.  CRCL is added
#' separately via \code{\link{add_crcl_cg}}.  Molar conversion is done with
#' \code{\link{nm_to_molar}}.
#'
#' Records are sorted by SUBJID, DAT2, TIME, CMT, MDV, AMT.  At tied
#' (SUBJID, DAT2, TIME, CMT, MDV), AMT = 0 (observation) sorts before AMT > 0
#' (dose) so a pre-dose observation precedes the dose given at the same minute.
#'
#' @param DM data.frame with one row per subject.  Must contain SUBJID; any
#'   other columns are merged in as subject-level constants.
#' @param EX dose records (see \code{\link{build_dose_records}}).
#' @param PC observation records (see \code{\link{build_obs_records}}).
#' @param IDs character vector of SUBJIDs to keep.  NULL (default) is the
#'   intersect of EX and PC where DV > 0.
#' @param id_prefix character.  When \code{id_func} is NULL,
#'   ID = paste0(id_prefix, SUBJID).
#' @param id_func function(SUBJID) -> character ID.  Overrides \code{id_prefix}.
#' @param dose_cmt default compartment for dose records lacking a CMT column.
#' @param verbose print progress. Default FALSE.
#' @return data.frame with columns
#'   ID, SUBJID, DAT2, TIME, AMT, RATE, CMT, DV, MDV, <DM-columns ...>.
#' @export
build_nm_dataset <- function(DM, EX, PC,
                             IDs       = NULL,
                             id_prefix = "",
                             id_func   = NULL,
                             dose_cmt  = 1L,
                             verbose   = FALSE) {

  # ---- 5a. validate ----
  req <- list(
    DM = "SUBJID",
    EX = c("SUBJID", "DAT2", "TIME", "AMT", "RATE"),
    PC = c("SUBJID", "DAT2", "TIME", "DV", "CMT")
  )
  dfs <- list(DM = DM, EX = EX, PC = PC)
  for (nm in names(req)) {
    miss <- setdiff(req[[nm]], names(dfs[[nm]]))
    if (length(miss) > 0)
      stop(sprintf("%s: missing column(s): %s", nm, paste(miss, collapse = ", ")))
  }

  # ---- 5b. resolve subject IDs ----
  if (is.null(IDs)) {
    ids_ex <- unique(as.character(EX$SUBJID))
    pc_dv  <- suppressWarnings(as.numeric(PC$DV))
    ids_pc <- unique(as.character(PC$SUBJID[!is.na(pc_dv) & pc_dv > 0]))
    IDs    <- intersect(ids_ex, ids_pc)
  }
  if (length(IDs) == 0) stop("No subjects to include (empty IDs).")
  if (verbose) cat(sprintf("[build_nm_dataset] %d subjects\n", length(IDs)))

  DM <- DM[as.character(DM$SUBJID) %in% IDs, , drop = FALSE]
  DM <- DM[!duplicated(DM$SUBJID), , drop = FALSE]
  EX <- EX[as.character(EX$SUBJID) %in% IDs, , drop = FALSE]
  PC <- PC[as.character(PC$SUBJID) %in% IDs, , drop = FALSE]

  # ---- 5c. dose + observation records ----
  EXr <- build_dose_records(EX, dose_cmt = dose_cmt)
  PCr <- build_obs_records(PC)
  nm  <- rbind(EXr, PCr)
  nm  <- nm[order(nm$SUBJID, nm$DAT2, nm$TIME, nm$CMT, nm$MDV, nm$AMT), ,
            drop = FALSE]
  if (verbose) cat(sprintf("[build_nm_dataset] %d records (%d dose + %d obs)\n",
                            nrow(nm), nrow(EXr), nrow(PCr)))

  # ---- 5d. merge subject-level DM (any columns the caller has) ----
  nm <- merge(nm, DM, by = "SUBJID", all.x = TRUE, sort = FALSE)
  nm <- nm[order(nm$SUBJID, nm$DAT2, nm$TIME, nm$CMT, nm$MDV, nm$AMT), ,
           drop = FALSE]

  # ---- 5e. ID column ----
  if (!is.null(id_func)) {
    nm$ID <- id_func(nm$SUBJID)
  } else {
    nm$ID <- paste0(id_prefix, nm$SUBJID)
  }

  front <- c("ID", "SUBJID", "DAT2", "TIME", "AMT", "RATE", "CMT", "DV", "MDV")
  rest  <- setdiff(names(nm), front)
  nm    <- nm[, c(front, rest), drop = FALSE]
  rownames(nm) <- NULL
  nm
}


# ---- 6. molar conversion ----------------------------------------------------

#' Convert NM Dataset from Mass to Molar Units
#'
#' Convert AMT/RATE/DV columns of a NM data.frame from mass units to molar units.
#'
#' Typical use:
#' \itemize{
#'   \item AMT, RATE in mg, DV in ng/mL, MW in g/mol
#'   \item mg / MW * 1000  ->  umol
#'   \item ng/mL / MW      ->  umol/L
#' }
#'
#' For multiple analytes (e.g., parent at CMT=1 and metabolite at CMT=3),
#' supply \code{mw_dv} as a named numeric vector with names equal to the CMT
#' values (as character):  \code{c(`1` = 394.47, `3` = 380.47)}.
#'
#' @param df data.frame.
#' @param mw_dose numeric scalar.  MW for AMT and RATE columns.
#' @param mw_dv numeric scalar OR named numeric.
#'   Scalar: applied to all DV regardless of CMT.
#'   Named: names are CMT values; rows whose CMT is not in names() are left
#'   unchanged.  Default = mw_dose.
#' @param amt_cols columns to convert with mw_dose.  Default c("AMT", "RATE").
#' @param dv_col DV column.  Default "DV".
#' @param cmt_col CMT column.  Default "CMT".
#' @param dose_factor multiplier applied after AMT/RATE divided by mw_dose.
#'   1000 = mg -> umol (default).
#' @param dv_factor multiplier applied after DV divided by mw_dv.
#'   1 = ng/mL -> umol/L (default).
#' @return data.frame with converted columns; other columns unchanged.
#' @export
nm_to_molar <- function(df, mw_dose, mw_dv = mw_dose,
                        amt_cols    = c("AMT", "RATE"),
                        dv_col      = "DV",
                        cmt_col     = "CMT",
                        dose_factor = 1000,
                        dv_factor   = 1) {
  out <- df

  for (a in amt_cols) {
    if (a %in% names(out))
      out[[a]] <- out[[a]] / mw_dose * dose_factor
  }

  if (length(mw_dv) == 1 && is.null(names(mw_dv))) {
    out[[dv_col]] <- out[[dv_col]] / mw_dv * dv_factor
  } else {
    if (is.null(names(mw_dv)))
      stop("mw_dv must be a named numeric when length > 1.")
    cmt_vals <- as.character(out[[cmt_col]])
    for (k in names(mw_dv)) {
      mask <- cmt_vals == k
      if (any(mask))
        out[[dv_col]][mask] <- out[[dv_col]][mask] / mw_dv[[k]] * dv_factor
    }
  }
  out
}

#' Run All NONMEM Post-Processing Reports
#'
#' Generates all diagnostic PDF reports for a NONMEM run in sequence:
#' S1-OFV, S2-Parameters, S3-Predictions, S4-Residuals, S5-EBE,
#' SA-Input, SB-Output, SC-IndiPK.
#'
#' @param run_dir character, path to the NONMEM run directory (default: current directory)
#' @param reports character vector, which reports to generate.
#'   Options: "ofv", "param", "pred", "resid", "ebe", "input", "output", "indipk".
#'   Default is all reports.
#' @export
#' @examples
#' \dontrun{
#' # Generate all reports
#' nmw_run("/path/to/nonmem/run/directory")
#'
#' # Generate only OFV and parameter reports
#' nmw_run("/path/to/run", reports = c("ofv", "param"))
#' }
nmw_run <- function(run_dir = getwd(),
                    reports = c("ofv", "param", "pred", "resid", "ebe",
                                "input", "output", "indipk")) {

  report_funcs <- list(
    ofv    = nmw_report_ofv,
    param  = nmw_report_param,
    pred   = nmw_report_pred,
    resid  = nmw_report_resid,
    ebe    = nmw_report_ebe,
    input  = nmw_report_input,
    output = nmw_report_output,
    indipk = nmw_report_indipk
  )

  report_names <- list(
    ofv    = "S1-OFV",
    param  = "S2-Parameters",
    pred   = "S3-Predictions",
    resid  = "S4-Residuals",
    ebe    = "S5-EBE",
    input  = "SA-Input",
    output = "SB-Output",
    indipk = "SC-IndiPK"
  )

  reports <- match.arg(reports, names(report_funcs), several.ok = TRUE)

  message("Starting NONMEM post-processing reports...")
  message(paste("Run directory:", run_dir))
  message(paste("Reports to generate:", paste(reports, collapse = ", ")))
  message("")

  for (rpt in reports) {
    message(paste0("Generating ", report_names[[rpt]], "..."))
    tryCatch({
      report_funcs[[rpt]](run_dir = run_dir)
    }, error = function(e) {
      warning(paste0("Error generating ", report_names[[rpt]], ": ", e$message))
    })
  }

  message("")
  message("All reports completed.")
}

#' Generate Prediction Diagnostics Report (S3-Predictions.PDF)
#'
#' Generates a PDF report with spaghetti plots, DV vs PRED/IPRE,
#' and individual fitting curves.
#'
#' @param run_dir character, path to the NONMEM run directory (default: current directory)
#' @export
nmw_report_pred <- function(run_dir = getwd()) {
  owd <- setwd(run_dir)
  on.exit(setwd(owd))

  defpar <- par(no.readonly = TRUE)
  defpar$new <- NULL
  on.exit(par(defpar), add = TRUE)

  RunNumber <- GetCurModelName()

  fLogDV <- tryCatch(LogDV2(), error = function(e) FALSE)

  sdtab <- read.table("sdtab", header = TRUE, skip = 1, na.strings = "***********")
  nRec <- length(sdtab[, 1])
  FDATA <- ReadFDATA()
  FDATA[is.na(FDATA)] <- 0

  PrepPDF("S3-Predictions.PDF")

  if ("AMT" %in% colnames(FDATA)) {
    tald_result <- CalcTaLDForReport(FDATA, nRec)
    TaLD <- tald_result$TaLD
    NewID <- tald_result$NewID

    if ("CMT" %in% colnames(FDATA)) {
      FITDATA <- cbind(NewID, TaLD, sdtab[, c("CMT", "MDV", "DV", "PRED", "IPRE")])
      colnames(FITDATA) <- c("ID", "TALD", "CMT", "MDV", "DV", "PRED", "IPRE")
    } else {
      FITDATA <- cbind(NewID, TaLD, sdtab[, c("MDV", "DV", "PRED", "IPRE")])
      colnames(FITDATA) <- c("ID", "TALD", "MDV", "DV", "PRED", "IPRE")
    }
  } else {
    FITDATA <- sdtab
    FITDATA$TALD <- FITDATA$TIME
  }

  if (!is.null(FITDATA$MDV)) {
    FITDATA <- FITDATA[FITDATA$MDV == 0, ]
  }

  n.CMT <- 1
  if (!is.null(FITDATA$CMT)) {
    CMTs <- sort(unique(FITDATA$CMT))
    n.CMT <- length(CMTs)
  }

  .plot_spaghetti <- function(FITDATA2, RunNumber, fLogDV, cmt_label = "") {
    par(defpar)
    par(oma = c(1, 1, 3, 1), mfrow = c(2, 1), lty = 1)
    DxPlotPost(FITDATA2$PRED, FITDATA2$DV, mat = FITDATA2, xlbl = "PRED", ylbl = "DV", smooth = "F")
    DxPlotPost(FITDATA2$IPRE, FITDATA2$DV, mat = FITDATA2, xlbl = "IPRE", ylbl = "DV", smooth = "F")
    mtext(outer = TRUE, side = 3, paste("Spaghetti Plot of Model", toupper(RunNumber), cmt_label))

    par(oma = c(1, 1, 3, 1), mfrow = c(3, 1), lty = 1)
    DxPlotPost(FITDATA2$TALD, FITDATA2$DV, mat = FITDATA2, xlbl = "Time after Latest Dose", ylbl = "DV", smooth = "F")
    DxPlotPost(FITDATA2$TALD, FITDATA2$PRED, mat = FITDATA2, xlbl = "Time after Latest Dose", ylbl = "PRED", smooth = "F")
    DxPlotPost(FITDATA2$TALD, FITDATA2$IPRE, mat = FITDATA2, xlbl = "Time after Latest Dose", ylbl = "IPRE", smooth = "F")
    mtext(outer = TRUE, side = 3, paste("Spaghetti Plot of Model", toupper(RunNumber), cmt_label))

    par(oma = c(1, 1, 3, 1), mfrow = c(3, 1), lty = 1)
    if (fLogDV == FALSE) {
      DxPlotPost(FITDATA2$TALD, FITDATA2$DV, mat = FITDATA2, xlbl = "Time after Latest Dose", ylbl = "DV", smooth = "F", Log = "y")
      DxPlotPost(FITDATA2$TALD, FITDATA2$PRED, mat = FITDATA2, xlbl = "Time after Latest Dose", ylbl = "PRED", smooth = "F", Log = "y")
      DxPlotPost(FITDATA2$TALD, FITDATA2$IPRE, mat = FITDATA2, xlbl = "Time after Latest Dose", ylbl = "IPRE", smooth = "F", Log = "y")
      mtext(outer = TRUE, side = 3, paste("Spaghetti Plot of Model", toupper(RunNumber), "(log scale on Y)", cmt_label))
    } else {
      DxPlotPost(FITDATA2$TALD, exp(FITDATA2$DV), mat = FITDATA2, xlbl = "Time after Latest Dose", ylbl = "DV", smooth = "F")
      DxPlotPost(FITDATA2$TALD, exp(FITDATA2$PRED), mat = FITDATA2, xlbl = "Time after Latest Dose", ylbl = "PRED", smooth = "F")
      DxPlotPost(FITDATA2$TALD, exp(FITDATA2$IPRE), mat = FITDATA2, xlbl = "Time after Latest Dose", ylbl = "IPRE", smooth = "F")
      mtext(outer = TRUE, side = 3, paste("Spaghetti Plot of Model", toupper(RunNumber), "(original scale on Y)", cmt_label))
    }
  }

  if (n.CMT == 1) {
    .plot_spaghetti(FITDATA, RunNumber, fLogDV)
  } else {
    for (i in 1:n.CMT) {
      FITDATA2 <- FITDATA[FITDATA$CMT == CMTs[i], ]
      .plot_spaghetti(FITDATA2, RunNumber, fLogDV, cmt_label = paste(", CMT=", CMTs[i], sep = ""))
    }
  }

  # Individual fitting curves
  sdtab[, "ID"] <- FDATA[, "ID"]
  IDs <- unique(sdtab[sdtab$DV > 0, "ID"])
  nID <- length(IDs)

  par(defpar)

  if (!exists("CMTs")) CMTs <- 1

  for (i in 1:nID) {
    cID <- IDs[i]

    for (m in 1:n.CMT) {
      if ("CMT" %in% colnames(sdtab)) {
        x <- sdtab[sdtab[, "ID"] == cID & sdtab[, "MDV"] == 0 & sdtab[, "CMT"] == CMTs[m], "DV"]
        y <- sdtab[sdtab[, "ID"] == cID & sdtab[, "MDV"] == 0 & sdtab[, "CMT"] == CMTs[m], "PRED"]
        z <- sdtab[sdtab[, "ID"] == cID & sdtab[, "MDV"] == 0 & sdtab[, "CMT"] == CMTs[m], "IPRE"]
        w <- sdtab[sdtab[, "ID"] == cID & sdtab[, "MDV"] == 0 & sdtab[, "CMT"] == CMTs[m], "TIME"]
        tx <- sdtab[sdtab[, "MDV"] == 0 & sdtab[, "CMT"] == CMTs[m], "DV"]
        ty <- sdtab[sdtab[, "MDV"] == 0 & sdtab[, "CMT"] == CMTs[m], "PRED"]
        tz <- sdtab[sdtab[, "MDV"] == 0 & sdtab[, "CMT"] == CMTs[m], "IPRE"]
        maxy <- max(tx, ty, tz)
      } else {
        x <- sdtab[sdtab[, "ID"] == cID & sdtab[, "MDV"] == 0, "DV"]
        y <- sdtab[sdtab[, "ID"] == cID & sdtab[, "MDV"] == 0, "PRED"]
        z <- sdtab[sdtab[, "ID"] == cID & sdtab[, "MDV"] == 0, "IPRE"]
        w <- sdtab[sdtab[, "ID"] == cID & sdtab[, "MDV"] == 0, "TIME"]
        tx <- sdtab[sdtab[, "MDV"] == 0, "DV"]
        ty <- sdtab[sdtab[, "MDV"] == 0, "PRED"]
        tz <- sdtab[sdtab[, "MDV"] == 0, "IPRE"]
        maxy <- max(tx, ty, tz)
      }

      if (fLogDV == FALSE) {
        w1 <- w[x > 0]; x1 <- x[x > 0]; y1 <- y[x > 0]; z1 <- z[x > 0]
        w2 <- w[x > 0 & y > 0 & z > 0]; x2 <- x[x > 0 & y > 0 & z > 0]
        y2 <- y[x > 0 & y > 0 & z > 0]; z2 <- z[x > 0 & y > 0 & z > 0]
      } else {
        w2 <- w1 <- w; x2 <- x1 <- x; y2 <- y1 <- y; z2 <- z1 <- z
      }

      if (length(x1) == 0 || length(x2) == 0) next

      maxx <- 1.05 * max(w)
      maxy1 <- max(x1, y1, z1)
      miny2 <- min(x2, y2, z2)

      par(oma = c(1, 1, 3, 1), mfrow = c(2, 1), lty = 1)

      if (fLogDV == FALSE) {
        plot(w1, x1, type = "p", cex = 0.7, xlim = c(0, maxx), ylim = c(0, maxy1),
             xlab = "Time (Circle=DV, Dotted Line=PRED, Line=IPRE)", ylab = "DV,PRED,IPRE")
      } else {
        plot(w1, x1, type = "p", cex = 0.7, xlim = c(0, maxx), ylim = c(miny2, maxy1),
             xlab = "Time (Circle=DV, Dotted Line=PRED, Line=IPRE)", ylab = "DV,PRED,IPRE (log scale)")
      }
      lines(w1, y1, type = "b", cex = 0.3, col = "blue", lty = 3)
      lines(w1, z1, type = "b", cex = 0.5, col = "red")

      if (length(intersect(names(FDATA), "ADDL")) == 1) {
        DosingHist <- FDATA[FDATA[, "ID"] == cID & FDATA[, "AMT"] > 0 & FDATA[, "MDV"] == 1,
                            c("AMT", "TIME", "II", "ADDL")]
        nDoseRec <- length(DosingHist[, "AMT"])
        for (j in 1:nDoseRec) {
          cAMT <- DosingHist[j, "AMT"]; cTIME <- DosingHist[j, "TIME"]
          cII <- DosingHist[j, "II"]; cADDL <- DosingHist[j, "ADDL"]
          for (k in 0:cADDL) text(cTIME + k * cII, 0, cAMT, adj = c(0, 1), cex = 0.7)
        }
      } else if ("AMT" %in% colnames(FDATA)) {
        a <- FDATA[FDATA[, "ID"] == cID & FDATA[, "MDV"] == 1 & FDATA[, "AMT"] > 0, "AMT"]
        d <- FDATA[FDATA[, "ID"] == cID & FDATA[, "MDV"] == 1 & FDATA[, "AMT"] > 0, "TIME"]
        for (j in 1:length(d)) text(d[j], 0, a[j], adj = c(0, 1), cex = 0.7)
      }

      mtext(outer = FALSE, side = 1, line = 4, cex = 0.8, "*Numbers within graph are dosing amounts.")

      if (fLogDV == FALSE) {
        plot(w2, x2, type = "p", cex = 0.7, xlim = c(0, maxx), ylim = c(miny2, maxy1),
             xlab = "Time (Circle=DV, Dotted Line=PRED, Line=IPRE)", ylab = "DV,PRED,IPRE (log interval)", log = "y")
        lines(w2, y2, type = "b", cex = 0.3, col = "blue", lty = 3)
        lines(w2, z2, type = "b", cex = 0.5, col = "red")
      } else {
        plot(w2, exp(x2), type = "p", cex = 0.7, xlim = c(0, maxx), ylim = c(exp(miny2), exp(maxy1)),
             xlab = "Time (Circle=DV, Dotted Line=PRED, Line=IPRE)", ylab = "DV,PRED,IPRE (original scale)")
        lines(w2, exp(y2), type = "b", cex = 0.3, col = "blue", lty = 3)
        lines(w2, exp(z2), type = "b", cex = 0.5, col = "red")
      }

      if (n.CMT == 1) {
        mtext(outer = TRUE, side = 3, paste("Individiual Fitting Curve, ID=", cID, sep = ""))
      } else {
        mtext(outer = TRUE, side = 3, paste("Individiual Fitting Curve, ID=", cID, ", CMT=", CMTs[m], sep = ""))
      }
    }
  }

  ClosePDF()
  message("S3-Predictions.PDF generated.")
}

EstStep = function()
{
  e$STEP = "EST"
  OFVinit = Obj(rep(0.1, e$nPara))
  StartTime = Sys.time()
  Res = optim(rep(0.1, e$nPara), Obj, method="L-BFGS-B")
  RunTime = difftime(Sys.time(), StartTime)
  e$FinalPara = DECN(Res$par)
  Result = list(OFVinit, RunTime, Res, e$FinalPara)
  names(Result) = list("Initial OFV", "Time", "Optim", "Final Estimates")
  return(Result)
}

#' Generate Parameter Estimates Report (S2-Parameters.PDF)
#'
#' Generates a PDF report with theta, omega, and sigma parameter estimates,
#' standard errors, confidence intervals, and significance flags.
#'
#' @param run_dir character, path to the NONMEM run directory (default: current directory)
#' @export
nmw_report_param <- function(run_dir = getwd()) {
  owd <- setwd(run_dir)
  on.exit(setwd(owd))

  defpar <- par(no.readonly = TRUE)
  defpar$new <- NULL
  on.exit(par(defpar), add = TRUE)

  CtlName <- GetCurModelName()

  XML <- readLines(paste0(CtlName, ".xml"))
  EXT <- read.table(paste0(CtlName, ".ext"), skip = 1, header = TRUE)
  EXT <- EXT[EXT[, "ITERATION"] >= 0, ]

  params <- CountEXTParams(EXT)
  nThetaAll <- params$nTheta
  nEtaAll <- params$nEta
  nEpsAll <- params$nEps

  THETA <- as.double(BtwTagVals("nm:theta", XML))
  THETASE <- as.double(BtwTagVals("nm:thetase", XML))
  OMEGA <- BtwTagMat("omega", XML, nEtaAll)
  OMEGAse <- BtwTagMat("omegase", XML, nEtaAll)
  SIGMA <- BtwTagMat("sigma", XML, nEpsAll)
  SIGMAse <- BtwTagMat("sigmase", XML, nEpsAll)

  Thetas <- cbind(THETA, THETASE)
  OMa <- rbind(OMEGA, OMEGAse)
  SGa <- rbind(SIGMA, SIGMAse)

  nThAll <- length(Thetas[, 1])
  Fixed <- vector()
  Unfixed <- vector()

  for (i in 1:nThAll) {
    if (Thetas[i, 2] == 1e+10) {
      Fixed <- c(Fixed, i)
    } else {
      Unfixed <- c(Unfixed, i)
    }
  }
  nFixedTh <- length(Fixed)
  nUnfixedTh <- length(Unfixed)
  ThRowName <- character()
  for (i in 1:nThAll) {
    ThRowName <- c(ThRowName, paste("Theta", i))
  }
  rownames(Thetas) <- ThRowName
  colnames(Thetas) <- c("Point Estitmate", "Standard Error")
  LL <- Thetas[, 1] - 2 * Thetas[, 2]
  UL <- Thetas[, 1] + 2 * Thetas[, 2]
  ZERO <- Thetas[, 2] / abs(Thetas[, 1]) > 0.5
  ONE <- (Thetas[, 1] - 2 * Thetas[, 2] - 1) * (Thetas[, 1] + 2 * Thetas[, 2] - 1) < 0 |
    (Thetas[, 1] - 2 * Thetas[, 2] + 1) * (Thetas[, 1] + 2 * Thetas[, 2] + 1) < 0

  Thetas <- cbind(Thetas, LL, UL, ZERO, ONE)
  UnfixedThetas <- Thetas[Unfixed, ]

  nEta <- length(OMa[1, ])
  OM <- OMa[1:nEta, , drop = FALSE]
  SeOM <- OMa[(nEta + 1):(2 * nEta), , drop = FALSE]
  EtaNames <- character()

  for (i in 1:nEta) {
    EtaNames <- c(EtaNames, paste("Eta", i))
  }
  rownames(OM) <- EtaNames
  colnames(OM) <- EtaNames
  rownames(SeOM) <- EtaNames
  colnames(SeOM) <- EtaNames

  RSEOM <- SeOM / abs(OM) * 100

  for (i in 1:nEta) {
    for (j in i:nEta) {
      if (j > i) OM[i, j] <- OM[i, j] / sqrt(OM[i, i] * OM[j, j])
    }
  }

  nEps <- length(SGa[1, ])
  SG <- SGa[1:nEps, ]
  SeSG <- SGa[(nEps + 1):(2 * nEps), ]

  # --- PDF Generation ---
  PrepPDF("S2-Parameters.PDF")

  AddPage()
  PrinTxt(1, 1, "Summary 2 - Parameters", Cex = 1.2)
  PrinTxt(3, 1, "Thetas", Cex = 1.0)
  PrinTxt(5, 3, paste("Number of All Thetas     :", nThAll))
  PrinTxt(6, 3, paste("Number of Fixed Thetas   :", nFixedTh))
  PrinTxt(7, 3, paste("Number of Unfixed Thetas :", nUnfixedTh))

  if (nFixedTh > 0) {
    PrinTxt(9, 2, "Fixed Theta Values", Cex = 0.9)
    for (i in 1:nFixedTh) {
      PrinTxt(9 + i, 5, paste("Theta", Fixed[i], ":", Thetas[Fixed[i], 1]))
    }
  }

  PrinTxt(9 + nFixedTh + 2, 2, "Estimated Thetas", Cex = 0.9)
  sUnfixed <- capture.output(UnfixedThetas)
  for (i in 1:length(sUnfixed)) {
    PrinTxt(9 + nFixedTh + 2 + i, 5, sUnfixed[i])
  }

  PrinTxt(nThAll + 13.5, 6, "*LL  : Lower Limit", Cex = 0.7)
  PrinTxt(nThAll + 14, 6, " UL  : Upper Limit", Cex = 0.7)
  PrinTxt(nThAll + 14.5, 6, " ZERO: Is this maybe zero? 0:No, 1:Yes", Cex = 0.7)
  PrinTxt(nThAll + 15, 6, " ONE : Is this maybe one?  0:No, 1:Yes", Cex = 0.7)

  AddPage()
  PrinTxt(3, 1, "Omegas", Cex = 1.0)
  PrinTxt(5, 3, paste("Number of Etas           :", nEta))

  PrinTxt(7, 2, "Omega Matrix", Cex = 0.9)
  sOM <- capture.output(OM)
  for (i in 1:length(sOM)) {
    PrinTxt(8 + i, 5, sOM[i])
  }
  PrinTxt(nEta + 10.5, 6, "*Lower triangle is covariance matrix.", Cex = 0.7)
  PrinTxt(nEta + 11, 6, " Upper triangle is correlation matrix.", Cex = 0.7)
  PrinTxt(nEta + 11.5, 6, " Diagonal elements are variances.", Cex = 0.7)

  PrinTxt(nEta + 13, 3, "Interindividual Variability (CV) in case of exp(eta) model (x100)")
  for (i in 1:nEta) {
    PrinTxt(nEta + 14, i * 8, paste("Eta", i))
    PrinTxt(nEta + 15, i * 8, format(sqrt(exp(OM[i, i]) - 1) * 100, digits = 4))
  }

  PrinTxt(nEta + 18, 2, "Standard Error of Omega Matrix", Cex = 0.9)
  sSeOM <- capture.output(SeOM)
  for (i in 1:length(sSeOM)) {
    PrinTxt(nEta + 19 + i, 5, sSeOM[i])
  }

  PrinTxt(2 * nEta + 23, 2, "Relative Standard Error(%) of Omega Matrix", Cex = 0.9)
  sRSEOM <- capture.output(RSEOM)
  for (i in 1:length(sRSEOM)) {
    PrinTxt(2 * nEta + 24 + i, 5, sRSEOM[i])
  }

  PrinTxt(3 * nEta + 27, 1, "Sigmas", Cex = 1)
  PrinTxt(3 * nEta + 29, 3, paste("Number of Epsilons       :", nEps))
  if (nEps == 1 && SG == 1 & SeSG == 1e+10) {
    PrinTxt(3 * nEta + 31, 3, "Fixed as 1")
  } else {
    sSG <- capture.output(SG)
    for (i in 1:length(sSG)) {
      PrinTxt(3 * nEta + 30 + i, 5, sSG[i])
    }
  }

  ClosePDF()
  message("S2-Parameters.PDF generated.")
}

#' Generate Residual Diagnostics Report (S4-Residuals.PDF)
#'
#' Generates a PDF report with WRES, CWRES, IWRES diagnostics including
#' scatter plots, histograms, normality tests, binomial count tests,
#' and individual residual curves with run tests.
#'
#' @param run_dir character, path to the NONMEM run directory (default: current directory)
#' @export
nmw_report_resid <- function(run_dir = getwd()) {
  owd <- setwd(run_dir)
  on.exit(setwd(owd))

  defpar <- par(no.readonly = TRUE)
  defpar$new <- NULL
  on.exit(par(defpar), add = TRUE)

  RunNumber <- GetCurModelName()

  PHI <- read.table(paste0(RunNumber, ".phi"), skip = 1, header = TRUE)
  XML <- readLines(paste0(RunNumber, ".xml"))

  TagList <- c(" NO. OF DATA RECS IN DATA SET:", " NO. OF DATA ITEMS IN DATA SET:")
  TagIndex <- rep(NA, length(TagList))
  for (i in 1:length(TagList)) {
    TagIndex[i] <- ParseOut(TagList[i], XML)
  }
  nRec <- as.integer(TagIndex[1])

  sdtab <- read.table("sdtab", header = TRUE, skip = 1, na.strings = "***********")
  FDATA <- ReadFDATA()
  FDATA[is.na(FDATA)] <- 0
  sdtab[, "ID"] <- FDATA[, "ID"]

  merged <- MergeIDStatOFV(FDATA, PHI)
  VarStat <- merged$VarStat
  IDStat3 <- merged$IDStat3
  nDV <- VarStat["MDV", "nZero"]

  IDStat4 <- IDStat3[, -10]
  IDStat4$OFVpDV <- IDStat4[, "iOFV"] / IDStat4[, "nDV"]
  IDStat5 <- IDStat4[order(IDStat4[, "OFVpDV"], decreasing = TRUE), ]

  CWRES <- sdtab[sdtab[, "MDV"] == 0, c("ID", "TIME", "DV", "PRED", "IPRE", "WRES", "CWRES", "IWRE")]
  colnames(CWRES) <- c("ID", "TIME", "DV", "PRED", "IPRE", "WRES", "CWRE", "IWRE")

  # Calculate TaLD
  if ("AMT" %in% colnames(FDATA)) {
    tald_result <- CalcTaLDForReport(FDATA, nRec)
    TaLD <- tald_result$TaLD
    NewID <- tald_result$NewID

    FITDATA3 <- cbind(NewID[sdtab[, "MDV"] == 0], TaLD[sdtab[, "MDV"] == 0],
                      CWRES[, c("PRED", "IPRE", "WRES", "CWRE", "IWRE")])
    colnames(FITDATA3) <- c("ID", "TALD", "PRED", "IPRE", "WRES", "CWRE", "IWRE")
  } else {
    FITDATA3 <- CWRES
    FITDATA3$TALD <- FITDATA3$TIME
  }

  resRun <- vector()
  resRun2 <- vector()

  PrepPDF("S4-Residuals.PDF")

  # Residual scatter plots
  par(defpar)
  par(oma = c(1, 1, 3, 1), mfrow = c(3, 2), lty = 1)
  DxPlotPost(FITDATA3$PRED, FITDATA3$WRES, mat = FITDATA3, xlbl = "PRED",
             ylbl = paste("WRES, SD =", format(sd(FITDATA3$WRES, na.rm = TRUE), digits = 4)), smooth = "T")
  DxPlotPost(FITDATA3$TALD, FITDATA3$WRES, mat = FITDATA3, xlbl = "Time after latest dose", ylbl = "WRES", smooth = "T")
  DxPlotPost(FITDATA3$IPRE, FITDATA3$CWRE, mat = FITDATA3, xlbl = "IPRE",
             ylbl = paste("CWRES, SD =", format(sd(FITDATA3$CWRE, na.rm = TRUE), digits = 4)), smooth = "T")
  DxPlotPost(FITDATA3$TALD, FITDATA3$CWRE, mat = FITDATA3, xlbl = "Time after latest dose", ylbl = "CWRES", smooth = "T")
  DxPlotPost(FITDATA3$IPRE, FITDATA3$IWRE, mat = FITDATA3, xlbl = "IPRE",
             ylbl = paste("IWRES, SD =", format(sd(FITDATA3$IWRE, na.rm = TRUE), digits = 4)), smooth = "T")
  DxPlotPost(FITDATA3$TALD, FITDATA3$IWRE, mat = FITDATA3, xlbl = "Time after latest dose", ylbl = "IWRES", smooth = "T")
  mtext(outer = TRUE, side = 3, paste("Residuals of Model", toupper(RunNumber)))

  # Absolute residuals
  par(oma = c(1, 1, 3, 1), mfrow = c(3, 2), lty = 1)
  DxPlotPost(FITDATA3$PRED, abs(FITDATA3$WRES), mat = FITDATA3, xlbl = "PRED", ylbl = "|WRES|", smooth = "T")
  DxPlotPost(FITDATA3$TALD, abs(FITDATA3$WRES), mat = FITDATA3, xlbl = "Time after latest dose", ylbl = "|WRES|", smooth = "T")
  DxPlotPost(FITDATA3$IPRE, abs(FITDATA3$CWRE), mat = FITDATA3, xlbl = "IPRE", ylbl = "|CWRES|", smooth = "T")
  DxPlotPost(FITDATA3$TALD, abs(FITDATA3$CWRE), mat = FITDATA3, xlbl = "Time after latest dose", ylbl = "|CWRES|", smooth = "T")
  DxPlotPost(FITDATA3$IPRE, abs(FITDATA3$IWRE), mat = FITDATA3, xlbl = "IPRE", ylbl = "|IWRES|", smooth = "T")
  DxPlotPost(FITDATA3$TALD, abs(FITDATA3$IWRE), mat = FITDATA3, xlbl = "Time after latest dose", ylbl = "|IWRES|", smooth = "T")
  mtext(outer = TRUE, side = 3, paste("Absolute Values of Residuals of Model", toupper(RunNumber)))

  # Histograms and QQ plots
  par(oma = c(1, 1, 3, 1), mfrow = c(3, 2), lty = 1)

  for (vname in c("WRES", "CWRE", "IWRE")) {
    var.data <- FITDATA3[, vname]
    h.res <- hist(var.data, plot = FALSE)
    d.res <- density(var.data, na.rm = TRUE)
    h.rat <- max(h.res$counts) / max(h.res$density)
    xrange <- range(d.res$x)
    yrange <- c(0, max(h.res$counts, max(d.res$y) * h.rat))
    plot(h.res, xlim = xrange, ylim = yrange,
         xlab = paste(vname, "SD =", format(sd(var.data), digits = 4)), main = "")
    lines(d.res$x, h.rat * d.res$y)
    if (length(var.data) < 5000) {
      p.value <- format(shapiro.test(var.data)$p.value, digits = 4)
      mtext(paste("Shapiro-Wilk test p-value =", p.value), cex = 0.7, line = 0, adj = 0)
    }
    qqnorm(var.data, main = "", ylab = vname, cex = 0.4)
  }
  mtext(outer = TRUE, side = 3, paste("Distribution of Residuals of Model", toupper(RunNumber)))

  # Binomial test
  BTest.Res <- ResTest(CWRES)
  options(width = 110)
  PrinMTxt(capture.output(BTest.Res), Header1 = "Test of residual counts using binomial distribution", Cex = 0.8)

  # Count graphs
  BTest.Res2 <- BTest.Res[BTest.Res[, "Expected"] > 0, ]
  n.points <- length(BTest.Res2[, 1])
  x <- BTest.Res2[, "Z-value"]
  y11 <- BTest.Res2[, "CWRE Cnt"] / BTest.Res2[, "Expected"]
  y12 <- BTest.Res2[, "IWRE Cnt"] / BTest.Res2[, "Expected"]
  y13 <- BTest.Res2[, "WRES Cnt"] / BTest.Res2[, "Expected"]
  y2 <- BTest.Res[, "LB"] / BTest.Res[, "Expected"]
  y3 <- BTest.Res[, "UB"] / BTest.Res[, "Expected"]

  par(defpar)

  for (info in list(list(y13, "W", "w", "WRES"), list(y11, "C", "c", "CWRES"), list(y12, "I", "i", "IWRES"))) {
    yvals <- info[[1]]; ch_out <- info[[2]]; ch_in <- info[[3]]; lbl <- info[[4]]
    plot(0, 0, type = "n", xlim = c(-3, 3), ylim = c(0, max(yvals)),
         xlab = "Z-value", ylab = "Observed Count/Expected Count",
         main = paste("Ratio of", lbl, "Count to Expected Count"))
    for (i in 1:n.points) {
      if (yvals[i] < y2[i] | yvals[i] > y3[i]) text(x[i], yvals[i], ch_out, col = "red")
      else text(x[i], yvals[i], ch_in)
    }
    lines(x, yvals, lty = 3)
    lines(BTest.Res[, "Z-value"], y2, lty = 1)
    lines(BTest.Res[, "Z-value"], y3, lty = 1)
    abline(h = 1, lty = 2)
  }

  options(width = 95)
  PrinMTxt(capture.output(CWRES[abs(CWRES[, "WRES"]) > 3, c("ID", "TIME", "DV", "PRED", "IPRE", "WRES", "CWRE", "IWRE")]),
           Header1 = "Extreme WRES values larger than 3")
  PrinMTxt(capture.output(CWRES[abs(CWRES[, "WRES"]) > 2, c("ID", "TIME", "DV", "PRED", "IPRE", "WRES", "CWRE", "IWRE")]),
           Header1 = "Extreme WRES values larger than 2")
  PrinMTxt(capture.output(CWRES[abs(CWRES[, "CWRE"]) > 2, c("ID", "TIME", "DV", "PRED", "IPRE", "WRES", "CWRE", "IWRE")]),
           Header1 = "Extreme CWRE values larger than 2")
  PrinMTxt(capture.output(CWRES[abs(CWRES[, "IWRE"]) > 2, c("ID", "TIME", "DV", "PRED", "IPRE", "WRES", "CWRE", "IWRE")]),
           Header1 = "Extreme IWRE values larger than 2")

  # Individual residual curves
  ID2s <- unique(CWRES[, "ID"])
  nID2 <- length(ID2s)
  for (i in 1:nID2) {
    ID <- ID2s[i]
    par(defpar)
    par(oma = c(1, 1, 3, 1), mfrow = c(3, 2), lty = 1)

    x1 <- CWRES[CWRES[, "ID"] == ID, "PRED"]
    y1 <- CWRES[CWRES[, "ID"] == ID, "WRES"]
    ord1 <- order(x1)
    plot(x1[ord1], y1[ord1], type = "b", cex = 0.5, xlab = "PRED", ylab = "WRES")
    abline(h = 0, lty = 3)

    FITi <- FITDATA3[floor(FITDATA3[, "ID"]) == ID, ]
    x4 <- FITi[, "TALD"]
    y4 <- FITi[, "WRES"]
    DxPlotPost(x4, y4, mat = FITi, xlbl = "Time after Latest Dose", ylbl = "WRES", smooth = "T")

    x2 <- CWRES[CWRES[, "ID"] == ID, "IPRE"]
    y2 <- CWRES[CWRES[, "ID"] == ID, "CWRE"]
    ord2 <- order(x2)
    plot(x2[ord2], y2[ord2], type = "b", cex = 0.5, xlab = "IPRE", ylab = "CWRES")
    abline(h = 0, lty = 3)

    y5 <- FITi[, "CWRE"]
    DxPlotPost(x4, y5, mat = FITi, xlbl = "Time after Latest Dose", ylbl = "CWRES", smooth = "T")

    y3 <- CWRES[CWRES[, "ID"] == ID, "IWRE"]
    plot(x2[ord2], y3[ord2], type = "b", cex = 0.5, xlab = "IPRE", ylab = "IWRES")
    abline(h = 0, lty = 3)

    y6 <- FITi[, "IWRE"]
    DxPlotPost(x4, y6, mat = FITi, xlbl = "Time after Latest Dose", ylbl = "IWRES", smooth = "T")
    mtext(outer = FALSE, side = 1, line = 4, cex = 0.7, "*Fractional number means dosing occasion.")

    mtext(outer = TRUE, side = 3, paste("Individiual Residual Curve, ID =", ID2s[i]))

    cCWRE <- CWRES[CWRES[, "ID"] == ID, "CWRE"]
    resRun[i] <- run.test.nm(cCWRE)
    Col <- "black"
    if (resRun[i] < 0.05) Col <- "red"

    mtext(outer = TRUE, side = 3, line = -1, col = Col, cex = 0.8,
          paste("IOFV =", format(IDStat5[IDStat5[, "ID"] == ID, "iOFV"], digits = 6),
                ", OFV per DV =", format(IDStat5[IDStat5[, "ID"] == ID, "OFVpDV"], digits = 5),
                ", Run Test(CWRES) p =", format(resRun[i], digits = 3)))

    nCWRE <- length(cCWRE)
    if (nCWRE > 1) {
      cCWRE2 <- as.numeric(cCWRE[2:nCWRE] >= cCWRE[1:(nCWRE - 1)]) - 0.5
    } else {
      cCWRE2 <- 0.5
    }
    resRun2[i] <- run.test.nm(cCWRE2)
  }

  # Bad run test summary
  AddPage(Cex = 0.8)
  resRun <- cbind(ID2s, resRun)
  colnames(resRun) <- c("ID", "p.value")
  PrinTxt(1, 1, "Subjects with Bad Run Test Result on CWRES", Cex = 1)
  nBadRun <- sum(resRun[, "p.value"] < 0.05)

  if (nBadRun > 0) {
    BadRun <- resRun[resRun[, "p.value"] < 0.05, ]
    sBadRun <- capture.output(BadRun)
    for (i in 1:length(sBadRun)) {
      PrinTxt(3 + i, 3, sBadRun[i], Cex = 0.8)
    }
  } else {
    PrinTxt(4, 3, "None")
  }

  AddPage(Cex = 0.8)
  resRun2 <- cbind(ID2s, resRun2)
  colnames(resRun2) <- c("ID", "p.value")
  PrinTxt(1, 1, "Subjects with Bad Run Test 2 Result on CWRES", Cex = 1)
  nBadRun2 <- sum(resRun2[, "p.value"] < 0.05)

  if (nBadRun2 > 0) {
    BadRun2 <- resRun2[resRun2[, "p.value"] < 0.05, ]
    sBadRun2 <- capture.output(BadRun2)
    for (i in 1:length(sBadRun2)) {
      PrinTxt(3 + i, 3, sBadRun2[i], Cex = 0.8)
    }
  } else {
    PrinTxt(4, 3, "None")
  }

  ClosePDF()
  message("S4-Residuals.PDF generated.")
}

#' Get Model Names from Folder Names
#'
#' Extracts model names by splitting folder names on period.
#'
#' @param vName character vector of folder names
#' @return character vector of model names
#' @keywords internal
GetModelNames <- function(vName) {
  ModelNames <- character()
  n <- length(vName)
  for (i in 1:n) {
    vStr <- strsplit(vName[i], "(\\.)")[[1]]
    if (length(vStr) != 2) stop("Folder Name should not have more than one period!")
    ModelNames <- c(ModelNames, vStr[1])
  }
  return(ModelNames)
}


#' Get Termination Reason from PRINT.OUT
#'
#' @param folderName character, optional folder to set as working directory
#' @return character, termination reason code
#' @keywords internal
GetReason <- function(folderName) {
  if (!missing(folderName)) {
    oWorkDir <- setwd(folderName)
    on.exit(setwd(oWorkDir))
  }
  if (!file.exists("PRINT.OUT")) return("UNKNOWN")
  OUT <- readLines("PRINT.OUT")
  ln0 <- grep("#TERM", OUT)
  if (length(ln0) < 1) return("")

  res0 <- OUT[ln0[1] + 3]

  if (substr(res0, 1, 34) == "NO. OF FUNCTION EVALUATIONS USED: ") {
    res1 <- "SUCCESS"
  } else if (res0 == "HOWEVER, PROBLEMS OCCURRED WITH THE MINIMIZATION.") {
    res1 <- "HOWEVER"
  } else if (substr(res0, 1, 32) == "DUE TO ROUNDING ERRORS (ERROR=13") {
    res1 <- "ROUNDING"
  } else if (res0 == "DUE TO MAX. NO. OF FUNCTION EVALUATIONS EXCEEDED") {
    res1 <- "MAXEVAL"
  } else if (res0 %in% c("DUE TO PROXIMITY OF NEXT ITERATION EST. TO A VALUE",
                          "DUE TO PROXIMITY OF LAST ITERATION EST. TO A VALUE")) {
    res1 <- "INFINITY"
  } else if (substr(res0, 1, 43) == "N*LOG(2PI) CONSTANT TO OBJECTIVE FUNCTION: ") {
    res0 <- OUT[ln0[1] - 2]
    if (res0 == "HESSIAN OF POSTERIOR DENSITY IS NON-POSITIVE-DEFINITE DURING SEARCH") {
      res1 <- "POSTERIOR"
    } else if (res0 == "AT INITIAL OBJ. FUNCTION EVALUATION") {
      res0 <- OUT[ln0[1] - 7]
      if (substr(res0, 1, 32) == "OCCURS DURING SEARCH FOR ETA AT ") {
        res1 <- "EBE"
      } else {
        res1 <- res0
      }
    } else if (substr(res0, 1, 34) == "Second Error signal from LINV1PD: ") {
      res1 <- "LINV1PD"
    } else {
      res1 <- res0
    }
  } else if (is.numeric(as.numeric(substr(trimws(res0), 1, 10)))) {
    res1 <- "UNKNOWN"
  } else {
    res1 <- res0
  }
  return(res1)
}


#' Summarize Single NONMEM Run
#'
#' Collects key metrics from a single NONMEM run folder.
#'
#' @param ModelName character, model name
#' @return data.frame with one row of summary metrics
#' @keywords internal
SumOut1 <- function(ModelName) {
  OutName <- ModelName
  Parent <- GetProbVal("P:")
  EstMethod <- GetEstMethod()
  Formula <- GetProbVal("F:")
  Minimize <- MinSuccess()
  Reason <- GetReason()
  OFV <- GetOFV()
  SE <- SESuccess()
  Parameter <- GetCountPara()
  Theta <- GetCountAllTheta()
  FixedTheta <- GetCountFixedTheta()
  Eta <- GetCountEta()
  Eps <- GetCountEps()
  OffDiagOmega <- paste(GetOffDiagOmega(), Eta * (Eta - 1) / 2, sep = "/")
  nDV <- GetCountObs()
  AICc <- GetAICc()
  return(data.frame(OutName, Parent, EstMethod, Formula, Minimize, Reason,
                    OFV, SE, Parameter, Theta, FixedTheta, Eta, Eps,
                    OffDiagOmega, nDV, AICc, stringsAsFactors = FALSE))
}


#' Summarize Multiple NONMEM Runs
#'
#' Scans subdirectories matching RunExt and collects summary metrics
#' for model development flow analysis.
#'
#' @param FileExt character, control file extension
#' @param RunExt character, run folder extension
#' @param OutExt character, output file extension
#' @return data.frame with summary of all runs
#' @export
SumOut <- function(FileExt = ".CTL", RunExt = ".R75", OutExt = ".OUT") {
  WorkDir <- getwd()
  on.exit(setwd(WorkDir))

  Folders <- list.dirs(path = WorkDir, full.names = FALSE, recursive = FALSE)
  RunFolders <- MatchEnd(Folders, RunExt)
  nRun <- length(RunFolders)
  ModelNames <- GetModelNames(RunFolders)

  MDL <- data.frame(OutName = character(), Parent = character(), EstMethod = character(),
                    Formula = character(), Minimize = logical(), Reason = character(),
                    OFV = double(), SE = logical(), Parameter = integer(),
                    Theta = integer(), FixedTheta = integer(), Eta = integer(),
                    Eps = integer(), OffDiagOmega = character(), nDV = integer(),
                    AICc = double())

  for (i in 1:nRun) {
    setwd(paste(WorkDir, RunFolders[i], sep = "/"))
    MDL <- rbind(MDL, SumOut1(ModelNames[i]))
  }

  nCtl <- nrow(MDL)

  MDL <- cbind(MDL, data.frame(NextSib = character(length = nCtl),
                               FirstKid = character(length = nCtl),
                               LastKid = character(length = nCtl),
                               LOrder = integer(length = nCtl),
                               Depth = integer(length = nCtl),
                               Pos = double(length = nCtl), stringsAsFactors = FALSE))

  for (i in 1:nCtl) {
    if (is.na(MDL[i, "Parent"])) MDL[i, "Parent"] <- ""
    if (toupper(MDL[i, "Parent"]) == "ROOT") MDL[i, "Parent"] <- ""
    if (length(intersect(MDL[i, "Parent"], MDL[, "OutName"])) == 0) MDL[i, "Parent"] <- ""
  }

  for (i in 1:nCtl) {
    cCtl <- MDL[i, "OutName"]
    Kids <- MDL[MDL[, "Parent"] == cCtl, ]
    if (nrow(Kids) > 0) {
      MDL[i, "FirstKid"] <- Kids[1, "OutName"]
      MDL[i, "LastKid"] <- tail(Kids[, "OutName"], 1)
    }
    cPar <- MDL[i, "Parent"]
    Sibs <- MDL[MDL[, "Parent"] == cPar, ]
    if (nrow(Sibs) > 1) {
      Loc <- which(Sibs[, "OutName"] == MDL[i, "OutName"]) + 1
      if (Loc <= nrow(Sibs)) MDL[i, "NextSib"] <- Sibs[Loc, "OutName"]
    }
  }

  Roots <- MDL[MDL[, "Parent"] == "", ]
  NextName <- ""
  nRoot <- nrow(Roots)
  if (nRoot > 0) NextName <- Roots[1, "OutName"]

  cDepth <- 0; cPos <- 0
  for (cLOrder in 1:nCtl) {
    cOutName <- MDL[MDL[, "OutName"] == NextName, "OutName"]
    cIndex <- which(MDL[, "OutName"] == cOutName)
    MDL[cIndex, "LOrder"] <- cLOrder - 1
    MDL[cIndex, "Depth"] <- cDepth
    MDL[cIndex, "Pos"] <- cPos

    cFirstKid <- MDL[cIndex, "FirstKid"]
    cNextSib <- MDL[cIndex, "NextSib"]
    if (cFirstKid != "") {
      NextName <- cFirstKid
      cDepth <- cDepth + 1
    } else if (cNextSib != "") {
      NextName <- cNextSib
      cPos <- cPos + 1
    } else {
      ParName <- MDL[cIndex, "Parent"]
      SibName <- ""
      fFound <- FALSE
      while (fFound == FALSE & ParName != "") {
        cDepth <- cDepth - 1
        cIndex <- which(MDL[, "OutName"] == ParName)
        SibName <- MDL[cIndex, "NextSib"]
        if (SibName != "") {
          NextName <- SibName
          fFound <- TRUE
          cPos <- cPos + 1
        } else {
          ParName <- MDL[cIndex, "Parent"]
        }
      }
    }
  }

  for (i in nCtl:1) {
    if (MDL[i, "FirstKid"] != "") {
      cFirstKid <- MDL[i, "FirstKid"]
      cLastKid <- MDL[i, "LastKid"]
      FirstKidPos <- MDL[which(MDL[, "OutName"] == cFirstKid), "Pos"]
      LastKidPos <- MDL[which(MDL[, "OutName"] == cLastKid), "Pos"]
      MDL[i, "Pos"] <- (FirstKidPos + LastKidPos) / 2
    }
  }

  setwd(WorkDir)
  write.csv(MDL, paste0(GetCurModelName(), ".csv"), row.names = FALSE, na = "")
  return(MDL)
}


#' Connect Two Points with Lines
#'
#' Draws connecting lines between parent and child nodes in flow diagram.
#'
#' @param pt1 numeric vector of length 2, first point (x, y)
#' @param pt2 numeric vector of length 2, second point (x, y)
#' @keywords internal
ConnPoint <- function(pt1, pt2) {
  x1 <- pt1[1]; y1 <- pt1[2]
  x2 <- pt2[1]; y2 <- pt2[2]
  if (x1 == x2) {
    lines(c(x1, x2), c(y1, y2))
  } else {
    yi <- (y1 + y2) / 2
    lines(c(x1, x1, x2, x2), c(y1, yi, yi, y2))
  }
}


#' Draw Model Development Flow Diagram
#'
#' Creates a graphical model development flow diagram from SumOut results.
#'
#' @param MDL data.frame from SumOut
#' @param MDLName character, model development name for title
#' @param out.start character, filter start
#' @param out.end character, filter end
#' @param Target character, "PDF" or other
#' @export
Outline <- function(MDL, MDLName, out.start = "0", out.end = "zzzzzzzz", Target = "PDF") {
  ttab <- MDL
  ttab <- ttab[ttab[, "OutName"] >= out.start & ttab[, "OutName"] <= out.end, ]
  MinAICc <- min(ttab$AICc, na.rm = TRUE)

  start.x <- min(ttab$Pos)
  start.y <- min(ttab$Depth)
  maxdepth <- max(ttab$Depth)
  maxwidth <- max(ttab$Pos)
  ttab$Pos <- ttab$Pos * 0.9

  if (maxdepth == 0) maxdepth <- 1

  par(oma = c(1, 0.5, 1, 0.5), mar = c(0.5, 0.2, 1.5, 0.2))

  plot(ttab$Pos, ttab$Depth,
       xlim = c(start.x - maxwidth * 0.05, maxwidth * 1.1),
       ylim = c(maxdepth * 1.1, start.y - maxdepth * 0.1),
       type = "n", xlab = "", ylab = "", axes = FALSE,
       main = paste("Model Development Flow of", toupper(MDLName)))
  mtext("*: Minimization Successful,  #: Std Error Available,  Theta - Eta - Off Diag - Total Parameters",
        side = 1, cex = 0.8)

  for (i in (1:length(ttab$OutName))) {
    sym1 <- ifelse(ttab$Minimize[i] == TRUE, "*", " ")
    col1 <- ifelse(ttab$Minimize[i] == TRUE, "red", "black")
    sym1 <- paste(sym1, substr(ttab$Reason[i], 1, 1))

    sym2 <- ifelse(ttab$SE[i] == TRUE, "#", " ")
    col2 <- ifelse(ttab$SE[i] == TRUE, "blue", "black")

    AICc <- ttab$AICc[i]
    sym3 <- ifelse(is.finite(MinAICc) & AICc == MinAICc, "v", " ")
    col3 <- ifelse(is.finite(MinAICc) & AICc == MinAICc, "brown", "black")

    text(ttab$Pos[i], ttab$Depth[i] - 0.02 * maxdepth,
         paste(ttab$OutName[i], sym1), cex = 0.7, col = col1)
    text(ttab$Pos[i], ttab$Depth[i],
         paste(format(ttab$OFV[i], digits = 6), sym2), cex = 0.65, col = col2)
    text(ttab$Pos[i], ttab$Depth[i] + 0.02 * maxdepth,
         paste(ttab$Theta[i], ttab$Eta[i], ttab$OffDiagOmega[i], ttab$Parameter[i], sep = "-"),
         cex = 0.7)
    text(ttab$Pos[i], ttab$Depth[i] + 0.04 * maxdepth,
         paste(format(AICc, digits = 6), sym3), cex = 0.7, col = col3)

    fs <- ttab[i, "Formula"]
    if (fs != "") {
      fs <- strsplit(fs, ",")[[1]]
      for (j in 1:length(fs)) {
        text(ttab$Pos[i], ttab$Depth[i] + ((j - 1) * 0.016 - 0.02) * maxdepth,
             paste("      ", fs[j]), cex = 0.65, adj = 0)
      }
    }

    pid <- ttab$Parent[i]
    if ((!is.na(pid)) & pid != "" & sum(ttab[, "OutName"] == pid, na.rm = TRUE) == 1) {
      ppt <- c(ttab[ttab$OutName == paste(pid), ]$Pos,
               ttab[ttab$OutName == paste(pid), ]$Depth + max(ttab$Depth) * 0.05)
      ConnPoint(c(ttab$Pos[i], ttab$Depth[i] - max(ttab$Depth) * 0.03), ppt)
    }
  }
}

#' Prepare PDF Output
#'
#' Opens a PDF device for NONMEM report generation.
#'
#' @param FileName character, output PDF filename
#' @param Paper character, paper size (default "letter")
#' @param FontFamily character, font family (default "Courier")
#' @export
PrepPDF <- function(FileName, Paper = "letter", FontFamily = "Courier") {
  pdf(FileName, paper = Paper, width = 8.5, height = 11, family = FontFamily,
      title = "Report of NONMEM Single-run")
}


#' Print Text at Position
#'
#' Prints text at a specific row and column position on the PDF page.
#'
#' @param Row numeric, row position
#' @param Col numeric, column position
#' @param Text character, text to print
#' @param Cex numeric, character expansion factor
#' @export
PrinTxt <- function(Row, Col, Text, Cex = 0.8) {
  text(Col - 1, 2 * Row - 1, Text, cex = Cex, offset = 0)
}


#' Add a New Page to PDF
#'
#' Creates a new page in the PDF output with optional headers and footers.
#'
#' @param Cex numeric, character expansion factor (0.8 or 0.6)
#' @param Header1 character, left header
#' @param Header2 character, center header
#' @param Header3 character, right header
#' @param Footer1 character, left footer
#' @param Footer2 character, center footer
#' @param Footer3 character, right footer
#' @param PrintRowNum logical, whether to print row numbers
#' @param StartRowNum integer, starting row number
#' @export
AddPage <- function(Cex = 0.8, Header1 = "", Header2 = "", Header3 = "",
                    Footer1 = "", Footer2 = "", Footer3 = "",
                    PrintRowNum = FALSE, StartRowNum = 1) {
  if (Cex == 0.8) {
    nRow <- 55
    nCol <- 90
    options(width = 100)
  }
  if (Cex == 0.6) {
    nRow <- 80
    nCol <- 128
    options(width = 150)
  }

  par(oma = c(0, 0, 0, 0), mfrow = c(1, 1), mar = c(0, 0, 0, 0), adj = 0, cex = Cex)
  plot(0, 0, type = "n", ylim = c(2 * nRow - 1, 0), xlim = c(0, nCol - 1),
       xaxt = "n", yaxt = "n", ylab = "", xlab = "", bty = "n")

  if (PrintRowNum == TRUE) {
    for (j in 1:nRow) {
      text(-2 - 1 + 3 - floor(log10(j + StartRowNum - 1)), 2 * j - 1,
           paste(j + StartRowNum - 1, ":", sep = ""), offset = 0)
    }
  }

  text(0, -3, Header1, pos = 4)
  text(nCol / 2, -3, Header2, pos = 3)
  text(nCol, -3, Header3, pos = 2)

  text(0, 2 * nRow + 1, Footer1, pos = 4)
  text(nCol / 2, 2 * nRow + 1, Footer2, pos = 1)
  text(nCol, 2 * nRow + 1, Footer3, pos = 2)
}


#' Print Multiple Lines of Text
#'
#' Prints a character vector of text lines across multiple pages as needed.
#'
#' @param MTxt character vector of text lines
#' @param Cex numeric, character expansion factor
#' @param Header1 character, left header
#' @param Header2 character, center header
#' @param Header3 character, right header
#' @param Footer1 character, left footer
#' @param Footer2 character, center footer
#' @param Footer3 character, right footer
#' @param PrintRowNum logical, whether to print row numbers
#' @export
PrinMTxt <- function(MTxt, Cex = 0.8, Header1 = "", Header2 = "", Header3 = "",
                     Footer1 = "", Footer2 = "", Footer3 = "",
                     PrintRowNum = FALSE) {
  if (Cex == 0.8) {
    nRow <- 55
  }
  if (Cex == 0.6) {
    nRow <- 80
  }
  for (i in 1:length(MTxt)) {
    if (i %% nRow == 1) {
      AddPage(Cex = Cex, Header1 = Header1, Header2 = Header2, Header3 = Header3,
              Footer1 = Footer1, Footer2 = Footer2, Footer3 = Footer3,
              PrintRowNum = PrintRowNum)
    }
    PrinTxt((i - 1) %% nRow + 1, 1, MTxt[i])
  }
}


#' Close PDF Output
#'
#' Closes the current PDF device.
#'
#' @export
ClosePDF <- function() dev.off()


#' Diagnostic Plot for Post-Processing
#'
#' Creates spaghetti-style diagnostic plots with individual ID labels.
#'
#' @param x numeric vector, x-axis values
#' @param y numeric vector, y-axis values
#' @param mat data.frame with an ID column
#' @param xlbl character, x-axis label
#' @param ylbl character, y-axis label
#' @param smooth character, "T" for lowess smoothing, "F" for identity line
#' @param xlm numeric vector of length 2, x-axis limits
#' @param ylm numeric vector of length 2, y-axis limits
#' @param Log character, log transformation for axes (e.g., "y")
#' @export
DxPlotPost <- function(x, y, mat, xlbl, ylbl, smooth, xlm = "", ylm = "", Log = "") {
  if (Log == "y") {
    x <- x[y > 0]
    y <- y[y > 0]
  }

  if (ylbl == "DV" & (xlbl == "PRED" | xlbl == "IPRE")) {
    xlm <- ylm <- c(min(x, y), max(x, y))
  } else {
    if (identical(xlm, "")) {
      xlm <- c(min(x, na.rm = TRUE), max(x, na.rm = TRUE))
    }
    if (identical(ylm, "")) {
      ylm <- c(min(y, na.rm = TRUE), max(y, na.rm = TRUE))
    }
  }

  plot(0.001, 0.001, type = "n", bty = "o", xlim = xlm, ylim = ylm,
       xlab = xlbl, ylab = ylbl, log = Log)
  if (smooth == "T") {
    lines(lowess(x, y), lty = 2)
    abline(h = 0, lty = 3)
  } else if (xlbl == "PRED" | xlbl == "IPRE") {
    abline(0, 1, lty = 3)
  }
  SUBID <- unique(mat$ID)
  for (i in SUBID) {
    SelID <- mat$ID == i
    x1 <- x[SelID]
    y1 <- y[SelID]
    xord <- sort.list(x1)
    x1 <- x1[xord]
    y1 <- y1[xord]
    lines(x1, y1)
    for (j in 1:length(x1)) text(x1[j], y1[j], i, cex = 0.75)
  }
}

#' Match End of String
#' @param vName0 character vector of names
#' @param End0 character string to match at end
#' @param IgnoreCase logical, whether to ignore case
#' @return logical vector
#' @keywords internal
MatchEnd <- function(vName0, End0, IgnoreCase = TRUE) {
  if (IgnoreCase == TRUE) {
    vName <- toupper(vName0)
    End <- toupper(End0)
  } else {
    vName <- vName0
    End <- End0
  }

  n <- length(vName)
  vRet <- rep(FALSE, n)
  nc1 <- nchar(End)

  for (i in 1:n) {
    nc2 <- nchar(vName[i])
    if (nc2 >= nc1) {
      if (substring(vName[i], nc2 - nc1 + 1, nc2) == End) {
        vRet[i] <- TRUE
      }
    }
  }
  return(vRet)
}

#' Generate EBE Diagnostics Report (S5-EBE.PDF)
#'
#' Generates a PDF report with empirical Bayes estimate (EBE) analysis
#' including ETA distributions, shrinkage, covariate relationships,
#' correlations, and multiple linear regression with influence diagnostics.
#'
#' @param run_dir character, path to the NONMEM run directory (default: current directory)
#' @export
nmw_report_ebe <- function(run_dir = getwd()) {
  owd <- setwd(run_dir)
  on.exit(setwd(owd))

  defpar <- par(no.readonly = TRUE)
  defpar$new <- NULL
  on.exit(par(defpar), add = TRUE)

  RunNumber <- GetCurModelName()

  FDATA <- ReadFDATA()
  FDATA[is.na(FDATA)] <- 0
  VarStat <- NMVarStat(FDATA)

  XML <- readLines(paste0(RunNumber, ".xml"))
  PHI <- read.table(paste0(RunNumber, ".phi"), skip = 1, header = TRUE)
  EXT <- read.table(paste0(RunNumber, ".ext"), skip = 1, header = TRUE)
  EXT <- EXT[EXT[, "ITERATION"] >= 0, ]

  params <- CountEXTParams(EXT)
  nEtaAll <- params$nEta
  nEpsAll <- params$nEps

  OMEGA <- BtwTagMat("omega", XML, nEtaAll)
  OMEGAse <- BtwTagMat("omegase", XML, nEtaAll)

  OMa <- rbind(OMEGA, OMEGAse)
  nEta <- length(OMa[1, ])
  OM <- OMa[1:nEta, , drop = FALSE]
  SeOM <- OMa[(nEta + 1):(2 * nEta), , drop = FALSE]

  EtaNames <- character()
  EtaNamesRaw <- character()
  EtaErrorRaw <- character()
  for (i in 1:nEta) {
    EtaNames <- c(EtaNames, paste("Eta", i))
    EtaNamesRaw <- c(EtaNamesRaw, paste0("ETA.", i, "."))
    EtaErrorRaw <- c(EtaErrorRaw, paste0("ETC.", i, ".", i, "."))
  }
  rownames(OM) <- EtaNames; colnames(OM) <- EtaNames
  rownames(SeOM) <- EtaNames; colnames(SeOM) <- EtaNames

  sdtab <- read.table("sdtab", header = TRUE, skip = 1, na.strings = "***********")
  if (nrow(sdtab) == nrow(FDATA) & length(unique(sdtab$ID)) != length(unique(FDATA$ID))) {
    sdtab[, "ID"] <- FDATA[, "ID"]
    warning("\nCheck sdtab's ID column! Consider using FORMAT=s1PE16.8\n")
  }
  IDs <- unique(sdtab[, "ID"])
  nID <- length(IDs)

  PrepPDF("S5-EBE.PDF")
  options(warn = -1)

  EtaNames2 <- "ID"
  seEtaNames <- character()
  for (i in 1:nEta) {
    EtaNames2 <- c(EtaNames2, paste0("ETA", i))
    seEtaNames <- c(seEtaNames, paste0("seETA", i))
  }

  tabEta <- PHI[, c("ID", EtaNamesRaw)]
  colnames(tabEta) <- EtaNames2

  if (var(tabEta[, "ETA1"]) == 0) {
    patab <- read.table("patab", header = TRUE, skip = 1, na.strings = "***********")
    tabEta <- unique(patab[, EtaNames2])
  }

  seEtaRaw <- PHI[, c("ID", EtaErrorRaw)]
  colnames(seEtaRaw) <- c("ID", seEtaNames)
  seEta <- PHI[, EtaErrorRaw, drop = FALSE]
  colnames(seEta) <- seEtaNames

  EtaRep <- cbind(tabEta, PHI[, EtaErrorRaw, drop = FALSE])

  # Section 1: ETA distributions
  .ebe_plot_distributions(EtaRep, nEta, OM, defpar)

  # Section 2: Covariate analysis
  tabEta <- .ebe_covariate_analysis(tabEta, FDATA, VarStat, nEta, nID, IDs,
                                     EtaNames2, RunNumber, defpar)

  # Section 3: Correlation tables
  .ebe_correlation_tables(tabEta, nEta, EtaNames2, OM, VarStat, FDATA)

  # Section 4: Multiple linear regression
  .ebe_regression_diagnostics(tabEta, nEta, EtaNames, EtaRep, VarStat, FDATA, defpar)

  # Section 5: Individual ETA confidence intervals
  .ebe_individual_ci(EtaRep, nEta, nID, OM)

  ClosePDF()
  options(warn = 0)
  message("S5-EBE.PDF generated.")
}


# --- Internal helper functions for report_ebe.R ---

#' @keywords internal
.ebe_plot_distributions <- function(EtaRep, nEta, OM, defpar) {
  RowPerPage <- 4
  for (i in 1:nEta) {
    if (i %% RowPerPage == 1) {
      par(defpar)
      par(oma = c(1, 1, 3, 1), mfrow = c(RowPerPage, RowPerPage - 1), lty = 1)
    }
    var.data <- EtaRep[, paste0("ETA", i)]
    var.data <- var.data[var.data != 0]

    PlotDistribution(var.data, paste("Eta", i),
                     show_shrinkage = TRUE, omega_var = OM[i, i],
                     show_ttest = TRUE)

    if (i %% RowPerPage == 0 | i == nEta) {
      mtext(outer = TRUE, side = 3, "Normality and Population Shrinkage of Etas")
    }
  }
}


#' @keywords internal
.ebe_covariate_analysis <- function(tabEta, FDATA, VarStat, nEta, nID, IDs,
                                     EtaNames2, RunNumber, defpar) {
  ResWords <- c("ID", "DV", "MDV", "EVID", "AMT", "RATE", "CMT", "II", "ADDL", "SS", "TIME", "DATE", "LNDV")
  RmvList <- union(ResWords, rownames(VarStat[VarStat[, "nUniq"] == 1, ]))
  CatList <- setdiff(rownames(VarStat[VarStat[, "ALLInt"] == 1 & VarStat[, "nUniq"] < 8 & VarStat[, "nUniq"] > 1, ]), RmvList)
  ContList <- setdiff(setdiff(rownames(VarStat), RmvList), c(CatList, "SITE", "CENT"))

  CovariateList <- union(CatList, ContList)
  tabCovariate <- matrix(nrow = nID, ncol = length(CovariateList))
  colnames(tabCovariate) <- CovariateList
  nRec <- dim(FDATA)[1]

  if (length(CovariateList) > 0) {
    PrevID <- 0; j <- 1
    for (i in 1:nRec) {
      if (PrevID != FDATA[i, "ID"]) {
        for (k in 1:length(CovariateList)) {
          tabCovariate[j, CovariateList[k]] <- FDATA[i, CovariateList[k]]
        }
        j <- j + 1
      }
      PrevID <- FDATA[i, "ID"]
    }
  }

  tabEta <- cbind(tabEta, tabCovariate)
  VarEta <- names(tabEta)
  Cat1 <- intersect(VarEta, CatList)
  Cont1 <- intersect(VarEta, ContList)
  nCat <- length(Cat1)
  nCont <- length(Cont1)

  Cat2Rmv <- NULL
  if (nCat > 0) {
    for (i in 1:nEta) {
      for (j in 1:nCat) {
        if (length(unique(tabEta[, Cat1[j]])) == 1) Cat2Rmv <- union(Cat2Rmv, Cat1[j])
      }
    }
  }
  Cat1 <- setdiff(Cat1, Cat2Rmv)
  nCat <- length(Cat1)

  # Categorical covariates boxplots
  if (nCat > 0) {
    par(defpar)
    par(oma = c(1, 1, 3, 1), mar = c(4, 4, 2, 2), mfrow = c(nEta, nCat), lty = 1)
    for (i in 1:nEta) {
      for (j in 1:nCat) {
        form <- paste0("ETA", i, "~", Cat1[j])
        boxplot(formula = formula(form), data = tabEta, xlab = Cat1[j], ylab = paste("ETA", i))
        Anova.pvalue <- anova(lm(formula(form), data = tabEta))[[5]][1]
        if (is.finite(Anova.pvalue) & Anova.pvalue < 0.05) { txtcol <- "red" }
        else { txtcol <- "black" }
        mtext(side = 3, cex = 0.6, col = txtcol, paste("ANOVA p =", format(Anova.pvalue, digits = 4)))
      }
    }
    mtext(outer = TRUE, side = 3, "ETA vs Categorical Variables")
  }

  # Pair plots
  EBEpair(tabEta[, c("ID", Cont1, EtaNames2[-1])], RunNumber)

  # ETA vs continuous covariates grid
  IDdep <- rownames(VarStat[VarStat[, "DepID"] == 1, ])
  IDTimeMDVdep <- rownames(VarStat[VarStat[, "DepIDTimeMDV"] == 1, ])
  IDTimeEVIDdep <- rownames(VarStat[VarStat[, "DepIDTimeEVID"] == 1, ])
  OnlyIDTime0 <- union(IDTimeMDVdep, IDTimeEVIDdep)
  OnlyIDTime1 <- setdiff(OnlyIDTime0, IDdep)
  OnlyIDTime2 <- setdiff(OnlyIDTime1, ResWords)

  OUTLIER_PCT <- 0.025

  par(oma = c(1, 1, 3, 1), mar = c(0, 0, 0, 0), mfrow = c(nEta + 1, nCont + 1), lty = 1)
  for (i in 1:nEta) {
    for (j in 0:nCont) {
      if (j == 0) {
        plot(0, 0, type = "n", ylim = c(0, 1), xlim = c(0, 1), xaxt = "n", yaxt = "n",
             ylab = "", xlab = "", bty = "n")
        text(0, 0.5, paste0("ETA", i), adj = 0)
      } else {
        Xaxt <- ifelse(i < nEta, "n", "s")
        Xlab <- ifelse(i < nEta, "", Cont1[j])
        Yaxt <- ifelse(j == 1, "s", "n")
        Ylab <- ifelse(j == 1, paste("ETA", i), "")

        x <- tabEta[, Cont1[j]]
        minx <- min(x, na.rm = TRUE); maxx <- max(x, na.rm = TRUE)
        y <- tabEta[, paste0("ETA", i)]

        IDList <- tabEta[, "ID"]
        n <- length(x)
        ordx <- order(x); ordy <- order(y)

        OutSide <- union(union(ordx[1:max(1, round(OUTLIER_PCT * n))], ordx[round((1 - OUTLIER_PCT) * n):n]),
                         union(ordy[1:max(1, round(OUTLIER_PCT * n))], ordy[round((1 - OUTLIER_PCT) * n):n]))
        InSide <- intersect(ordx[max(2, round(OUTLIER_PCT * n)):min(round((1 - OUTLIER_PCT) * n), n - 1)],
                            ordy[max(2, round(OUTLIER_PCT * n)):min(round((1 - OUTLIER_PCT) * n), n - 1)])
        plot(x, y, type = "n", xlim = c(minx - 0.1 * (maxx - minx), maxx + 0.1 * (maxx - minx)),
             xaxt = Xaxt, xlab = Xlab, yaxt = Yaxt, ylab = Ylab)
        text(x[OutSide], y[OutSide], IDList[OutSide], cex = 0.8)
        points(x[InSide], y[InSide], cex = 0.5)

        if (length(intersect(Cont1[j], OnlyIDTime2)) == 1) {
          for (k in 1:nID) {
            cID <- IDs[k]
            x2 <- unique(FDATA[FDATA[, "ID"] == cID, Cont1[j]])
            nx2 <- length(x2)
            y2 <- rep(tabEta[tabEta[, "ID"] == cID, paste0("ETA", i)], nx2)
            lines(x2, y2, lty = 1)
            points(x2[-1], y2[-1], cex = 0.5, pch = 2)
          }
        }
        abline(h = 0, lty = 2)
        lines(lowess(x, y), col = "red", lty = 1)
      }
    }
  }

  plot(0, 0, type = "n", ylim = c(0, 1), xlim = c(0, 1), xaxt = "n", yaxt = "n",
       ylab = "", xlab = "", bty = "n")
  if (nCont > 0) {
    for (j in 1:nCont) {
      plot(0, 0, type = "n", ylim = c(0, 1), xlim = c(0, 1), xaxt = "n", yaxt = "n",
           ylab = "", xlab = "", bty = "n")
      text(0.5, 0.8, Cont1[j])
    }
  }

  tabEta
}


#' @keywords internal
.ebe_correlation_tables <- function(tabEta, nEta, EtaNames2, OM, VarStat, FDATA) {
  ResWords <- c("ID", "DV", "MDV", "EVID", "AMT", "RATE", "CMT", "II", "ADDL", "SS", "TIME", "DATE", "LNDV")
  RmvList <- union(ResWords, rownames(VarStat[VarStat[, "nUniq"] == 1, ]))
  ContList <- setdiff(setdiff(rownames(VarStat), RmvList), c(
    setdiff(rownames(VarStat[VarStat[, "ALLInt"] == 1 & VarStat[, "nUniq"] < 8 & VarStat[, "nUniq"] > 1, ]), RmvList),
    "SITE", "CENT"))
  Cont1 <- intersect(names(tabEta), ContList)

  options(width = 95)
  corEtaA <- cor(tabEta[, c(Cont1, EtaNames2[-1])])
  covEta2 <- cov(tabEta[, EtaNames2[-1], drop = FALSE])
  corEta2 <- cov2cor(covEta2)
  corOM <- cov2cor(OM)

  oAll <- list(corEtaA, covEta2, OM, covEta2 / OM, corEta2, corOM, corEta2 / corOM)
  names(oAll) <- c("Correlation of Covariates and EBE", "Covariance of EBE", "Omega Matrix",
                    "Ratios of Cov(EBE)/OM", "Correlation of EBE", "Correlation from Omega Matrix",
                    "Ratios of Cor(EBE)/(Cor from OM)")
  PrinMTxt(capture.output(oAll), Header1 = "Estimation vs EBE")
}


#' @keywords internal
.ebe_regression_diagnostics <- function(tabEta, nEta, EtaNames, EtaRep, VarStat, FDATA, defpar) {
  ResWords <- c("ID", "DV", "MDV", "EVID", "AMT", "RATE", "CMT", "II", "ADDL", "SS", "TIME", "DATE", "LNDV")
  RmvList <- union(ResWords, rownames(VarStat[VarStat[, "nUniq"] == 1, ]))
  CatList <- setdiff(rownames(VarStat[VarStat[, "ALLInt"] == 1 & VarStat[, "nUniq"] < 8 & VarStat[, "nUniq"] > 1, ]), RmvList)
  ContList <- setdiff(setdiff(rownames(VarStat), RmvList), c(CatList, "SITE", "CENT"))

  Cat1 <- intersect(names(tabEta), CatList)
  Cont1 <- intersect(names(tabEta), ContList)

  # Remove single-level categoricals
  Cat2Rmv <- NULL
  nCat <- length(Cat1)
  if (nCat > 0) {
    for (i in 1:nEta) {
      for (j in 1:nCat) {
        if (length(unique(tabEta[, Cat1[j]])) == 1) Cat2Rmv <- union(Cat2Rmv, Cat1[j])
      }
    }
  }
  Cat1 <- setdiff(Cat1, Cat2Rmv)

  tabEta2 <- RemoveNA(tabEta)

  EtaZero <- rep(FALSE, nEta)
  for (j in 1:nEta) {
    EtaZero[j] <- (all.equal(tabEta[, paste0("ETA", j)], rep(0, nrow(tabEta))) == TRUE)
  }

  CovarList <- c(Cat1, Cont1)

  ESQ_SSE_THRESHOLD <- 0.15
  DFFITS_THRESHOLD <- 1
  COVRATIO_MULTIPLIER <- 3

  if (length(CovarList) > 0 & !all(EtaZero)) {
    modelstr <- character()
    for (i in 1:length(CovarList)) {
      modelstr <- paste(modelstr, CovarList[i])
      if (i < length(CovarList)) modelstr <- paste(modelstr, "+")
    }

    result <- list()
    result2 <- list()

    for (j in 1:nEta) {
      if (!EtaZero[j]) {
        formstr <- paste0("ETA", j, "~", modelstr)
        result[j] <- list(summary(lm(formula(formstr), data = tabEta)))
        result2[j] <- list(mlr2(tabEta2[, paste0("ETA", j)], tabEta2[, CovarList]))
      } else {
        result[j] <- NA
        result2[j] <- NA
      }
    }

    names(result) <- EtaNames
    names(result2) <- EtaNames

    for (i in 1:nEta) {
      if (!EtaZero[i]) {
        sOut <- capture.output(result[[i]])
        PrinMTxt(sOut[4:length(sOut)], Header1 = paste("Multiple Linear Regression : ETA", i))
        sOut2 <- capture.output(result2[[i]])
        PrinMTxt(sOut2, Header1 = paste("Multiple Linear Regression - Influence : ETA", i))

        D <- result2[[i]][[2]][, "Cook's D"]
        y.hat <- result2[[i]][[2]][, "Yhat"]
        sdr <- result2[[i]][[2]][, "R-Student"]
        h <- result2[[i]][[2]][, "hat"]
        e <- result2[[i]][[2]][, "Residual"]
        SSE <- result2[[i]][[6]]
        COVRATIO <- result2[[i]][[2]][, "COV-Ratio"]
        DFFITS <- result2[[i]][[2]][, "DFFITS"]
        n <- result2[[i]][[3]]
        p <- result2[[i]][[4]]

        par(defpar)
        par(mfrow = c(2, 2), oma = c(1, 1, 3, 1))
        plot(D, type = "n", xlab = "Index", ylab = "Cook's Distance")
        for (jj in 1:n) {
          if (D[jj] == max(D)) text(jj, D[jj], jj)
          else points(jj, D[jj])
        }

        plot(y.hat, sdr, type = "n", xlab = "Predicted Value", ylab = "Studentized deleted residuals")
        for (jj in 1:n) {
          if (abs(sdr[jj]) > 2) text(y.hat[jj], sdr[jj], jj)
          else points(y.hat[jj], sdr[jj])
        }

        plot(h, e^2 / SSE, type = "n", xlab = "hat", ylab = "e^2/SSE")
        for (jj in 1:n) {
          if (e[jj]^2 / SSE > ESQ_SSE_THRESHOLD) text(h[jj], e[jj]^2 / SSE, jj)
          else points(h[jj], e[jj]^2 / SSE)
        }

        if (all(is.finite(COVRATIO))) {
          plot(COVRATIO, DFFITS, type = "n", xlab = "COVRATIO", ylab = "DFFITS")
          for (jj in 1:n) {
            if (abs(DFFITS[jj]) > DFFITS_THRESHOLD | abs(COVRATIO[jj] - 1) > COVRATIO_MULTIPLIER * p / n) {
              text(COVRATIO[jj], DFFITS[jj], jj)
            } else {
              points(COVRATIO[jj], DFFITS[jj])
            }
          }
        }

        mtext(paste("Influence Diagnostics on Eta", i), outer = TRUE, side = 3)
      }
    }
  }
}


#' @keywords internal
.ebe_individual_ci <- function(EtaRep, nEta, nID, OM) {
  LL <- matrix(nrow = nID, ncol = nEta)
  UL <- matrix(nrow = nID, ncol = nEta)
  ZERO <- matrix(nrow = nID, ncol = nEta)
  RSE <- matrix(nrow = nID, ncol = nEta)
  SHR <- matrix(nrow = nID, ncol = nEta)

  LLnames <- character(); ULnames <- character()
  ZEROnames <- character(); RSEnames <- character(); SHRnames <- character()
  for (i in 1:nEta) {
    LLnames <- c(LLnames, paste0("LL", i))
    ULnames <- c(ULnames, paste0("UL", i))
    ZEROnames <- c(ZEROnames, paste0("ZERO", i))
    RSEnames <- c(RSEnames, paste0("RSE", i))
    SHRnames <- c(SHRnames, paste0("SHR", i))
  }
  colnames(LL) <- LLnames; colnames(UL) <- ULnames
  colnames(ZERO) <- ZEROnames; colnames(RSE) <- RSEnames; colnames(SHR) <- SHRnames

  for (i in 1:nEta) {
    LL[, i] <- EtaRep[, i + 1] - 2 * EtaRep[, nEta + i + 1]
    UL[, i] <- EtaRep[, i + 1] + 2 * EtaRep[, nEta + i + 1]
    ZERO[, i] <- (LL[, i] * UL[, i] < 0)
    RSE[, i] <- EtaRep[, nEta + i + 1] / abs(EtaRep[, i + 1])
    SHR[, i] <- EtaRep[, nEta + i + 1] / sqrt(OM[i, i])
  }

  write.csv(cbind(EtaRep, LL, UL, ZERO, RSE, SHR), file = "EBEALL.CSV", row.names = FALSE)

  for (i in 1:nEta) {
    tabHeader <- c("ID", paste0("ETA", i), paste0("seETA", i), paste0("LL", i),
                   paste0("UL", i), paste0("ZERO", i), paste0("RSE", i), paste0("SHR", i))
    tabEtaOut <- cbind(EtaRep[, c(1, i + 1, i + nEta + 1)], LL[, i], UL[, i], ZERO[, i], RSE[, i], SHR[, i])
    colnames(tabEtaOut) <- tabHeader
    PrinMTxt(capture.output(tabEtaOut), Header1 = paste("ETA", i))
  }
}

#' Generate Individual PK Parameter Report (SC-IndiPK.PDF)
#'
#' Generates a PDF report with individual PK parameter distributions
#' including summary statistics, histograms, and QQ plots.
#'
#' @param run_dir character, path to the NONMEM run directory (default: current directory)
#' @export
nmw_report_indipk <- function(run_dir = getwd()) {
  owd <- setwd(run_dir)
  on.exit(setwd(owd))

  defpar <- par(no.readonly = TRUE)
  defpar$new <- NULL
  on.exit(par(defpar), add = TRUE)

  PrepPDF("SC-IndiPK.PDF")
  IndiPK <- unique(read.table("patab", skip = 1, header = TRUE))

  PKParas <- setdiff(colnames(IndiPK), "ID")
  nPKPara <- length(PKParas)

  if ("ETA1" %in% PKParas) IndiPK <- IndiPK[IndiPK$ETA1 != 0, ]

  RowPerPage <- 3
  for (i in 1:nPKPara) {
    if (i %% RowPerPage == 1) {
      par(oma = c(1, 1, 3, 1), mfrow = c(RowPerPage, 3), lty = 1)
    }
    var.data <- IndiPK[, PKParas[i]]

    PlotDistribution(var.data, PKParas[i], show_cv = TRUE)

    if (i %% RowPerPage == 0 | i == nPKPara) {
      mtext(outer = TRUE, side = 3, "PK Parameter distribution")
    }
  }

  PrinMTxt(capture.output(IndiPK), Cex = 0.8)
  ClosePDF()
  message("SC-IndiPK.PDF generated.")
}

TabStep = function()
{
# Uses results in environment
  THETA = e$FinalPara[1:e$nTheta]
  OM = ltv2mat(e$FinalPara[(e$nTheta+1):(e$nTheta + e$nEta*(e$nEta + 1)/2)])
  SG = diag(e$FinalPara[(e$nTheta + e$nEta*(e$nEta + 1)/2 + 1):(e$nTheta + e$nEta*(e$nEta + 1)/2 + e$nEps)])

  tSD = vector()
  for (i in 1:e$nID) {
    cID  = e$IDs[i]
    DATAi = e$DataRef[[i]]
    FGH0 = e$PRED(THETA, rep(0, e$nEta), DATAi )
    G0i  = FGH0[, e$GNames, drop=FALSE]
    H0i  = FGH0[, e$HNames, drop=FALSE]
    R0i  = DATAi[, "DV"] - FGH0[,"F"]
    C0i  = G0i %*% OM %*% t(G0i) + diag(diag(H0i %*% SG %*% t(H0i)))
    WRES = SqrtInvCov(C0i) %*% R0i

    if (e$METHOD != "ZERO") {
      cEBE = e$EBE[e$EBE[,"ID"]==cID, 2:(e$nEta + 1)]
      FGH1 = e$PRED(THETA, cEBE, DATAi)
      G1i  = FGH1[, e$GNames, drop=FALSE]
      H1i  = FGH1[, e$HNames, drop=FALSE]
      R1i  = DATAi[, "DV"] - FGH1[,"F"]
      if (e$INTER == FALSE) {
        C1i  = G1i %*% OM %*% t(G1i) + diag(diag(H0i %*% SG %*% t(H0i)))        
      } else {
        C1i  = G1i %*% OM %*% t(G1i) + diag(diag(H1i %*% SG %*% t(H1i)))
      }
      CWRES = SqrtInvCov(C1i) %*% (R1i + G1i %*% cEBE)
    }

    if (e$METHOD == "ZERO") {
      tSD = rbind(tSD, cbind(DATAi[, c("ID", "TIME", "DV")], PRED=FGH0[,"F"], RES=R0i, WRES, G0i, H0i))
    } else {
      tSD = rbind(tSD, cbind(DATAi[, c("ID", "TIME", "DV")], PRED=FGH0[,"F"], RES=R0i, WRES, CIPREDI=FGH1[,"F"], CIRESI=R1i, CWRES, G1i, H1i))
    }
  }

#  Result = list(THETA, OM, SG, tSD)
#  names(Result) = c("THETA", "OMEGA", "SIGMA", "sdtab")
  return(tSD)
}

#' Post-Hoc OFV and CWRES Calculation
#'
#' Calculates individual OFV and CWRES using METHOD=COND or METHOD=ZERO.
#'
#' @param ctlname character, control file name (without extension)
#' @param sdtab data.frame, standard output table
#' @return list with OFV, OFVi, CWRES
#' @keywords internal
OBJfpost <- function(ctlname, sdtab) {
  FCON <- readLines("FCON")
  iESTM <- 0
  for (i in 1:length(FCON)) {
    if (substring(FCON[i], 1, 4) == "ESTM") iESTM <- i
  }
  ESTM <- read.fwf("FCON", rep(4, 20), skip = iESTM - 1, nrow = 1)
  fMETH <- ESTM[1, 9] + 1

  TDAT <- sdtab
  TDAT <- TDAT[TDAT[, "MDV"] == 0, ]
  IDs <- unique(TDAT[, "ID"])

  OM <- as.matrix(read.table("fort.53"))
  NETA <- length(OM[1, ])
  OM <- OM[1:NETA, ]

  if (file.exists("fort.54")) {
    SG <- as.matrix(read.table("fort.54"))
    NEPS <- length(SG[1, ])
    SG <- SG[1:NEPS, ]
  } else {
    NEPS <- 1
    SG <- matrix(1)
  }

  etanames <- "ID"
  for (i in 1:NETA) {
    etanames <- c(etanames, paste("ETA", i, sep = ""))
  }

  if (fMETH == 0) {
    ETAr <- matrix(rep(0, NETA * length(IDs)), ncol = NETA)
    ETA <- cbind(IDs, ETAr)
    colnames(ETA) <- c(etanames)
  } else {
    ETAs <- read.table(paste(ctlname, ".ETA", sep = ""), header = TRUE, skip = 1)
    ETA <- as.matrix(ETAs[, etanames])
  }

  gnames <- vector()
  for (i in 1:NETA) gnames <- append(gnames, paste("G", i, "1", sep = ""))
  hnames <- vector()
  for (i in 1:NEPS) hnames <- append(hnames, paste("H", i, "1", sep = ""))

  CWRE <- vector()
  OFVi <- cbind(IDs, rep(NA, length(IDs)))

  DATcolnames <- colnames(TDAT)
  nDATcol <- length(DATcolnames)

  for (i in 1:length(IDs)) {
    DATi <- matrix(as.matrix(TDAT[TDAT[, "ID"] == IDs[i], ]), ncol = nDATcol)
    colnames(DATi) <- DATcolnames
    Yi <- DATi[, "DV"]
    F0i <- DATi[, "PRED"]
    F1i <- DATi[, "IPRE"]
    Gi <- matrix(as.matrix(DATi[, gnames]), ncol = NETA)
    Hi <- matrix(as.matrix(DATi[, hnames]), ncol = NEPS)
    RES0i <- Yi - F0i
    RES1i <- Yi - F1i
    COVi <- Gi %*% OM %*% t(Gi) + diag(diag(Hi %*% SG %*% t(Hi)), nrow = length(Yi))

    if (fMETH == 0) {
      WRESi <- SqrtInvCov(COVi) %*% RES0i
    } else {
      ETAi <- ETA[ETA[, "ID"] == IDs[i], 2:(NETA + 1)]
      WRESi <- SqrtInvCov(COVi) %*% (RES1i + Gi %*% ETAi)
    }
    OFVi[i, 2] <- determinant(COVi, logarithm = TRUE)$modulus[1] + t(WRESi) %*% WRESi
    CWRE <- append(CWRE, WRESi)
  }
  output <- list(sum(OFVi[, 2]), OFVi, cbind(TDAT, CWRE))
  names(output) <- list("OFV", "OFVi", "CWRES")
  return(output)
}


#' Check if Model Uses Log-Transformed DV (from Control File)
#'
#' @param CtlFileName character, path to NONMEM control file
#' @return logical, TRUE if LOG(F) is found after $PRED or $ERROR
#' @export
LogDV <- function(CtlFileName) {
  ModFile <- toupper(readLines(CtlFileName))

  iPRED <- grep("$PRED", ModFile, fixed = TRUE)
  iERROR <- grep("$ERROR", ModFile, fixed = TRUE)
  iLOGF <- grep("LOG(F)", ModFile, fixed = TRUE)

  if (length(iPRED) == 1) {
    if (iLOGF > iPRED) return(TRUE)
  }

  if (length(iERROR) == 1) {
    if (iLOGF > iERROR) return(TRUE)
  }
  return(FALSE)
}


#' Check if Model Uses Log-Transformed DV (from FSUBS)
#'
#' @return logical, TRUE if LOG(F) is found after SUBROUTINE PRED or ERROR in FSUBS
#' @export
LogDV2 <- function() {
  if (!file.exists("FSUBS")) return(FALSE)
  ModFile <- readLines("FSUBS")

  iPRED <- grep("SUBROUTINE PRED (", ModFile, fixed = TRUE)
  iERROR <- grep("SUBROUTINE ERROR (", ModFile, fixed = TRUE)
  iLOGF <- grep("LOG(F)", ModFile, fixed = TRUE)

  if (length(iLOGF) == 1) {
    if (length(iPRED) == 1) {
      if (iLOGF > iPRED) return(TRUE)
    }

    if (length(iERROR) == 1) {
      if (iLOGF > iERROR) return(TRUE)
    }
  }

  return(FALSE)
}

#' Read FDATA.CSV with Auto-Detection of Header Format
#'
#' Reads FDATA.CSV, auto-detecting whether it has a NONMEM-style
#' 2-row header (skip=2) or a standard 1-row header (skip=0).
#'
#' @param file character, path to the CSV file (default: "FDATA.CSV")
#' @return data.frame
#' @keywords internal
ReadFDATA <- function(file = "FDATA.CSV") {
  first_line <- readLines(file, n = 1)
  if (grepl("^[A-Za-z]", first_line)) {
    return(read.csv(file))
  } else {
    return(read.csv(file, skip = 2))
  }
}


#' Add Dose Number, Time of Latest Dose, and Time after Latest Dose
#'
#' Adds DoNo (Dosing Occasion Number), ToLD (Time of Latest Dose), and
#' TaLD (Time after Latest Dose) columns to a NONMEM dataset.
#'
#' @param NMData data.frame of NONMEM dataset
#' @param ID character, column name for subject ID
#' @param TIME character, column name for time
#' @param AMT character, column name for dose amount
#' @param II character, column name for interdose interval
#' @param ADDL character, column name for additional doses
#' @param MDV character, column name for missing dependent variable flag
#' @return data.frame with DoNo, ToLD, TaLD columns added
#' @export
AddDoNoTaLD <- function(NMData, ID = "ID", TIME = "TIME", AMT = "AMT",
                        II = "II", ADDL = "ADDL", MDV = "MDV") {
  nRec <- NROW(NMData)
  DoNo <- vector(length = nRec)
  ToLD <- vector(length = nRec)
  TaLD <- vector(length = nRec)

  LastID <- -1
  for (i in 1:nRec) {
    CurID <- NMData[i, ID]
    if (LastID != CurID) {
      pToLD <- 0
      cDoNo <- 0
    }

    if (length(intersect(names(NMData), ADDL)) == 1) {
      DosingHist <- NMData[NMData[, ID] == CurID & NMData[, AMT] > 0 & NMData[, MDV] == 1,
                           c(TIME, AMT, II, ADDL)]
      nDoseRec <- length(DosingHist[, AMT])
      for (j in 1:nDoseRec) {
        cADDL <- DosingHist[j, ADDL]
        if (cADDL > 0) {
          cAMT <- DosingHist[j, AMT]
          cTIME <- DosingHist[j, TIME]
          cII <- DosingHist[j, II]
          for (k in 1:cADDL) DosingHist <- rbind(DosingHist, c(cTIME + k * cII, cAMT, NA, NA))
        }
      }
      DosingHist2 <- DosingHist[order(DosingHist[, TIME]), c(TIME, AMT)]
    }

    if (NMData[i, AMT] > 0 & NMData[i, MDV] == 1) {
      cToLD <- NMData[i, TIME]
      cDoNo <- cDoNo + 1
    } else {
      if (length(intersect(names(NMData), ADDL)) == 1) {
        cTime <- NMData[i, TIME]
        cToLD <- max(DosingHist2[DosingHist2[, TIME] < cTime, TIME])
        cDoNo <- length(DosingHist2[DosingHist2[, TIME] < cTime, TIME])
      } else {
        cToLD <- pToLD
      }
    }
    DoNo[i] <- cDoNo
    ToLD[i] <- cToLD
    LastID <- CurID
    pToLD <- cToLD
  }

  TaLD <- NMData[, TIME] - ToLD
  return(cbind(NMData, DoNo, ToLD, TaLD))
}


#' Expand Dose History with ADDL/II Records
#'
#' Expands compressed dosing records (ADDL/II) into individual dose records.
#'
#' @param DoseHistTab data.frame with columns TIME, AMT, II, ADDL
#' @return data.frame with expanded dose records
#' @export
ExpandDoseHist <- function(DoseHistTab) {
  RetTab <- DoseHistTab
  nDoseRec <- dim(RetTab)[1]
  TempRec <- matrix(nrow = 1, ncol = dim(RetTab)[2])
  colnames(TempRec) <- colnames(RetTab)

  for (j in 1:nDoseRec) {
    cADDL <- RetTab[j, "ADDL"]
    if (cADDL > 0) {
      cTIME <- RetTab[j, "TIME"]
      cAMT <- RetTab[j, "AMT"]
      cII <- RetTab[j, "II"]
      for (k in 1:cADDL) {
        RetTab[j, "II"] <- 0
        RetTab[j, "ADDL"] <- 0
        TempRec <- RetTab[j, ]
        TempRec[1, "TIME"] <- cTIME + k * cII
        TempRec[1, "AMT"] <- cAMT
        RetTab <- rbind(RetTab, TempRec)
      }
    }
  }

  RetTab <- RetTab[order(RetTab[, "TIME"]), ]

  # Check too short interval
  IIs <- unique(DoseHistTab[DoseHistTab[, "II"] > 0, "II"])
  mII <- min(IIs, na.rm = TRUE)

  nRetRec <- dim(RetTab)[1]
  pTIME <- -1 * max(IIs, na.rm = TRUE)
  pAMT <- 0
  for (j in 1:nRetRec) {
    cTIME <- RetTab[j, "TIME"]
    cAMT <- RetTab[j, "AMT"]
    if ((cTIME - pTIME) < 0.5 * mII & cAMT > 0 & pAMT > 0) {
      message("Warning: Maybe too short interval")
      message(paste(capture.output(RetTab[j, ]), collapse = "\n"))
    }
    pTIME <- cTIME
    pAMT <- cAMT
  }

  return(RetTab)
}


#' Remove Rows with NA Values
#'
#' @param RawData data.frame
#' @return data.frame with rows containing NA removed
#' @export
RemoveNA <- function(RawData) {
  nRow <- nrow(RawData)
  Index <- vector(length = nRow)
  for (i in 1:nRow) {
    Index[i] <- !any(is.na(RawData[i, ]))
  }
  return(RawData[Index, ])
}


#' Remove Columns from Table
#'
#' @param Tab data.frame
#' @param OldCols character vector of column names to remove
#' @return data.frame with specified columns removed
#' @keywords internal
RmvCol <- function(Tab, OldCols) {
  return(Tab[, setdiff(colnames(Tab), OldCols)])
}


#' Rename Columns
#'
#' @param ColList character vector of column names
#' @param OldCol old column name
#' @param NewCol new column name
#' @return character vector with renamed columns
#' @keywords internal
RenCol <- function(ColList, OldCol, NewCol) {
  nCol <- length(ColList)
  for (i in 1:nCol) if (ColList[i] == OldCol) ColList[i] <- NewCol
  return(ColList)
}


#' Remove Fixed (Non-Varying) Columns
#'
#' Removes columns where all values are the same across rows.
#'
#' @param Table data.frame or matrix
#' @return data.frame with fixed columns removed
#' @keywords internal
RmvFixed <- function(Table) {
  ColNames <- colnames(Table)
  NotFixed <- rep(FALSE, length(ColNames))
  for (i in 1:dim(Table)[1]) {
    for (j in 1:length(ColNames)) {
      if (Table[i, ColNames[j]] != Table[1, ColNames[j]]) NotFixed[j] <- TRUE
    }
  }
  return(Table[, ColNames[NotFixed]])
}


#' Remove Zero Rows/Cols from Symmetric Matrix
#'
#' @param SymmMat symmetric matrix
#' @return matrix with zero rows/columns removed
#' @keywords internal
RmvZero <- function(SymmMat) {
  nRow <- dim(SymmMat)[1]
  NonZero <- rep(FALSE, nRow)

  for (i in 1:nRow) {
    for (j in 1:i) {
      if (SymmMat[i, j] != 0) NonZero[i] <- TRUE
    }
  }
  return(as.matrix(SymmMat[NonZero, NonZero]))
}

test_that("RemoveNA removes rows with NA", {
  df <- data.frame(a = c(1, NA, 3), b = c(4, 5, NA))
  result <- RemoveNA(df)
  expect_equal(nrow(result), 1)
  expect_equal(result$a, 1)
  expect_equal(result$b, 4)
})

test_that("RemoveNA keeps all rows when no NA", {
  df <- data.frame(a = c(1, 2, 3), b = c(4, 5, 6))
  result <- RemoveNA(df)
  expect_equal(nrow(result), 3)
})

test_that("RmvFixed removes constant columns", {
  m <- data.frame(a = c(1, 1, 1), b = c(1, 2, 3), c = c(5, 5, 5))
  result <- nmw:::RmvFixed(m)
  # When only one column remains, R drops to a vector
  expect_equal(as.numeric(result), c(1, 2, 3))
})

test_that("RmvFixed keeps multiple varying columns", {
  m <- data.frame(a = c(1, 1, 1), b = c(1, 2, 3), c = c(5, 6, 7))
  result <- nmw:::RmvFixed(m)
  expect_equal(ncol(result), 2)
  expect_true("b" %in% colnames(result))
  expect_true("c" %in% colnames(result))
})

test_that("RmvZero removes zero rows/cols from symmetric matrix", {
  m <- matrix(c(1, 0, 0.5, 0, 0, 0, 0.5, 0, 2), nrow = 3)
  result <- nmw:::RmvZero(m)
  expect_equal(nrow(result), 2)
  expect_equal(ncol(result), 2)
})

test_that("RmvZero handles all-nonzero matrix", {
  m <- matrix(c(1, 0.5, 0.5, 2), nrow = 2)
  result <- nmw:::RmvZero(m)
  expect_equal(dim(result), c(2, 2))
})

test_that("RmvCol removes specified columns", {
  df <- data.frame(a = 1:3, b = 4:6, c = 7:9)
  result <- nmw:::RmvCol(df, c("b"))
  expect_equal(ncol(result), 2)
  expect_false("b" %in% colnames(result))
})

test_that("RenCol renames specified column", {
  cols <- c("a", "b", "c")
  result <- nmw:::RenCol(cols, "b", "B_new")
  expect_equal(result, c("a", "B_new", "c"))
})

test_that("ExpandDoseHist expands ADDL records", {
  dose_tab <- data.frame(
    TIME = c(0),
    AMT = c(100),
    II = c(24),
    ADDL = c(2)
  )
  result <- ExpandDoseHist(dose_tab)
  expect_equal(nrow(result), 3)  # original + 2 ADDL
  expect_equal(sort(result$TIME), c(0, 24, 48))
})

test_that("MatchEnd matches string endings", {
  names <- c("file.csv", "data.txt", "report.csv", "notes.doc")
  result <- MatchEnd(names, ".csv")
  expect_equal(result, c(TRUE, FALSE, TRUE, FALSE))
})

test_that("MatchEnd is case insensitive by default", {
  names <- c("file.CSV", "data.txt")
  result <- MatchEnd(names, ".csv")
  expect_equal(result, c(TRUE, FALSE))
})

test_that("MatchEnd respects case when IgnoreCase=FALSE", {
  names <- c("file.CSV", "data.csv")
  result <- MatchEnd(names, ".csv", IgnoreCase = FALSE)
  expect_equal(result, c(FALSE, TRUE))
})

AddCox = function(nmData, coxData, coxCol, dateCol = "DATE", idCol = "ID")
{
  coxData = coxData[order(coxData[, idCol], coxData[, dateCol]), ]
  nmData[, coxCol] = NA
  nrNm = NROW(nmData)
  for (i in 1:nrNm) {
    cID = nmData[i, idCol]
    cDATE = nmData[i, dateCol] 
    cDAT = coxData[coxData[, idCol] == cID & coxData[, dateCol] <= cDATE, , drop=F]
    nrcDAT = NROW(cDAT)
    if (nrcDAT > 0) {
      nmData[i, coxCol] = cDAT[nrcDAT, coxCol]
    } else {
      cDAT = coxData[coxData[, idCol] == cID, , drop=F]
      if (NROW(cDAT) == 0) { nmData[i, coxCol] = NA
      } else { nmData[i, coxCol] = cDAT[1, coxCol] }
    }
  }
  
  if (sum(is.na(nmData[, coxCol])) > 0) warning("There is NA value in the result!")
  return(nmData)
}

#' Get Current Model Name from Working Directory
#'
#' Detects model name by finding the .xml file in the current directory.
#' Falls back to extracting from folder name if no .xml file is found.
#'
#' @return character, the model name
#' @keywords internal
GetCurModelName <- function() {
  xmlFiles <- list.files(pattern = "\\.xml$", ignore.case = TRUE)
  if (length(xmlFiles) >= 1) {
    return(sub("\\.xml$", "", xmlFiles[1], ignore.case = TRUE))
  }
  vPath <- strsplit(getwd(), "/")[[1]]
  n <- length(vPath)
  return(strsplit(vPath[n], "(\\.)")[[1]][1])
}


#' Get Objective Function Value
#'
#' Retrieves the final OFV from the .ext file in the current directory.
#'
#' @return numeric, the objective function value
#' @export
GetOFV <- function() {
  EXTName <- paste0(GetCurModelName(), ".ext")
  if (length(intersect(toupper(list.files()), toupper(EXTName))) == 1) {
    EXT <- read.table(EXTName, skip = 1, header = TRUE)
    if ("OBJ" %in% colnames(EXT)) {
      EXT0 <- EXT[EXT[, "ITERATION"] >= 0, ]
      nRow <- NROW(EXT0)
      if (nRow == 0) {
        return(EXT[1, "OBJ"])
      } else {
        return(tail(EXT0[, "OBJ"], 1))
      }
    }
  }
  return(NA)
}


#' Get Estimation Method
#'
#' Identifies the estimation method from the .ext file header.
#'
#' @return character, abbreviation of the estimation method
#' @export
GetEstMethod <- function() {
  EXTName <- paste0(GetCurModelName(), ".ext")
  if (length(intersect(toupper(list.files()), toupper(EXTName))) == 1) {
    EstString <- strsplit(readLines(EXTName, n = 1), ":")[[1]][2]
    EstMethod <- c("FO", "FOI", "FOCE", "FOCEI", "L", "LI",
                   "FO", "FOI", "FOCE", "FOCEI", "L", "LI")
    names(EstMethod) <- c(
      " First Order",
      " First Order with Interaction",
      " First Order Conditional Estimation",
      " First Order Conditional Estimation with Interaction",
      " Laplacian Conditional Estimation",
      " Laplacian Conditional Estimation with Interaction",
      " First Order (Evaluation)",
      " First Order with Interaction (Evaluation)",
      " First Order Conditional Estimation (Evaluation)",
      " First Order Conditional Estimation with Interaction (Evaluation)",
      " Laplacian Conditional Estimation (Evaluation)",
      " Laplacian Conditional Estimation with Interaction (Evaluation)")
    if (!is.na(EstString)) return(EstMethod[[EstString]])
  }
  return(NA)
}


#' Get PROBLEM Statement Value
#'
#' @param Tag character, the tag to search for in FCON
#' @return character, the extracted value
#' @keywords internal
GetProbVal <- function(Tag) {
  if (length(intersect(toupper(list.files()), "FCON")) == 1) {
    FCON <- readLines("FCON")
    PROB <- strsplit(trimws(substr(FCON[grep("PROB", FCON)], 9, 80)), " ")[[1]]
    Loc <- grep(Tag, PROB)
    if (Loc > 0) {
      return(substr(PROB[Loc], 3, nchar(PROB[Loc])))
    } else {
      return("")
    }
  } else {
    return(NA)
  }
}


#' Check for Tag in File
#'
#' @param FileName character, filename
#' @param Tag character, text to search for
#' @return logical or NA
#' @keywords internal
FileTag <- function(FileName, Tag) {
  if (length(intersect(toupper(list.files()), toupper(FileName))) == 1) {
    if (!is.na(pmatch(Tag, readLines(FileName)))) {
      return(TRUE)
    } else {
      return(FALSE)
    }
  } else {
    return(NA)
  }
}


#' Check Minimization Success
#'
#' @return logical, TRUE if minimization was successful
#' @export
MinSuccess <- function() {
  return(FileTag("PRINT.OUT", "MINIMIZATION SUCCESSFUL"))
}


#' Check Standard Error Success
#'
#' @return logical, TRUE if standard errors were computed successfully
#' @export
SESuccess <- function() {
  return(FileTag("PRINT.OUT", "********************                            STANDARD ERROR OF ESTIMATE"))
}


#' Get Count of Parameters
#'
#' @return integer, total number of estimated parameters, or NA if file not found
#' @export
GetCountPara <- function() {
  GRDName <- paste0(GetCurModelName(), ".grd")
  if (!file.exists(GRDName)) return(NA)
  GRD <- read.table(GRDName, skip = 1, header = TRUE)
  ncol(GRD) - 1
}


#' Get Count of All Thetas
#'
#' @return integer, total number of thetas, or NA if file not found
#' @keywords internal
GetCountAllTheta <- function() {
  EXT <- ReadEXTFile()
  if (is.null(EXT)) return(NA)
  length(grep("THETA", colnames(EXT)))
}


#' Get Count of Unfixed Thetas
#'
#' @return integer
#' @keywords internal
GetCountUnfixedTheta <- function() {
  return(GetCountPara() - GetCountOmega() - GetCountUnfixedEps())
}


#' Get Count of Fixed Thetas
#'
#' @return integer
#' @keywords internal
GetCountFixedTheta <- function() {
  return(GetCountAllTheta() - GetCountUnfixedTheta())
}


#' Get Count of Etas
#'
#' @return integer, number of eta parameters, or NA if file not found
#' @export
GetCountEta <- function() {
  PHIName <- paste0(GetCurModelName(), ".phi")
  if (!file.exists(PHIName)) return(NA)
  PHI <- read.table(PHIName, skip = 1, header = TRUE)
  length(grep("ETA", colnames(PHI)))
}


#' Get Count of Unfixed Omega Elements
#'
#' @return integer, or NA if file not found
#' @keywords internal
GetCountOmega <- function() {
  EXT <- ReadEXTFile()
  if (is.null(EXT)) return(NA)
  tOM <- EXT[EXT$ITERATION == -1000000000, grep("OMEGA", colnames(EXT))]
  sum(tOM != 0)
}


#' Get Off-Diagonal Omega Count
#'
#' @return integer
#' @keywords internal
GetOffDiagOmega <- function() {
  Parameter <- GetCountPara()
  Theta <- GetCountAllTheta()
  FixedTheta <- GetCountFixedTheta()
  nEta <- GetCountEta()
  return(Parameter - (Theta - FixedTheta) - nEta)
}


#' Get Count of Epsilons
#'
#' @return integer, number of epsilon parameters, or NA if file not found
#' @export
GetCountEps <- function() {
  EXT <- ReadEXTFile()
  if (is.null(EXT)) return(NA)
  length(grep("SIGMA", colnames(EXT)))
}


#' Get Count of Unfixed Epsilons
#'
#' @return integer, or NA if file not found
#' @keywords internal
GetCountUnfixedEps <- function() {
  EXT <- ReadEXTFile()
  if (is.null(EXT)) return(NA)
  if (!("ITERATION" %in% colnames(EXT))) return(NA)
  EXT <- EXT[EXT[, "ITERATION"] >= 0, ]
  nRow <- NROW(EXT)
  if (nRow == 0) return(0)
  ColNo <- grep("SIGMA", colnames(EXT))
  nCol <- length(ColNo)

  nUnfixed <- 0
  for (i in 1:nCol) {
    for (j in 2:nRow) {
      if (EXT[j, ColNo[i]] != EXT[j - 1, ColNo[i]]) {
        nUnfixed <- nUnfixed + 1
        break
      }
    }
  }
  nUnfixed
}


#' Get Count of Observations
#'
#' @return integer, total number of observation records
#' @export
GetCountObs <- function() {
  if (file.exists("PRINT.OUT")) {
    PRINTOUT <- readLines("PRINT.OUT")
    Tag <- "TOT. NO. OF OBS RECS:"
    LineNo <- grep(Tag, PRINTOUT)
    if (length(LineNo) > 0) {
      return(as.integer(trimws(substr(PRINTOUT[LineNo[1]], nchar(Tag) + 1, 80))))
    } else {
      return(NA)
    }
  } else {
    return(NA)
  }
}


#' Get Corrected AIC (AICc)
#'
#' @return numeric, the corrected AIC value
#' @export
GetAICc <- function() {
  n <- GetCountObs()
  p <- GetCountPara()
  ofv <- GetOFV()
  return(ofv + 2 * p + 2 * p * (p + 1) / (n - p - 1))
}

---
title: "nmw: NONMEM Post-Run Processing"
author: "Kyun-Seop Bae"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{nmw: NONMEM Post-Run Processing}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  eval = FALSE
)
```

## Overview

`nmw` package는 NONMEM 실행 결과에 대한 post-run processing을 수행합니다.
NONMEM run directory의 output 파일들을 읽어 총 8종의 진단 PDF 보고서를 자동 생성합니다.

### 생성되는 보고서

| 보고서 | 파일명 | 내용 |
|--------|--------|------|
| S1-OFV | S1-OFV.PDF | Objective Function Value 진단 |
| S2-Parameters | S2-Parameters.PDF | 모수 추정값 및 표준오차 |
| S3-Predictions | S3-Predictions.PDF | 예측 진단 (DV vs PRED, IPRE) |
| S4-Residuals | S4-Residuals.PDF | 잔차 진단 (WRES, CWRES, IWRES) |
| S5-EBE | S5-EBE.PDF | Empirical Bayes Estimate 진단 |
| SA-Input | SA-Input.PDF | 입력 데이터 요약 |
| SB-Output | SB-Output.PDF | NONMEM output 파일 내용 |
| SC-IndiPK | SC-IndiPK.PDF | 개체별 약동학 파라미터 |


## Installation

```{r install}
install.packages("path/to/nmw_0.1.0", repos = NULL, type = "source")
```

## Quick Start

### 전체 보고서 생성

```{r basic}
library(nmw)

# NONMEM run directory 경로를 지정하여 전체 보고서 생성
nmw_run("/path/to/nonmem/run/2006.R75")
```

`run_dir`에는 NONMEM 실행이 완료된 디렉토리 경로를 지정합니다.
해당 디렉토리에 `.xml`, `.ext`, `.grd`, `.phi`, `sdtab`, `patab`, `cotab`,
`FDATA.CSV`, `PRINT.OUT` 등의 파일이 존재해야 합니다.

### 선택적 보고서 생성

```{r selective}
# OFV와 파라미터 보고서만 생성
nmw_run("/path/to/run", reports = c("ofv", "param"))

# 잔차와 EBE 보고서만 생성
nmw_run("/path/to/run", reports = c("resid", "ebe"))
```

### 개별 보고서 함수

각 보고서를 개별적으로 생성할 수도 있습니다:

```{r individual}
# S1-OFV.PDF
nmw_report_ofv(run_dir = "/path/to/run")

# S2-Parameters.PDF
nmw_report_param(run_dir = "/path/to/run")

# S3-Predictions.PDF
nmw_report_pred(run_dir = "/path/to/run")

# S4-Residuals.PDF
nmw_report_resid(run_dir = "/path/to/run")

# S5-EBE.PDF
nmw_report_ebe(run_dir = "/path/to/run")

# SA-Input.PDF
nmw_report_input(run_dir = "/path/to/run")

# SB-Output.PDF
nmw_report_output(run_dir = "/path/to/run")

# SC-IndiPK.PDF
nmw_report_indipk(run_dir = "/path/to/run")
```


## 보고서 상세 내용

### S1-OFV: Objective Function Value 진단

- Total OFV, AICc, BIC
- Condition number
- 개체별 OFV/DV의 box plot, histogram, Q-Q plot
- Shapiro-Wilk 정규성 검정
- Gradient 분석

### S2-Parameters: 모수 추정값

- THETA, OMEGA, SIGMA 추정값 및 표준오차
- Wald-type significance test
- Unitary test
- OMEGA matrix의 plausibility 검토
- Correlation matrix

### S3-Predictions: 예측 진단

- Superposed fitting curve (DV, PRED, IPRE vs TaLD)
- DV vs PRED, DV vs IPRE scatter plot
- 개체별 fitting curve (DV, PRED, IPRE vs TIME)
- Log scale 포함
- Run test 결과 표시

### S4-Residuals: 잔차 진단

- WRES, IWRES, CWRES vs PRED, IPRE, TIME
- 절대 잔차 vs PRED, IPRE, TIME
- 잔차의 histogram 및 정규성 검정
- Epsilon shrinkage
- Mean percent residual
- Run test

### S5-EBE: Empirical Bayes Estimate 진단

- ETA의 histogram, Q-Q plot, 정규성 검정
- Population shrinkage
- ETA vs covariates (전체 및 개별 pair plot)
- Correlation matrix
- Multiple linear regression (Cook's Distance, DFFITS, COVRATIO, DFBETAS)
- 개체별 PK 파라미터의 histogram 및 요약 통계
- Individual shrinkage

### SA-Input: 입력 데이터 요약

- 투여 기록 없는 개체, DV 없는 개체 탐지
- 중복 레코드 검출
- 변수별 요약 통계
- 범주형/연속형 변수 구분

### SB-Output: NONMEM Output

- Control file 내용 (줄 번호 포함)
- NONMEM output 파일 내용

### SC-IndiPK: 개체별 약동학 파라미터

- 개체별 PK 파라미터 분포
- Histogram 및 요약 통계
- CV% 계산


## Utility Functions

패키지에는 보고서 외에도 유용한 유틸리티 함수들이 포함되어 있습니다.

### 데이터 처리

```{r utils}
# Dose Number, Time of Latest Dose, Time after Latest Dose 추가
FDATA <- read.csv("FDATA.CSV", skip = 2)
FDATA <- AddDoNoTaLD(FDATA)

# ADDL/II 레코드 확장
FDATA_expanded <- ExpandDoseHist(FDATA)

# NA 행 제거
FDATA_clean <- RemoveNA(FDATA, "DV")
```

### NONMEM 파일 파싱

```{r parse}
# XML에서 태그 값 추출
XML <- readLines("model.xml")
vals <- BtwTagVals("theta", XML)
mat <- BtwTagMat("omega", XML)

# .ext 파일에서 OFV 추출
setwd("/path/to/run")
ofv <- GetOFV()

# 추정 방법 확인
method <- GetEstMethod()

# AICc 계산
aicc <- GetAICc()

# 파라미터 수 확인
n_theta <- GetCountPara()
n_eta <- GetCountEta()
n_eps <- GetCountEps()
n_obs <- GetCountObs()
```

### 통계 검정

```{r tests}
# Run test
run.test.nm(residuals)

# 잔차 검정 (Shapiro-Wilk + Run test)
ResTest(residuals)

# Multiple linear regression with influence diagnostics
result <- mlr2(dat, y_col, x_cols)
```

### 모델 개발 요약

```{r summary}
# 여러 NONMEM run의 요약 비교
MDL <- SumOut(FileExt = ".CTL", RunExt = ".R75", OutExt = ".OUT")

# 모델 개발 흐름도 그리기
Outline(MDL, MDLName = "ProjectName")
```

### EBE Pair Plot

```{r ebepair}
# ETA vs Covariates pair plot
EBEpair(tabEta, RunNumber = "2006")

# 개별 covariate별 pair plot
EBEpair(tabEta, RunNumber = "2006", each = TRUE)
```

### PDF 보고서 직접 작성

```{r pdf}
# PDF 생성
PrepPDF("MyReport.PDF")

# 텍스트 출력
PrinTxt("Title Text", Cex = 1.2)

# 여러 줄 텍스트 출력
PrinMTxt(capture.output(summary(data)), Header1 = "Data Summary")

# 페이지 추가
AddPage()

# PDF 닫기
ClosePDF()
```


## 필요한 NONMEM Output 파일

`nmw_run()` 실행을 위해 run directory에 다음 파일들이 필요합니다:

| 파일 | 설명 | 사용 보고서 |
|------|------|------------|
| `{model}.xml` | NONMEM XML output | 전체 |
| `{model}.ext` | 추정값 파일 | S1, S2 |
| `{model}.grd` | Gradient 파일 | S1 |
| `{model}.phi` | 개체별 파라미터 파일 | S1, S5, SC |
| `sdtab` | Standard table (DV, PRED, IPRE, WRES 등) | S3, S4 |
| `patab` | Parameter table (ETA, 개체별 PK 파라미터) | S5, SC |
| `cotab` | Covariate table | S5 |
| `FDATA.CSV` | 입력 데이터 | SA |
| `PRINT.OUT` | NONMEM print output | SB |

### Run Directory 명명 규칙

Run directory 이름은 `{ModelName}.{Extension}` 형식이어야 합니다.
예: `2006.R75` (ModelName = "2006", Extension = "R75")

패키지는 디렉토리 이름에서 모델 이름을 추출하여 해당 `.xml`, `.ext` 등의
파일을 찾습니다.


## NONMEM Control File 규칙

이 패키지는 아래와 같은 NONMEM control file 작성 규칙을 따를 때 정상 동작합니다:

- `$PROBLEM` 행에 `P:` 태그로 parent model 지정 (예: `P:1001`)
- `$PROBLEM` 행에 `F:` 태그로 formula 지정 (예: `F:CL~WT`)
- `sdtab`, `patab`, `cotab` 파일명 사용
- `FDATA.CSV` 형식의 입력 데이터 (2행 skip)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetOffDiagOmega}
\alias{GetOffDiagOmega}
\title{Get Off-Diagonal Omega Count}
\usage{
GetOffDiagOmega()
}
\value{
integer
}
\description{
Get Off-Diagonal Omega Count
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_parse.R
\name{ParseItemWidth}
\alias{ParseItemWidth}
\title{Parse Fixed-Width Format Item Widths from NONMEM Output}
\usage{
ParseItemWidth(RawRead)
}
\arguments{
\item{RawRead}{character vector of lines from NONMEM output}
}
\value{
list with LineCount and Widths
}
\description{
Parse Fixed-Width Format Item Widths from NONMEM Output
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_math.R
\name{OBJfpost}
\alias{OBJfpost}
\title{Post-Hoc OFV and CWRES Calculation}
\usage{
OBJfpost(ctlname, sdtab)
}
\arguments{
\item{ctlname}{character, control file name (without extension)}

\item{sdtab}{data.frame, standard output table}
}
\value{
list with OFV, OFVi, CWRES
}
\description{
Calculates individual OFV and CWRES using METHOD=COND or METHOD=ZERO.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetCountFixedTheta}
\alias{GetCountFixedTheta}
\title{Get Count of Fixed Thetas}
\usage{
GetCountFixedTheta()
}
\value{
integer
}
\description{
Get Count of Fixed Thetas
}
\keyword{internal}

\name{Objs}
\alias{ObjEta}
\alias{Oi}
\alias{Oi0}
\alias{Oi1}
\alias{Obj}
\alias{OiS0}
\alias{OiS1}
\alias{CalcSmat}
\alias{PostHocEta}
\title{Internal Obj Functions}
\description{Internal Obj functions}
\details{These are not to be called by the user.}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{RmvFixed}
\alias{RmvFixed}
\title{Remove Fixed (Non-Varying) Columns}
\usage{
RmvFixed(Table)
}
\arguments{
\item{Table}{data.frame or matrix}
}
\value{
data.frame with fixed columns removed
}
\description{
Removes columns where all values are the same across rows.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdf_report.R
\name{DxPlotPost}
\alias{DxPlotPost}
\title{Diagnostic Plot for Post-Processing}
\usage{
DxPlotPost(x, y, mat, xlbl, ylbl, smooth, xlm = "", ylm = "", Log = "")
}
\arguments{
\item{x}{numeric vector, x-axis values}

\item{y}{numeric vector, y-axis values}

\item{mat}{data.frame with an ID column}

\item{xlbl}{character, x-axis label}

\item{ylbl}{character, y-axis label}

\item{smooth}{character, "T" for lowess smoothing, "F" for identity line}

\item{xlm}{numeric vector of length 2, x-axis limits}

\item{ylm}{numeric vector of length 2, y-axis limits}

\item{Log}{character, log transformation for axes (e.g., "y")}
}
\description{
Creates spaghetti-style diagnostic plots with individual ID labels.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_stats.R
\name{AllNA}
\alias{AllNA}
\title{Check if All Values are NA}
\usage{
AllNA(Column)
}
\arguments{
\item{Column}{vector}
}
\value{
logical
}
\description{
Check if All Values are NA
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_parse.R
\name{ParseItemName}
\alias{ParseItemName}
\title{Parse Data Item Names from NONMEM Output}
\usage{
ParseItemName(RawRead)
}
\arguments{
\item{RawRead}{character vector of lines from NONMEM output}
}
\value{
character vector of item names
}
\description{
Parse Data Item Names from NONMEM Output
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{code_race}
\alias{code_race}
\title{Code RACE as Integer}
\usage{
code_race(x)
}
\arguments{
\item{x}{character or numeric vector of race values.}
}
\value{
integer vector.
}
\description{
Code RACE as integer: ASIAN -> 1, WHITE -> 2, BLACK -> 3, otherwise -> 4.
Numeric input is returned unchanged (cast to integer).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{parse_dtc}
\alias{parse_dtc}
\title{Parse SDTM-Style DTC Strings into POSIXct}
\usage{
parse_dtc(x, tz = "UTC")
}
\arguments{
\item{x}{character vector (or POSIXct/Date) of datetime values.}

\item{tz}{time zone. Default "UTC".}
}
\value{
POSIXct vector.
}
\description{
Parse SDTM-style datetime strings into POSIXct.
Accepts "YYYY-MM-DDTHH:MM:SS", "YYYY-MM-DDTHH:MM",
"YYYY-MM-DD HH:MM:SS", "YYYY-MM-DD HH:MM", or "YYYY-MM-DD".
POSIXct / Date inputs are returned coerced to POSIXct.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_parse.R
\name{BtwTagMat}
\alias{BtwTagMat}
\title{Extract Matrix from NONMEM XML Output}
\usage{
BtwTagMat(Tag, RawRead, nRow)
}
\arguments{
\item{Tag}{character, the tag name (e.g., "omega", "sigma")}

\item{RawRead}{character vector of XML lines}

\item{nRow}{integer, dimension of the matrix}
}
\value{
numeric matrix
}
\description{
Parses NONMEM XML output to extract omega or sigma matrices.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetCountUnfixedTheta}
\alias{GetCountUnfixedTheta}
\title{Get Count of Unfixed Thetas}
\usage{
GetCountUnfixedTheta()
}
\value{
integer
}
\description{
Get Count of Unfixed Thetas
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_pairs.R
\name{PanelChar}
\alias{PanelChar}
\title{Character Panel for Pairs Plot}
\usage{
PanelChar(x, y)
}
\arguments{
\item{x}{numeric vector}

\item{y}{numeric vector}
}
\description{
Character Panel for Pairs Plot
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_summary.R
\name{SumOut}
\alias{SumOut}
\title{Summarize Multiple NONMEM Runs}
\usage{
SumOut(FileExt = ".CTL", RunExt = ".R75", OutExt = ".OUT")
}
\arguments{
\item{FileExt}{character, control file extension}

\item{RunExt}{character, run folder extension}

\item{OutExt}{character, output file extension}
}
\value{
data.frame with summary of all runs
}
\description{
Scans subdirectories matching RunExt and collects summary metrics
for model development flow analysis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetCountEta}
\alias{GetCountEta}
\title{Get Count of Etas}
\usage{
GetCountEta()
}
\value{
integer, number of eta parameters, or NA if file not found
}
\description{
Get Count of Etas
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{dat2_time_to_posix}
\alias{dat2_time_to_posix}
\title{Combine Date and Time Strings into POSIXct}
\usage{
dat2_time_to_posix(dat2, time, tz = "UTC")
}
\arguments{
\item{dat2}{character vector of dates in "YYYY-MM-DD" format.}

\item{time}{character vector of times in "HH:MM" format.}

\item{tz}{time zone. Default "UTC".}
}
\value{
POSIXct vector.
}
\description{
Combine "YYYY-MM-DD" and "HH:MM" character vectors into POSIXct.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_stats.R
\name{FuncDep}
\alias{FuncDep}
\title{Test Functional Dependency}
\usage{
FuncDep(NMTable, DetNameList, DepNameList)
}
\arguments{
\item{NMTable}{data.frame}

\item{DetNameList}{character vector of determinant column names}

\item{DepNameList}{character vector of dependent column names}
}
\value{
logical, TRUE if functionally dependent
}
\description{
Tests whether DepNameList columns are functionally dependent on
DetNameList columns in a NONMEM table.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_stats.R
\name{NMIDStat}
\alias{NMIDStat}
\title{NONMEM Individual (ID) Statistics}
\usage{
NMIDStat(NMTable)
}
\arguments{
\item{NMTable}{data.frame of NONMEM data}
}
\value{
list with IDStat data.frame, sortedID logical, sortedDT logical
}
\description{
Computes per-subject statistics including record counts,
DV counts, dosing information, and sorting checks.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{locf_value}
\alias{locf_value}
\title{LOCF / LOCB Lookup}
\usage{
locf_value(target_t, source_t, source_v, fallback = NA_real_)
}
\arguments{
\item{target_t}{vector of target times (numeric or POSIXct).}

\item{source_t}{vector of source times (same type as target_t).}

\item{source_v}{numeric vector of source values, parallel to source_t.}

\item{fallback}{value used when no source is available. Default NA.}
}
\value{
numeric vector, length(target_t).
}
\description{
For each target time, return the most-recent source value at
source_t <= target_t (LOCF). If no preceding source exists, return the
earliest source_v at source_t > target_t (LOCB). Empty source returns
\code{fallback} for all targets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nmw_run.R
\name{nmw_run}
\alias{nmw_run}
\title{Run All NONMEM Post-Processing Reports}
\usage{
nmw_run(
  run_dir = getwd(),
  reports = c("ofv", "param", "pred", "resid", "ebe", "input", "output", "indipk")
)
}
\arguments{
\item{run_dir}{character, path to the NONMEM run directory (default: current directory)}

\item{reports}{character vector, which reports to generate.
Options: "ofv", "param", "pred", "resid", "ebe", "input", "output", "indipk".
Default is all reports.}
}
\description{
Generates all diagnostic PDF reports for a NONMEM run in sequence:
S1-OFV, S2-Parameters, S3-Predictions, S4-Residuals, S5-EBE,
SA-Input, SB-Output, SC-IndiPK.
}
\examples{
\dontrun{
# Generate all reports
nmw_run("/path/to/nonmem/run/directory")

# Generate only OFV and parameter reports
nmw_run("/path/to/run", reports = c("ofv", "param"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_parse.R
\name{BtwTagVal}
\alias{BtwTagVal}
\title{Extract Value Between XML Tags}
\usage{
BtwTagVal(Tag1, Tag2, RawRead)
}
\arguments{
\item{Tag1}{opening tag}

\item{Tag2}{closing tag}

\item{RawRead}{character vector of lines}
}
\value{
character, the value between tags
}
\description{
Extract Value Between XML Tags
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_ofv.R
\name{nmw_report_ofv}
\alias{nmw_report_ofv}
\title{Generate OFV Diagnostic Report (S1-OFV.PDF)}
\usage{
nmw_report_ofv(run_dir = getwd())
}
\arguments{
\item{run_dir}{character, path to the NONMEM run directory (default: current directory)}
}
\description{
Generates a PDF report with objective function value diagnostics including
total OFV, AICc, BIC, individual OFV per DV summary, and gradient analysis.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{ExpandDoseHist}
\alias{ExpandDoseHist}
\title{Expand Dose History with ADDL/II Records}
\usage{
ExpandDoseHist(DoseHistTab)
}
\arguments{
\item{DoseHistTab}{data.frame with columns TIME, AMT, II, ADDL}
}
\value{
data.frame with expanded dose records
}
\description{
Expands compressed dosing records (ADDL/II) into individual dose records.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{merge_cov_locf}
\alias{merge_cov_locf}
\title{Merge Time-Varying Covariates by LOCF}
\usage{
merge_cov_locf(
  nm,
  cov_df,
  value_cols,
  time_col,
  median_fb = TRUE,
  verbose = FALSE,
  subj_col = "SUBJID",
  dat_col = "DAT2",
  tim_col = "TIME"
)
}
\arguments{
\item{nm}{data.frame with key columns \code{subj_col}, \code{dat_col},
\code{tim_col}.}

\item{cov_df}{data.frame with \code{subj_col}, \code{time_col}, and one or
more value columns.}

\item{value_cols}{character vector of value columns in \code{cov_df} to
bring into \code{nm}.}

\item{time_col}{name of the time column in \code{cov_df}, parsed by
\code{\link{parse_dtc}}.}

\item{median_fb}{logical. If TRUE (default), fill remaining NA values with
the column-wise median of \code{cov_df}.}

\item{verbose}{logical. Print median fill counts. Default FALSE.}

\item{subj_col}{subject ID column name. Default "SUBJID".}

\item{dat_col}{date column name in \code{nm}. Default "DAT2".}

\item{tim_col}{time column name in \code{nm}. Default "TIME".}
}
\value{
\code{nm} with \code{value_cols} added.
}
\description{
Merge time-varying covariate columns into a NM-style data.frame using LOCF
(with LOCB for records before the first source). Subjects/records still NA
after LOCF/LOCB can be filled with the column-wise median of \code{cov_df}.
}

\name{InitStep}
\alias{InitStep}
\title{Initialization Step}
\description{
It receives parameters for the estimation and stores them into \code{e} environment.
}
\usage{
InitStep(DataAll, THETAinit, OMinit, SGinit, LB, UB, Pred, METHOD)
}
\arguments{
  \item{DataAll}{Data for all subjects. It should contain columns which \code{Pred} function uses.}
  \item{THETAinit}{Theta initial values}
  \item{OMinit}{Omega matrix initial values}
  \item{SGinit}{Sigma matrix initial values}
  \item{LB}{Lower bounds for theta vector}
  \item{UB}{Upper bounds for theta vector}
  \item{Pred}{Prediction function name}
  \item{METHOD}{one of the estimation methods \code{"ZERO"}, \code{"COND"}, or \code{"LAPL"}}
}
\details{
Prediction function should return not only prediction values(F or IPRED) but also G (first derivative with respect to etas) and H (first derivative of Y with respect to epsilon).
For the \code{"LAPL"}, prediction function should return second derivative with respect to eta also.
"INTERACTION" is \code{TRUE} for \code{"COND"} and \code{"LAPL"} option, and \code{FALSE} for \code{"ZERO"}.
Omega matrix should be full block one.
Sigma matrix should be diagonal one.
}
\value{
This does not return values, but stores necessary values into the environment \code{e}.
}
\references{
NONMEM Users Guide
}
\author{Kyun-Seop Bae <k@acr.kr>}
\examples{
DataAll = Theoph
colnames(DataAll) = c("ID", "BWT", "DOSE", "TIME", "DV")
DataAll[,"ID"] = as.numeric(as.character(DataAll[,"ID"]))

nTheta = 3
nEta = 3
nEps = 2

THETAinit = c(2, 50, 0.1) # Initial estimate
OMinit = matrix(c(0.2, 0.1, 0.1, 0.1, 0.2, 0.1, 0.1, 0.1, 0.2), nrow=nEta, ncol=nEta)
OMinit
SGinit = diag(c(0.1, 0.1))
SGinit

LB = rep(0, nTheta) # Lower bound
UB = rep(1000000, nTheta) # Upper bound

FGD = deriv(~DOSE/(TH2*exp(ETA2))*TH1*exp(ETA1)/(TH1*exp(ETA1) - TH3*exp(ETA3))*
             (exp(-TH3*exp(ETA3)*TIME)-exp(-TH1*exp(ETA1)*TIME)),
            c("ETA1","ETA2","ETA3"),
            function.arg=c("TH1", "TH2", "TH3", "ETA1", "ETA2", "ETA3", "DOSE", "TIME"),
            func=TRUE, hessian=TRUE)
H = deriv(~F + F*EPS1 + EPS2, c("EPS1", "EPS2"), function.arg=c("F", "EPS1", "EPS2"), func=TRUE)

PRED = function(THETA, ETA, DATAi)
{
  FGDres = FGD(THETA[1], THETA[2], THETA[3], ETA[1], ETA[2], ETA[3], DOSE=320, DATAi[,"TIME"])
  Gres = attr(FGDres, "gradient")
  Hres = attr(H(FGDres, 0, 0), "gradient")

  if (e$METHOD == "LAPL") {
    Dres = attr(FGDres, "hessian")
    Res = cbind(FGDres, Gres, Hres, Dres[,1,1], Dres[,2,1], Dres[,2,2], Dres[,3,])
    colnames(Res) = c("F", "G1", "G2", "G3", "H1", "H2", "D11", "D21", "D22", "D31", "D32", "D33")
  } else {
    Res = cbind(FGDres, Gres, Hres)
    colnames(Res) = c("F", "G1", "G2", "G3", "H1", "H2")
  }
  return(Res)
}

######### First Order Approximation Method
InitStep(DataAll, THETAinit=THETAinit, OMinit=OMinit, SGinit=SGinit, LB=LB, UB=UB,
         Pred=PRED, METHOD="ZERO")

######### First Order Conditional Estimation with Interaction Method
InitStep(DataAll, THETAinit=THETAinit, OMinit=OMinit, SGinit=SGinit, LB=LB, UB=UB,
         Pred=PRED, METHOD="COND")

######### Laplacian Approximation with Interaction Method
InitStep(DataAll, THETAinit=THETAinit, OMinit=OMinit, SGinit=SGinit, LB=LB, UB=UB,
         Pred=PRED, METHOD="LAPL")
}
\keyword{Initialization Step}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_tests.R
\name{run.p}
\alias{run.p}
\title{Run Test Probability}
\usage{
run.p(m, n, r)
}
\arguments{
\item{m}{count of fewer species (minimum value = 0)}

\item{n}{count of more frequent species (minimum value = 1)}

\item{r}{count of runs (minimum value = 1)}
}
\value{
numeric, probability
}
\description{
Calculates P(Run count <= r | m, n) for the runs test.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_report.R
\name{CountEXTParams}
\alias{CountEXTParams}
\title{Count THETA, ETA, EPS Parameters from EXT Data Frame}
\usage{
CountEXTParams(EXT)
}
\arguments{
\item{EXT}{data.frame from .ext file (already read)}
}
\value{
list with nTheta, nEta, nEps
}
\description{
Count THETA, ETA, EPS Parameters from EXT Data Frame
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_tests.R
\name{run.test.nm}
\alias{run.test.nm}
\title{Runs Test for NONMEM Residuals}
\usage{
run.test.nm(RES)
}
\arguments{
\item{RES}{numeric vector of residuals}
}
\value{
numeric, two-sided p-value
}
\description{
Performs a runs test on residuals. Zeros are omitted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_summary.R
\name{GetReason}
\alias{GetReason}
\title{Get Termination Reason from PRINT.OUT}
\usage{
GetReason(folderName)
}
\arguments{
\item{folderName}{character, optional folder to set as working directory}
}
\value{
character, termination reason code
}
\description{
Get Termination Reason from PRINT.OUT
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetOFV}
\alias{GetOFV}
\title{Get Objective Function Value}
\usage{
GetOFV()
}
\value{
numeric, the objective function value
}
\description{
Retrieves the final OFV from the .ext file in the current directory.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{RmvZero}
\alias{RmvZero}
\title{Remove Zero Rows/Cols from Symmetric Matrix}
\usage{
RmvZero(SymmMat)
}
\arguments{
\item{SymmMat}{symmetric matrix}
}
\value{
matrix with zero rows/columns removed
}
\description{
Remove Zero Rows/Cols from Symmetric Matrix
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_pred.R
\name{nmw_report_pred}
\alias{nmw_report_pred}
\title{Generate Prediction Diagnostics Report (S3-Predictions.PDF)}
\usage{
nmw_report_pred(run_dir = getwd())
}
\arguments{
\item{run_dir}{character, path to the NONMEM run directory (default: current directory)}
}
\description{
Generates a PDF report with spaghetti plots, DV vs PRED/IPRE,
and individual fitting curves.
}

\name{TrimOut}
\alias{TrimOut}
\title{Trimming and beautifying NONMEM original OUTPUT file}
\description{
TrimOut removes unnecessary parts from NONMEM original OUTPUT file.
}
\usage{
  TrimOut(inFile, outFile="PRINT.OUT")
}
\arguments{
  \item{inFile}{NONMEM original untidy OUTPUT file name}
  \item{outFile}{Output file name to be written}
}

\details{
NONMEM original OUTPUT file contains unnecessary parts such as CONTROL file content, Start/End Time, License Info, Print control characters such as "+", "0", "1". This function trims those.
}
\value{
  outFile will be written in the current working folder or designated folder.
  This returns TRUE if the process was smooth.
}
\author{Kyun-Seop Bae <k@acr.kr>}
\keyword{NONMEM OUTPUT}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_resid.R
\name{nmw_report_resid}
\alias{nmw_report_resid}
\title{Generate Residual Diagnostics Report (S4-Residuals.PDF)}
\usage{
nmw_report_resid(run_dir = getwd())
}
\arguments{
\item{run_dir}{character, path to the NONMEM run directory (default: current directory)}
}
\description{
Generates a PDF report with WRES, CWRES, IWRES diagnostics including
scatter plots, histograms, normality tests, binomial count tests,
and individual residual curves with run tests.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_report.R
\name{CalcTaLDForReport}
\alias{CalcTaLDForReport}
\title{Calculate TaLD (Time after Latest Dose) for Report Generation}
\usage{
CalcTaLDForReport(FDATA, nRec)
}
\arguments{
\item{FDATA}{data.frame of NONMEM data}

\item{nRec}{integer, number of records}
}
\value{
list with ToLD, TaLD, DOCC, NewID vectors
}
\description{
Shared logic between report_pred.R and report_resid.R.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_report.R
\docType{data}
\name{OFV_SCREEN_LAYOUT}
\alias{OFV_SCREEN_LAYOUT}
\title{OFV Report Split Screen Layout}
\format{
An object of class \code{matrix} (inherits from \code{array}) with 4 rows and 4 columns.
}
\usage{
OFV_SCREEN_LAYOUT
}
\description{
Layout matrix for the S1-OFV report's split screen display.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_summary.R
\name{Outline}
\alias{Outline}
\title{Draw Model Development Flow Diagram}
\usage{
Outline(MDL, MDLName, out.start = "0", out.end = "zzzzzzzz", Target = "PDF")
}
\arguments{
\item{MDL}{data.frame from SumOut}

\item{MDLName}{character, model development name for title}

\item{out.start}{character, filter start}

\item{out.end}{character, filter end}

\item{Target}{character, "PDF" or other}
}
\description{
Creates a graphical model development flow diagram from SumOut results.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdf_report.R
\name{PrinMTxt}
\alias{PrinMTxt}
\title{Print Multiple Lines of Text}
\usage{
PrinMTxt(
  MTxt,
  Cex = 0.8,
  Header1 = "",
  Header2 = "",
  Header3 = "",
  Footer1 = "",
  Footer2 = "",
  Footer3 = "",
  PrintRowNum = FALSE
)
}
\arguments{
\item{MTxt}{character vector of text lines}

\item{Cex}{numeric, character expansion factor}

\item{Header1}{character, left header}

\item{Header2}{character, center header}

\item{Header3}{character, right header}

\item{Footer1}{character, left footer}

\item{Footer2}{character, center footer}

\item{Footer3}{character, right footer}

\item{PrintRowNum}{logical, whether to print row numbers}
}
\description{
Prints a character vector of text lines across multiple pages as needed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_stats.R
\name{AllSame}
\alias{AllSame}
\title{Check if All Values are the Same}
\usage{
AllSame(Column)
}
\arguments{
\item{Column}{vector}
}
\value{
logical
}
\description{
Check if All Values are the Same
}
\keyword{internal}

\name{nmw-package}
\alias{nmw-package}
\alias{nmw}
\docType{package}
\title{Understanding Nonlinear Mixed Effects Modeling for Population Pharmacokinetics}
\description{This shows how NONMEM(R) <http://www.iconplc.com/innovation/nonmem/> software works. Additionally, provides functions for post-run processing of NONMEM output files, generating comprehensive PDF diagnostic reports.}
\details{
This package explains 'First Order(FO) approximation' method, 'First Order Conditional Estimation(FOCE)' method, and 'Laplacian(LAPL)' method of NONMEM software. It also provides post-run processing functions including PDF diagnostic reports for OFV analysis, parameter estimates, prediction diagnostics, residual diagnostics, EBE analysis, input data summary, and individual PK parameter distributions.
}
\author{Kyun-Seop Bae <k@acr.kr>}
\references{
\enumerate{  
  \item NONMEM Users guide
  \item Wang Y. Derivation of various NONMEM estimation methods. J Pharmacokinet Pharmacodyn. 2007.
  \item Kang D, Bae K, Houk BE, Savic RM, Karlsson MO. Standard Error of Empirical Bayes Estimate in NONMEM(R) VI. K J Physiol Pharmacol. 2012.
  \item Kim M, Yim D, Bae K. R-based reproduction of the estimation process hidden behind NONMEM Part 1: First order approximation method. 2015.
  \item Bae K, Yim D. R-based reproduction of the estimation process hidden behind NONMEM Part 2: First order conditional estimation. 2016.
}
}
\keyword{Population Pharmacokinetics}
\keyword{Nonlinear Mixed Effects Modeling}
\examples{
DataAll = Theoph
colnames(DataAll) = c("ID", "BWT", "DOSE", "TIME", "DV")
DataAll[,"ID"] = as.numeric(as.character(DataAll[,"ID"]))

nTheta = 3
nEta = 3
nEps = 2

THETAinit = c(2, 50, 0.1)
OMinit = matrix(c(0.2, 0.1, 0.1, 0.1, 0.2, 0.1, 0.1, 0.1, 0.2), nrow=nEta, ncol=nEta)
SGinit = diag(c(0.1, 0.1))

LB = rep(0, nTheta) # Lower bound
UB = rep(1000000, nTheta) # Upper bound

FGD = deriv(~DOSE/(TH2*exp(ETA2))*TH1*exp(ETA1)/(TH1*exp(ETA1) - TH3*exp(ETA3))*
             (exp(-TH3*exp(ETA3)*TIME)-exp(-TH1*exp(ETA1)*TIME)),
            c("ETA1","ETA2","ETA3"),
            function.arg=c("TH1", "TH2", "TH3", "ETA1", "ETA2", "ETA3", "DOSE", "TIME"),
            func=TRUE, hessian=TRUE)
H = deriv(~F + F*EPS1 + EPS2, c("EPS1", "EPS2"), function.arg=c("F", "EPS1", "EPS2"), func=TRUE)

PRED = function(THETA, ETA, DATAi)
{
  FGDres = FGD(THETA[1], THETA[2], THETA[3], ETA[1], ETA[2], ETA[3], DOSE=320, DATAi[,"TIME"]) 
  Gres = attr(FGDres, "gradient")
  Hres = attr(H(FGDres, 0, 0), "gradient")
  
  if (e$METHOD == "LAPL") {
    Dres = attr(FGDres, "hessian")
    Res = cbind(FGDres, Gres, Hres, Dres[,1,1], Dres[,2,1], Dres[,2,2], Dres[,3,])
    colnames(Res) = c("F", "G1", "G2", "G3", "H1", "H2", "D11", "D21", "D22", "D31", "D32", "D33") 
  } else {
    Res = cbind(FGDres, Gres, Hres)
    colnames(Res) = c("F", "G1", "G2", "G3", "H1", "H2") 
  }
  return(Res)
}

####### First Order Approximation Method # Commented out for the CRAN CPU time
#InitStep(DataAll, THETAinit=THETAinit, OMinit=OMinit, SGinit=SGinit, LB=LB, UB=UB, 
#         Pred=PRED, METHOD="ZERO")
#(EstRes = EstStep())           # 4 sec
#(CovRes = CovStep())           # 2 sec
#PostHocEta() # Using e$FinalPara from EstStep()
#TabStep()    

######## First Order Conditional Estimation with Interaction Method
#InitStep(DataAll, THETAinit=THETAinit, OMinit=OMinit, SGinit=SGinit, LB=LB, UB=UB, 
#         Pred=PRED, METHOD="COND")
#(EstRes = EstStep())           # 2 min
#(CovRes = CovStep())           # 1 min
#get("EBE", envir=e)
#TabStep()

######## Laplacian Approximation with Interaction Method
#InitStep(DataAll, THETAinit=THETAinit, OMinit=OMinit, SGinit=SGinit, LB=LB, UB=UB, 
#         Pred=PRED, METHOD="LAPL")
#(EstRes = EstStep())           # 4 min
#(CovRes = CovStep())           # 1 min
#get("EBE", envir=e)
#TabStep()
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{RemoveNA}
\alias{RemoveNA}
\title{Remove Rows with NA Values}
\usage{
RemoveNA(RawData)
}
\arguments{
\item{RawData}{data.frame}
}
\value{
data.frame with rows containing NA removed
}
\description{
Remove Rows with NA Values
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_pairs.R
\name{PanelHist}
\alias{PanelHist}
\title{Histogram Panel for Pairs Plot}
\usage{
PanelHist(x, ...)
}
\arguments{
\item{x}{numeric vector}

\item{...}{additional arguments}
}
\description{
Histogram Panel for Pairs Plot
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_ebe.R
\name{nmw_report_ebe}
\alias{nmw_report_ebe}
\title{Generate EBE Diagnostics Report (S5-EBE.PDF)}
\usage{
nmw_report_ebe(run_dir = getwd())
}
\arguments{
\item{run_dir}{character, path to the NONMEM run directory (default: current directory)}
}
\description{
Generates a PDF report with empirical Bayes estimate (EBE) analysis
including ETA distributions, shrinkage, covariate relationships,
correlations, and multiple linear regression with influence diagnostics.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdf_report.R
\name{PrinTxt}
\alias{PrinTxt}
\title{Print Text at Position}
\usage{
PrinTxt(Row, Col, Text, Cex = 0.8)
}
\arguments{
\item{Row}{numeric, row position}

\item{Col}{numeric, column position}

\item{Text}{character, text to print}

\item{Cex}{numeric, character expansion factor}
}
\description{
Prints text at a specific row and column position on the PDF page.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdf_report.R
\name{PrepPDF}
\alias{PrepPDF}
\title{Prepare PDF Output}
\usage{
PrepPDF(FileName, Paper = "letter", FontFamily = "Courier")
}
\arguments{
\item{FileName}{character, output PDF filename}

\item{Paper}{character, paper size (default "letter")}

\item{FontFamily}{character, font family (default "Courier")}
}
\description{
Opens a PDF device for NONMEM report generation.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{RmvCol}
\alias{RmvCol}
\title{Remove Columns from Table}
\usage{
RmvCol(Tab, OldCols)
}
\arguments{
\item{Tab}{data.frame}

\item{OldCols}{character vector of column names to remove}
}
\value{
data.frame with specified columns removed
}
\description{
Remove Columns from Table
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{build_obs_records}
\alias{build_obs_records}
\title{Build NONMEM Observation Records}
\usage{
build_obs_records(PC)
}
\arguments{
\item{PC}{data.frame with observation information.}
}
\value{
data.frame of observation records with columns
  SUBJID, DAT2, TIME, AMT, RATE, CMT, DV, MDV.
}
\description{
Build NONMEM observation records from a PC-like data.frame.
PC must contain: SUBJID, DAT2, TIME, DV, CMT.
Output rows have AMT = 0, RATE = 0; DV that is NA or 0 -> MDV = 1.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetCurModelName}
\alias{GetCurModelName}
\title{Get Current Model Name from Working Directory}
\usage{
GetCurModelName()
}
\value{
character, the model name
}
\description{
Detects model name by finding the .xml file in the current directory.
Falls back to extracting from folder name if no .xml file is found.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{ReadFDATA}
\alias{ReadFDATA}
\title{Read FDATA.CSV with Auto-Detection of Header Format}
\usage{
ReadFDATA(file = "FDATA.CSV")
}
\arguments{
\item{file}{character, path to the CSV file (default: "FDATA.CSV")}
}
\value{
data.frame
}
\description{
Reads FDATA.CSV, auto-detecting whether it has a NONMEM-style
2-row header (skip=2) or a standard 1-row header (skip=0).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{AddDoNoTaLD}
\alias{AddDoNoTaLD}
\title{Add Dose Number, Time of Latest Dose, and Time after Latest Dose}
\usage{
AddDoNoTaLD(
  NMData,
  ID = "ID",
  TIME = "TIME",
  AMT = "AMT",
  II = "II",
  ADDL = "ADDL",
  MDV = "MDV"
)
}
\arguments{
\item{NMData}{data.frame of NONMEM dataset}

\item{ID}{character, column name for subject ID}

\item{TIME}{character, column name for time}

\item{AMT}{character, column name for dose amount}

\item{II}{character, column name for interdose interval}

\item{ADDL}{character, column name for additional doses}

\item{MDV}{character, column name for missing dependent variable flag}
}
\value{
data.frame with DoNo, ToLD, TaLD columns added
}
\description{
Adds DoNo (Dosing Occasion Number), ToLD (Time of Latest Dose), and
TaLD (Time after Latest Dose) columns to a NONMEM dataset.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetCountAllTheta}
\alias{GetCountAllTheta}
\title{Get Count of All Thetas}
\usage{
GetCountAllTheta()
}
\value{
integer, total number of thetas, or NA if file not found
}
\description{
Get Count of All Thetas
}
\keyword{internal}

\name{MinUtil}
\alias{SqrtInvCov}
\alias{Mx}
\alias{mat2ltv}
\alias{ltv2mat}
\alias{mat2utv}
\alias{utv2mat}
\alias{ScaleVar}
\alias{DesclVar}
\alias{DECN}
\title{Internal Min Util Functions}
\description{Internal Min Util functions}
\details{These are not to be called by the user.}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_indipk.R
\name{nmw_report_indipk}
\alias{nmw_report_indipk}
\title{Generate Individual PK Parameter Report (SC-IndiPK.PDF)}
\usage{
nmw_report_indipk(run_dir = getwd())
}
\arguments{
\item{run_dir}{character, path to the NONMEM run directory (default: current directory)}
}
\description{
Generates a PDF report with individual PK parameter distributions
including summary statistics, histograms, and QQ plots.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_summary.R
\name{ConnPoint}
\alias{ConnPoint}
\title{Connect Two Points with Lines}
\usage{
ConnPoint(pt1, pt2)
}
\arguments{
\item{pt1}{numeric vector of length 2, first point (x, y)}

\item{pt2}{numeric vector of length 2, second point (x, y)}
}
\description{
Draws connecting lines between parent and child nodes in flow diagram.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_tests.R
\name{ResTest}
\alias{ResTest}
\title{Residual Count Test Using Binomial Distribution}
\usage{
ResTest(
  ResTab,
  TestCols = c("WRES", "CWRE", "IWRE"),
  ZVals = c(0, 1, 2, 3),
  PctVals = c(0.005, 0.01, 0.025, 0.05, 0.1, 0.2, 0.3, 0.4, 0.5)
)
}
\arguments{
\item{ResTab}{data.frame with residual columns}

\item{TestCols}{character vector of column names to test}

\item{ZVals}{numeric vector of Z-value thresholds}

\item{PctVals}{numeric vector of percentile values}
}
\value{
matrix of test results
}
\description{
Tests whether residual counts at various Z-value thresholds follow
expected binomial distributions.
}

\name{AddCox}
\alias{AddCox}
\title{Add a Covariate Column to an Existing NONMEM dataset}
\description{
A new covariate column can be added to an existing NONMEM dataset.
}
\usage{
  AddCox(nmData, coxData, coxCol, dateCol = "DATE", idCol = "ID")
}
\arguments{
  \item{nmData}{an existing NONMEM dataset}
  \item{coxData}{a data table containing a covariate column}
  \item{coxCol}{the covariate column name in the coxData table}
  \item{dateCol}{date column name in the NONMEM dataset and the covariate data table}
  \item{idCol}{ID column name in the NONMEM dataset and the covariate data table}
}

\details{
  It first carry forward for the missing data. If NA is remained, it carry backward.
}
\value{
  A new NONMEM dataset containing the covariate column
}
\author{Kyun-Seop Bae <k@acr.kr>}
\keyword{Data Preparation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_pairs.R
\name{PanelCor}
\alias{PanelCor}
\title{Correlation Panel for Pairs Plot}
\usage{
PanelCor(x, y, digits = 2, prefix = "", cex.cor)
}
\arguments{
\item{x}{numeric vector}

\item{y}{numeric vector}

\item{digits}{integer, number of digits for correlation}

\item{prefix}{character, prefix for text}

\item{cex.cor}{numeric, character expansion for correlation}
}
\description{
Correlation Panel for Pairs Plot
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{build_dose_records}
\alias{build_dose_records}
\title{Build NONMEM Dose Records}
\usage{
build_dose_records(EX, dose_cmt = 1L)
}
\arguments{
\item{EX}{data.frame with dose information.}

\item{dose_cmt}{default compartment for dose records lacking a CMT column.}
}
\value{
data.frame of dose records with columns
  SUBJID, DAT2, TIME, AMT, RATE, CMT, DV, MDV.
}
\description{
Build NONMEM dose records from an EX-like data.frame.
EX must contain: SUBJID, DAT2, TIME, AMT, RATE.
Optional: CMT (defaults to \code{dose_cmt}).
Output rows have MDV = 1, DV = 0.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_parse.R
\name{ParseOut}
\alias{ParseOut}
\title{Parse Value After Tag from NONMEM Output}
\usage{
ParseOut(TagString, RawRead)
}
\arguments{
\item{TagString}{character, the tag to search for}

\item{RawRead}{character vector of lines from NONMEM output}
}
\value{
character, the value after the tag
}
\description{
Extracts the value following a tag string (after the colon) in NONMEM output lines.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_math.R
\name{LogDV}
\alias{LogDV}
\title{Check if Model Uses Log-Transformed DV (from Control File)}
\usage{
LogDV(CtlFileName)
}
\arguments{
\item{CtlFileName}{character, path to NONMEM control file}
}
\value{
logical, TRUE if LOG(F) is found after $PRED or $ERROR
}
\description{
Check if Model Uses Log-Transformed DV (from Control File)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_data.R
\name{RenCol}
\alias{RenCol}
\title{Rename Columns}
\usage{
RenCol(ColList, OldCol, NewCol)
}
\arguments{
\item{ColList}{character vector of column names}

\item{OldCol}{old column name}

\item{NewCol}{new column name}
}
\value{
character vector with renamed columns
}
\description{
Rename Columns
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_string.R
\name{MatchEnd}
\alias{MatchEnd}
\title{Match End of String}
\usage{
MatchEnd(vName0, End0, IgnoreCase = TRUE)
}
\arguments{
\item{vName0}{character vector of names}

\item{End0}{character string to match at end}

\item{IgnoreCase}{logical, whether to ignore case}
}
\value{
logical vector
}
\description{
Match End of String
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetCountPara}
\alias{GetCountPara}
\title{Get Count of Parameters}
\usage{
GetCountPara()
}
\value{
integer, total number of estimated parameters, or NA if file not found
}
\description{
Get Count of Parameters
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{code_sex}
\alias{code_sex}
\title{Code SEX as Integer}
\usage{
code_sex(x)
}
\arguments{
\item{x}{character or numeric vector of sex values.}
}
\value{
integer vector.
}
\description{
Code SEX as integer: M -> 0, F -> 1, otherwise -> 2.
Numeric input is returned unchanged (cast to integer).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{SESuccess}
\alias{SESuccess}
\title{Check Standard Error Success}
\usage{
SESuccess()
}
\value{
logical, TRUE if standard errors were computed successfully
}
\description{
Check Standard Error Success
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetCountEps}
\alias{GetCountEps}
\title{Get Count of Epsilons}
\usage{
GetCountEps()
}
\value{
integer, number of epsilon parameters, or NA if file not found
}
\description{
Get Count of Epsilons
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_stats.R
\name{NMVarStat}
\alias{NMVarStat}
\title{NONMEM Variable Statistics}
\usage{
NMVarStat(NMTable)
}
\arguments{
\item{NMTable}{data.frame of NONMEM data}
}
\value{
matrix with variable statistics
}
\description{
Computes statistics for each variable in a NONMEM dataset including
counts of NA, zero, positive values, unique values, and functional
dependency checks.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_param.R
\name{nmw_report_param}
\alias{nmw_report_param}
\title{Generate Parameter Estimates Report (S2-Parameters.PDF)}
\usage{
nmw_report_param(run_dir = getwd())
}
\arguments{
\item{run_dir}{character, path to the NONMEM run directory (default: current directory)}
}
\description{
Generates a PDF report with theta, omega, and sigma parameter estimates,
standard errors, confidence intervals, and significance flags.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{crcl_cg}
\alias{crcl_cg}
\title{Cockcroft-Gault Creatinine Clearance}
\usage{
crcl_cg(age, bwt, sex, crea)
}
\arguments{
\item{age}{numeric, age in years.}

\item{bwt}{numeric, body weight in kg.}

\item{sex}{integer/numeric, 0 = male, 1 = female.}

\item{crea}{numeric, serum creatinine in mg/dL.}
}
\value{
numeric vector of CRCL in mL/min.
}
\description{
Cockcroft-Gault creatinine clearance (mL/min).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_stats.R
\name{IntStat}
\alias{IntStat}
\title{Integer Variable Statistics}
\usage{
IntStat(NMTable, VarStat)
}
\arguments{
\item{NMTable}{data.frame}

\item{VarStat}{matrix from NMVarStat}
}
\description{
Prints unique values for all integer variables.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_tests.R
\name{mlr2}
\alias{mlr2}
\title{Multiple Linear Regression with Influence Diagnostics}
\usage{
mlr2(y, x.raw, standardize = 0)
}
\arguments{
\item{y}{numeric vector of response}

\item{x.raw}{data.frame of predictors}

\item{standardize}{integer, standardization method (0=none, 1=center, 2=scale by mean, 3=z-score)}
}
\value{
list with model estimates and influence diagnostics
}
\description{
Performs multiple linear regression and computes influence diagnostics
including DFFITS, DFBETAS, Cook's Distance, COVRATIO, and R-Student.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{FileTag}
\alias{FileTag}
\title{Check for Tag in File}
\usage{
FileTag(FileName, Tag)
}
\arguments{
\item{FileName}{character, filename}

\item{Tag}{character, text to search for}
}
\value{
logical or NA
}
\description{
Check for Tag in File
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetCountUnfixedEps}
\alias{GetCountUnfixedEps}
\title{Get Count of Unfixed Epsilons}
\usage{
GetCountUnfixedEps()
}
\value{
integer, or NA if file not found
}
\description{
Get Count of Unfixed Epsilons
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_output.R
\name{nmw_report_output}
\alias{nmw_report_output}
\title{Generate NONMEM Output Report (SB-Output.PDF)}
\usage{
nmw_report_output(run_dir = getwd())
}
\arguments{
\item{run_dir}{character, path to the NONMEM run directory (default: current directory)}
}
\description{
Includes the NONMEM PRINT.OUT file content in a PDF report.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{MinSuccess}
\alias{MinSuccess}
\title{Check Minimization Success}
\usage{
MinSuccess()
}
\value{
logical, TRUE if minimization was successful
}
\description{
Check Minimization Success
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdf_report.R
\name{ClosePDF}
\alias{ClosePDF}
\title{Close PDF Output}
\usage{
ClosePDF()
}
\description{
Closes the current PDF device.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetProbVal}
\alias{GetProbVal}
\title{Get PROBLEM Statement Value}
\usage{
GetProbVal(Tag)
}
\arguments{
\item{Tag}{character, the tag to search for in FCON}
}
\value{
character, the extracted value
}
\description{
Get PROBLEM Statement Value
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{add_crcl_cg}
\alias{add_crcl_cg}
\title{Add CRCL Column via Cockcroft-Gault}
\usage{
add_crcl_cg(
  df,
  age_col = "AGE",
  bwt_col = "BWT",
  sex_col = "SEX",
  crea_col = "CREA",
  out_col = "CRCL"
)
}
\arguments{
\item{df}{data.frame containing AGE, BWT, SEX, CREA columns.}

\item{age_col}{name of the age column (years). Default "AGE".}

\item{bwt_col}{name of the body-weight column (kg). Default "BWT".}

\item{sex_col}{name of the sex column (0=M, 1=F). Default "SEX".}

\item{crea_col}{name of the creatinine column (mg/dL). Default "CREA".}

\item{out_col}{name of the output column. Default "CRCL".}
}
\value{
data.frame with the CRCL column added.
}
\description{
Add a CRCL column to a NM-style data.frame using Cockcroft-Gault.
Column names are configurable.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pdf_report.R
\name{AddPage}
\alias{AddPage}
\title{Add a New Page to PDF}
\usage{
AddPage(
  Cex = 0.8,
  Header1 = "",
  Header2 = "",
  Header3 = "",
  Footer1 = "",
  Footer2 = "",
  Footer3 = "",
  PrintRowNum = FALSE,
  StartRowNum = 1
)
}
\arguments{
\item{Cex}{numeric, character expansion factor (0.8 or 0.6)}

\item{Header1}{character, left header}

\item{Header2}{character, center header}

\item{Header3}{character, right header}

\item{Footer1}{character, left footer}

\item{Footer2}{character, center footer}

\item{Footer3}{character, right footer}

\item{PrintRowNum}{logical, whether to print row numbers}

\item{StartRowNum}{integer, starting row number}
}
\description{
Creates a new page in the PDF output with optional headers and footers.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report_input.R
\name{nmw_report_input}
\alias{nmw_report_input}
\title{Generate Input Data Summary Report (SA-Input.PDF)}
\usage{
nmw_report_input(run_dir = getwd())
}
\arguments{
\item{run_dir}{character, path to the NONMEM run directory (default: current directory)}
}
\description{
Generates a PDF report with input data validation including
individuals with no dosing/DV, duplicate records, and summary statistics.
}

\name{TabStep}
\alias{TabStep}
\title{Table Step}
\description{This produces standard table.}
\usage{TabStep()}
\details{
It does not have arguments.
All necessary arguments are stored in the \code{e} environment.
This is similar to other standard results table.
}
\value{
  A table with ID, TIME, DV, PRED, RES, WRES, derivatives of G and H.
  If the estimation method is other than 'ZERO'(First-order approximation), it includes CWRES, CIPREDI(formerly IPRED), CIRESI(formerly IRES). 
}
\references{NONMEM Users Guide}
\seealso{\code{\link{EstStep}}}
\author{Kyun-Seop Bae <k@acr.kr>}
\examples{
# Only After EstStep
#TabStep()
}
\keyword{Tabulation Step}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetEstMethod}
\alias{GetEstMethod}
\title{Get Estimation Method}
\usage{
GetEstMethod()
}
\value{
character, abbreviation of the estimation method
}
\description{
Identifies the estimation method from the .ext file header.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_summary.R
\name{GetModelNames}
\alias{GetModelNames}
\title{Get Model Names from Folder Names}
\usage{
GetModelNames(vName)
}
\arguments{
\item{vName}{character vector of folder names}
}
\value{
character vector of model names
}
\description{
Extracts model names by splitting folder names on period.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_summary.R
\name{SumOut1}
\alias{SumOut1}
\title{Summarize Single NONMEM Run}
\usage{
SumOut1(ModelName)
}
\arguments{
\item{ModelName}{character, model name}
}
\value{
data.frame with one row of summary metrics
}
\description{
Collects key metrics from a single NONMEM run folder.
}
\keyword{internal}

\name{EstStep}
\alias{EstStep}
\title{Estimation Step}
\description{This estimates upon the conditions with \code{InitStep}.}
\usage{EstStep()}
\details{
It does not have arguments.
All necessary arguments are stored in the \code{e} environment.
It assumes "INTERACTION" between eta and epsilon for \code{"COND"} and \code{"LAPL"} options.
The output is basically same to NONMEM output.
}
\value{
  \item{Initial OFV}{initial value of the objective function}
  \item{Time}{time consumed for this step}
  \item{Optim}{the raw output from \code{optim} function}
  \item{Final Estimates}{final estimates in the original scale}
}
\references{NONMEM Users Guide}
\seealso{\code{\link{InitStep}}}
\author{Kyun-Seop Bae <k@acr.kr>}
\examples{
# Only After InitStep
#EstStep()
}
\keyword{Estimation Step}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_parse.R
\name{BtwTagVals}
\alias{BtwTagVals}
\title{Extract Vector of Values Between XML Tags}
\usage{
BtwTagVals(Tag, RawRead)
}
\arguments{
\item{Tag}{character, the tag name (e.g., "nm:theta")}

\item{RawRead}{character vector of XML lines}
}
\value{
numeric vector of values
}
\description{
Parses NONMEM XML output to extract theta, omega, or sigma vectors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_pairs.R
\name{EBEpair}
\alias{EBEpair}
\title{EBE Pair Plot}
\usage{
EBEpair(tabEta, RunNumber, each = FALSE)
}
\arguments{
\item{tabEta}{data.frame with ID, covariates, and ETA columns}

\item{RunNumber}{character, model/run identifier}

\item{each}{logical, if TRUE creates separate plots per covariate}
}
\description{
Creates pairs plot of ETAs vs covariates with histograms and correlations.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{nm_to_molar}
\alias{nm_to_molar}
\title{Convert NM Dataset from Mass to Molar Units}
\usage{
nm_to_molar(
  df,
  mw_dose,
  mw_dv = mw_dose,
  amt_cols = c("AMT", "RATE"),
  dv_col = "DV",
  cmt_col = "CMT",
  dose_factor = 1000,
  dv_factor = 1
)
}
\arguments{
\item{df}{data.frame.}

\item{mw_dose}{numeric scalar.  MW for AMT and RATE columns.}

\item{mw_dv}{numeric scalar OR named numeric.
Scalar: applied to all DV regardless of CMT.
Named: names are CMT values; rows whose CMT is not in names() are left
unchanged.  Default = mw_dose.}

\item{amt_cols}{columns to convert with mw_dose.  Default c("AMT", "RATE").}

\item{dv_col}{DV column.  Default "DV".}

\item{cmt_col}{CMT column.  Default "CMT".}

\item{dose_factor}{multiplier applied after AMT/RATE divided by mw_dose.
1000 = mg -> umol (default).}

\item{dv_factor}{multiplier applied after DV divided by mw_dv.
1 = ng/mL -> umol/L (default).}
}
\value{
data.frame with converted columns; other columns unchanged.
}
\description{
Convert AMT/RATE/DV columns of a NM data.frame from mass units to molar units.
}
\details{
Typical use:
\itemize{
  \item AMT, RATE in mg, DV in ng/mL, MW in g/mol
  \item mg / MW * 1000  ->  umol
  \item ng/mL / MW      ->  umol/L
}

For multiple analytes (e.g., parent at CMT=1 and metabolite at CMT=3),
supply \code{mw_dv} as a named numeric vector with names equal to the CMT
values (as character):  \code{c(`1` = 394.47, `3` = 380.47)}.
}

\name{CombDmExPc}
\alias{CombDmExPc}
\title{Combine the demographics(DM), dosing(EX), and DV(PC) tables into a new NONMEM dataset}
\description{
A new NONMEM dataset can be created from the demographics, dosing, and DV tables.
}
\usage{
  CombDmExPc(dm, ex, pc)
}
\arguments{
  \item{dm}{A demographics table. It should contain a row per subject.}
  \item{ex}{An exposure table. Drug administration (dosing) history table.}
  \item{pc}{A DV(dependent variable) or PC(drug concentration) table}
}
\details{
  Combining a demographics, a dosing, and a concentration table can produce a new NONMEM dataset.
}
\value{
  A new NONMEM dataset
}
\author{Kyun-Seop Bae <k@acr.kr>}
\keyword{Data Preparation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_report.R
\name{MergeIDStatOFV}
\alias{MergeIDStatOFV}
\title{Merge IDStat with Individual OFV from PHI}
\usage{
MergeIDStatOFV(FDATA, PHI)
}
\arguments{
\item{FDATA}{data.frame of NONMEM data}

\item{PHI}{data.frame from .phi file}
}
\value{
list with IDStat3 (merged), VarStat, IDStat
}
\description{
Common pattern used in report_ofv.R and report_resid.R.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetCountObs}
\alias{GetCountObs}
\title{Get Count of Observations}
\usage{
GetCountObs()
}
\value{
integer, total number of observation records
}
\description{
Get Count of Observations
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_report.R
\name{ReadEXTFile}
\alias{ReadEXTFile}
\title{Read EXT File from Current Directory}
\usage{
ReadEXTFile()
}
\value{
data.frame or NULL if file not found
}
\description{
Read EXT File from Current Directory
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_read.R
\name{GetAICc}
\alias{GetAICc}
\title{Get Corrected AIC (AICc)}
\usage{
GetAICc()
}
\value{
numeric, the corrected AIC value
}
\description{
Get Corrected AIC (AICc)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_math.R
\name{LogDV2}
\alias{LogDV2}
\title{Check if Model Uses Log-Transformed DV (from FSUBS)}
\usage{
LogDV2()
}
\value{
logical, TRUE if LOG(F) is found after SUBROUTINE PRED or ERROR in FSUBS
}
\description{
Check if Model Uses Log-Transformed DV (from FSUBS)
}

\name{CovStep}
\alias{CovStep}
\title{Covariance Step}
\description{
It calculates standard errors and various variance matrices with the \code{e$FinalPara} after estimation step.
}
\usage{CovStep()}
\details{
Because \code{EstStep} uses nonlinear optimization, covariance step is separated from estimation step.
It calculates variance-covariance matrix of estimates in the original scale.
}
\value{
  \item{Time}{consumed time}
  \item{Standard Error}{standard error of the estimates in the order of theta, omega, and sigma}
  \item{Covariance Matrix of Estimates}{covariance matrix of estimates in the order of theta, omega, and sigma. This is inverse(R) x S x inverse(R) by default.}
  \item{Correlation Matrix of Estimates}{correlation matrix of estimates in the order of theta, omega, and sigma}
  \item{Inverse Covariance Matrix of Estimates}{inverse covariance matrix of estimates in the order of theta, omega, and sigma}
  \item{Eigen Values}{eigen values of covariance matrix}
  \item{R Matrix}{R matrix of NONMEM, the second derivative of log likelihood function with respect to estimation parameters}
  \item{S Matrix}{S matrix of NONMEM, sum of individual cross-product of the first derivative of log likelihood function with respect to estimation parameters}
}
\references{NONMEM Users Guide}
\author{Kyun-Seop Bae <k@acr.kr>}
\seealso{
\code{\link{EstStep}}, \code{\link{InitStep}} 
}
\examples{
# Only after InitStep and EstStep
#CovStep()
}
\keyword{Covariance Step}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils_report.R
\name{PlotDistribution}
\alias{PlotDistribution}
\title{Plot Distribution Panel (Summary + Histogram + QQ)}
\usage{
PlotDistribution(
  var.data,
  var.name,
  show_shrinkage = FALSE,
  omega_var = NULL,
  show_ttest = FALSE,
  show_cv = FALSE
)
}
\arguments{
\item{var.data}{numeric vector of data}

\item{var.name}{character, variable name for labeling}

\item{show_shrinkage}{logical, whether to show shrinkage}

\item{omega_var}{numeric, omega variance for shrinkage calc}

\item{show_ttest}{logical, whether to show t-test p-value}

\item{show_cv}{logical, whether to show CV percent}
}
\description{
Common 3-panel distribution plot used in report_ebe.R,
report_indipk.R, and report_resid.R.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nm_helpers.R
\name{build_nm_dataset}
\alias{build_nm_dataset}
\title{Build a NONMEM-Format Dataset from DM/EX/PC}
\usage{
build_nm_dataset(
  DM,
  EX,
  PC,
  IDs = NULL,
  id_prefix = "",
  id_func = NULL,
  dose_cmt = 1L,
  verbose = FALSE
)
}
\arguments{
\item{DM}{data.frame with one row per subject.  Must contain SUBJID; any
other columns are merged in as subject-level constants.}

\item{EX}{dose records (see \code{\link{build_dose_records}}).}

\item{PC}{observation records (see \code{\link{build_obs_records}}).}

\item{IDs}{character vector of SUBJIDs to keep.  NULL (default) is the
intersect of EX and PC where DV > 0.}

\item{id_prefix}{character.  When \code{id_func} is NULL,
ID = paste0(id_prefix, SUBJID).}

\item{id_func}{function(SUBJID) -> character ID.  Overrides \code{id_prefix}.}

\item{dose_cmt}{default compartment for dose records lacking a CMT column.}

\item{verbose}{print progress. Default FALSE.}
}
\value{
data.frame with columns
  ID, SUBJID, DAT2, TIME, AMT, RATE, CMT, DV, MDV, <DM-columns ...>.
}
\description{
Build a NONMEM-format dataset from cleaned DM/EX/PC tables.
Time-varying covariates (VS, LB, ...) are NOT merged here.  Use
\code{\link{merge_cov_locf}} afterwards to add them.  CRCL is added
separately via \code{\link{add_crcl_cg}}.  Molar conversion is done with
\code{\link{nm_to_molar}}.
}
\details{
Records are sorted by SUBJID, DAT2, TIME, CMT, MDV, AMT.  At tied
(SUBJID, DAT2, TIME, CMT, MDV), AMT = 0 (observation) sorts before AMT > 0
(dose) so a pre-dose observation precedes the dose given at the same minute.
}
